/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <string.h>

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikawidgets/pikawidgets.h"

#include "actions-types.h"

#include "widgets/pikacontainerview.h"
#include "widgets/pikacontainerview-utils.h"
#include "widgets/pikadialogfactory.h"
#include "widgets/pikadock.h"
#include "widgets/pikadockable.h"
#include "widgets/pikadockbook.h"
#include "widgets/pikadocked.h"
#include "widgets/pikasessioninfo.h"

#include "dockable-commands.h"


static PikaDockable * dockable_get_current (PikaDockbook *dockbook);


/*  public functions  */

void
dockable_add_tab_cmd_callback (PikaAction *action,
                               GVariant   *value,
                               gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);

  pika_dockbook_add_from_dialog_factory (dockbook,
                                         g_variant_get_string (value, NULL));
}

void
dockable_close_tab_cmd_callback (PikaAction *action,
                                 GVariant   *value,
                                 gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable = dockable_get_current (dockbook);

  if (dockable)
    gtk_container_remove (GTK_CONTAINER (dockbook),
                          GTK_WIDGET (dockable));
}

void
dockable_detach_tab_cmd_callback (PikaAction *action,
                                  GVariant   *value,
                                  gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable = dockable_get_current (dockbook);

  if (dockable)
    pika_dockable_detach (dockable);
}

void
dockable_lock_tab_cmd_callback (PikaAction *action,
                                GVariant   *value,
                                gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable = dockable_get_current (dockbook);

  if (dockable)
    {
      gboolean lock = g_variant_get_boolean (value);

      pika_dockable_set_locked (dockable, lock);
    }
}

void
dockable_toggle_view_cmd_callback (PikaAction *action,
                                   GVariant   *value,
                                   gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable;
  PikaViewType  view_type;
  gint          page_num;

  view_type = (PikaViewType) g_variant_get_int32 (value);

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (dockbook));

  dockable = (PikaDockable *)
    gtk_notebook_get_nth_page (GTK_NOTEBOOK (dockbook), page_num);

  if (dockable)
    {
      PikaDialogFactoryEntry *entry;

      pika_dialog_factory_from_widget (GTK_WIDGET (dockable), &entry);

      if (entry)
        {
          gchar *identifier;
          gchar *substring = NULL;

          identifier = g_strdup (entry->identifier);

          substring = strstr (identifier, "grid");

          if (substring && view_type == PIKA_VIEW_TYPE_GRID)
            {
              g_free (identifier);
              return;
            }

          if (! substring)
            {
              substring = strstr (identifier, "list");

              if (substring && view_type == PIKA_VIEW_TYPE_LIST)
                {
                  g_free (identifier);
                  return;
                }
            }

          if (substring)
            {
              PikaContainerView *old_view;
              GtkWidget         *new_dockable;
              PikaDock          *dock;
              gint               view_size = -1;

              if (view_type == PIKA_VIEW_TYPE_LIST)
                memcpy (substring, "list", 4);
              else if (view_type == PIKA_VIEW_TYPE_GRID)
                memcpy (substring, "grid", 4);

              old_view = pika_container_view_get_by_dockable (dockable);

              if (old_view)
                view_size = pika_container_view_get_view_size (old_view, NULL);

              dock         = pika_dockbook_get_dock (dockbook);
              new_dockable = pika_dialog_factory_dockable_new (pika_dock_get_dialog_factory (dock),
                                                               dock,
                                                               identifier,
                                                               view_size);

              if (new_dockable)
                {
                  PikaDocked *old;
                  PikaDocked *new;
                  gboolean    show;

                  pika_dockable_set_locked (PIKA_DOCKABLE (new_dockable),
                                            pika_dockable_get_locked (dockable));

                  old = PIKA_DOCKED (gtk_bin_get_child (GTK_BIN (dockable)));
                  new = PIKA_DOCKED (gtk_bin_get_child (GTK_BIN (new_dockable)));

                  show = pika_docked_get_show_button_bar (old);
                  pika_docked_set_show_button_bar (new, show);

                  /*  Maybe pika_dialog_factory_dockable_new() returned
                   *  an already existing singleton dockable, so check
                   *  if it already is attached to a dockbook.
                   */
                  if (! pika_dockable_get_dockbook (PIKA_DOCKABLE (new_dockable)))
                    {
                      gtk_notebook_insert_page (GTK_NOTEBOOK (dockbook),
                                                new_dockable, NULL,
                                                page_num);
                      gtk_widget_show (new_dockable);

                      gtk_container_remove (GTK_CONTAINER (dockbook),
                                            GTK_WIDGET (dockable));

                      gtk_notebook_set_current_page (GTK_NOTEBOOK (dockbook),
                                                     page_num);
                   }
                }
            }

          g_free (identifier);
        }
    }
}

void
dockable_view_size_cmd_callback (PikaAction *action,
                                 GVariant   *value,
                                 gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable = dockable_get_current (dockbook);
  gint          view_size;

  view_size = g_variant_get_int32 (value);

  if (dockable)
    {
      PikaContainerView *view = pika_container_view_get_by_dockable (dockable);

      if (view)
        {
          gint old_size;
          gint border_width;

          old_size = pika_container_view_get_view_size (view, &border_width);

          if (old_size != view_size)
            pika_container_view_set_view_size (view, view_size, border_width);
        }
    }
}

void
dockable_tab_style_cmd_callback (PikaAction *action,
                                 GVariant   *value,
                                 gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable = dockable_get_current (dockbook);
  PikaTabStyle  tab_style;

  tab_style = (PikaTabStyle) g_variant_get_int32 (value);

  if (dockable && pika_dockable_get_tab_style (dockable) != tab_style)
    {
      GtkWidget *tab_widget;

      pika_dockable_set_tab_style (dockable, tab_style);

      tab_widget = pika_dockbook_create_tab_widget (dockbook, dockable);

      gtk_notebook_set_tab_label (GTK_NOTEBOOK (dockbook),
                                  GTK_WIDGET (dockable),
                                  tab_widget);
    }
}

void
dockable_show_button_bar_cmd_callback (PikaAction *action,
                                       GVariant   *value,
                                       gpointer    data)
{
  PikaDockbook *dockbook = PIKA_DOCKBOOK (data);
  PikaDockable *dockable = dockable_get_current (dockbook);

  if (dockable)
    {
      PikaDocked *docked;
      gboolean    show;

      docked = PIKA_DOCKED (gtk_bin_get_child (GTK_BIN (dockable)));
      show   = g_variant_get_boolean (value);

      pika_docked_set_show_button_bar (docked, show);
    }
}


/*  private functions  */

static PikaDockable *
dockable_get_current (PikaDockbook *dockbook)
{
  gint page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (dockbook));

  return (PikaDockable *) gtk_notebook_get_nth_page (GTK_NOTEBOOK (dockbook),
                                                     page_num);
}
