/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikawidgets/pikawidgets.h"

#include "actions-types.h"

#include "core/pikacontext.h"
#include "core/pikadata.h"

#include "widgets/pikaactiongroup.h"
#include "widgets/pikahelp-ids.h"

#include "actions.h"
#include "data-commands.h"
#include "gradients-actions.h"
#include "gradients-commands.h"

#include "pika-intl.h"


static const PikaActionEntry gradients_actions[] =
{
  { "gradients-new", PIKA_ICON_DOCUMENT_NEW,
    NC_("gradients-action", "_New Gradient"), NULL, { NULL },
    NC_("gradients-action", "Create a new gradient"),
    data_new_cmd_callback,
    PIKA_HELP_GRADIENT_NEW },

  { "gradients-duplicate", PIKA_ICON_OBJECT_DUPLICATE,
    NC_("gradients-action", "D_uplicate Gradient"), NULL, { NULL },
    NC_("gradients-action", "Duplicate this gradient"),
    data_duplicate_cmd_callback,
    PIKA_HELP_GRADIENT_DUPLICATE },

  { "gradients-copy-location", PIKA_ICON_EDIT_COPY,
    NC_("gradients-action", "Copy Gradient _Location"), NULL, { NULL },
    NC_("gradients-action", "Copy gradient file location to clipboard"),
    data_copy_location_cmd_callback,
    PIKA_HELP_GRADIENT_COPY_LOCATION },

  { "gradients-show-in-file-manager", PIKA_ICON_FILE_MANAGER,
    NC_("gradients-action", "Show in _File Manager"), NULL, { NULL },
    NC_("gradients-action", "Show gradient file location in the file manager"),
    data_show_in_file_manager_cmd_callback,
    PIKA_HELP_GRADIENT_SHOW_IN_FILE_MANAGER },

  { "gradients-save-as-pov", PIKA_ICON_DOCUMENT_SAVE_AS,
    NC_("gradients-action", "Save as _POV-Ray..."), NULL, { NULL },
    NC_("gradients-action", "Save gradient as POV-Ray"),
    gradients_save_as_pov_ray_cmd_callback,
    PIKA_HELP_GRADIENT_SAVE_AS_POV },

  { "gradients-delete", PIKA_ICON_EDIT_DELETE,
    NC_("gradients-action", "_Delete Gradient"), NULL, { NULL },
    NC_("gradients-action", "Delete this gradient"),
    data_delete_cmd_callback,
    PIKA_HELP_GRADIENT_DELETE },

  { "gradients-refresh", PIKA_ICON_VIEW_REFRESH,
    NC_("gradients-action", "_Refresh Gradients"), NULL, { NULL },
    NC_("gradients-action", "Refresh gradients"),
    data_refresh_cmd_callback,
    PIKA_HELP_GRADIENT_REFRESH }
};

static const PikaStringActionEntry gradients_edit_actions[] =
{
  { "gradients-edit", PIKA_ICON_EDIT,
    NC_("gradients-action", "_Edit Gradient..."), NULL, { NULL },
    NC_("gradients-action", "Edit this gradient"),
    "pika-gradient-editor",
    PIKA_HELP_GRADIENT_EDIT }
};


void
gradients_actions_setup (PikaActionGroup *group)
{
  pika_action_group_add_actions (group, "gradients-action",
                                 gradients_actions,
                                 G_N_ELEMENTS (gradients_actions));

  pika_action_group_add_string_actions (group, "gradients-action",
                                        gradients_edit_actions,
                                        G_N_ELEMENTS (gradients_edit_actions),
                                        data_edit_cmd_callback);
}

void
gradients_actions_update (PikaActionGroup *group,
                          gpointer         user_data)
{
  PikaContext  *context  = action_data_get_context (user_data);
  PikaGradient *gradient = NULL;
  PikaData     *data     = NULL;
  GFile        *file     = NULL;

  if (context)
    {
      gradient = pika_context_get_gradient (context);

      if (action_data_sel_count (user_data) > 1)
        {
          gradient = NULL;
        }

      if (gradient)
        {
          data = PIKA_DATA (gradient);

          file = pika_data_get_file (data);
        }
    }

#define SET_SENSITIVE(action,condition) \
        pika_action_group_set_action_sensitive (group, action, (condition) != 0, NULL)

  SET_SENSITIVE ("gradients-edit",                 gradient);
  SET_SENSITIVE ("gradients-duplicate",            gradient);
  SET_SENSITIVE ("gradients-save-as-pov",          gradient);
  SET_SENSITIVE ("gradients-copy-location",        file);
  SET_SENSITIVE ("gradients-show-in-file-manager", file);
  SET_SENSITIVE ("gradients-delete",               gradient && pika_data_is_deletable (data));

#undef SET_SENSITIVE
}
