/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * gimphistogram module Copyright (C) 1999 Jay Cox <jaycox@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <cairo.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "core-types.h"

#include "gegl/pika-gegl-nodes.h"
#include "gegl/pikatilehandlervalidate.h"

#include "pikaasync.h"
#include "pikachannel.h"
#include "pikadrawable-filters.h"
#include "pikadrawable-histogram.h"
#include "pikahistogram.h"
#include "pikaimage.h"
#include "pikaprojectable.h"


/*  local function prototypes  */

static PikaAsync * pika_drawable_calculate_histogram_internal (PikaDrawable  *drawable,
                                                               PikaHistogram *histogram,
                                                               gboolean       with_filters,
                                                               gboolean       run_async);


/*  private functions  */


static PikaAsync *
pika_drawable_calculate_histogram_internal (PikaDrawable  *drawable,
                                            PikaHistogram *histogram,
                                            gboolean       with_filters,
                                            gboolean       run_async)
{
  PikaAsync   *async = NULL;
  PikaImage   *image;
  PikaChannel *mask;
  gint         x, y, width, height;

  if (! pika_item_mask_intersect (PIKA_ITEM (drawable), &x, &y, &width, &height))
    goto end;

  image = pika_item_get_image (PIKA_ITEM (drawable));
  mask  = pika_image_get_mask (image);

  if (FALSE)
    {
      GeglNode      *node = gegl_node_new ();
      GeglNode      *source;
      GeglNode      *histogram_sink;
      GeglProcessor *processor;

      if (with_filters)
        {
          source = pika_drawable_get_source_node (drawable);
        }
      else
        {
          source =
            pika_gegl_add_buffer_source (node,
                                         pika_drawable_get_buffer (drawable),
                                         0, 0);
        }

      histogram_sink =
        gegl_node_new_child (node,
                             "operation", "pika:histogram-sink",
                             "histogram", histogram,
                             NULL);

      gegl_node_link (source, histogram_sink);

      if (! pika_channel_is_empty (mask))
        {
          GeglNode *mask_source;
          gint      off_x, off_y;

          g_printerr ("adding mask aux\n");

          pika_item_get_offset (PIKA_ITEM (drawable), &off_x, &off_y);

          mask_source =
            pika_gegl_add_buffer_source (node,
                                         pika_drawable_get_buffer (PIKA_DRAWABLE (mask)),
                                         -off_x, -off_y);

          gegl_node_connect (mask_source, "output", histogram_sink, "aux");
        }

      processor = gegl_node_new_processor (histogram_sink,
                                           GEGL_RECTANGLE (x, y, width, height));

      while (gegl_processor_work (processor, NULL));

      g_object_unref (processor);
      g_object_unref (node);
    }
  else
    {
      GeglBuffer      *buffer      = pika_drawable_get_buffer (drawable);
      PikaProjectable *projectable = NULL;

      if (with_filters && pika_drawable_has_filters (drawable))
        {
          PikaTileHandlerValidate *validate;
          GeglNode                *node;

          node = pika_drawable_get_source_node (drawable);

          buffer = gegl_buffer_new (gegl_buffer_get_extent (buffer),
                                    gegl_buffer_get_format (buffer));

          validate =
            PIKA_TILE_HANDLER_VALIDATE (pika_tile_handler_validate_new (node));

          pika_tile_handler_validate_assign (validate, buffer);

          g_object_unref (validate);

          pika_tile_handler_validate_invalidate (validate,
                                                 gegl_buffer_get_extent (buffer));

#if 0
          /*  this would keep the buffer updated across drawable or
           *  filter changes, but the histogram is created in one go
           *  and doesn't need the signal connection
           */
          g_signal_connect_object (node, "invalidated",
                                   G_CALLBACK (pika_tile_handler_validate_invalidate),
                                   validate, G_CONNECT_SWAPPED);
#endif

          if (PIKA_IS_PROJECTABLE (drawable))
            projectable = PIKA_PROJECTABLE (drawable);
        }
      else
        {
          g_object_ref (buffer);
        }

      if (projectable)
        pika_projectable_begin_render (projectable);

      if (! pika_channel_is_empty (mask))
        {
          gint off_x, off_y;

          pika_item_get_offset (PIKA_ITEM (drawable), &off_x, &off_y);

          if (run_async)
            {
              async = pika_histogram_calculate_async (
                histogram, buffer,
                GEGL_RECTANGLE (x, y, width, height),
                pika_drawable_get_buffer (PIKA_DRAWABLE (mask)),
                GEGL_RECTANGLE (x + off_x, y + off_y,
                                width, height));
            }
          else
            {
              pika_histogram_calculate (
                histogram, buffer,
                GEGL_RECTANGLE (x, y, width, height),
                pika_drawable_get_buffer (PIKA_DRAWABLE (mask)),
                GEGL_RECTANGLE (x + off_x, y + off_y,
                                width, height));
            }
        }
      else
        {
          if (run_async)
            {
              async = pika_histogram_calculate_async (
                histogram, buffer,
                GEGL_RECTANGLE (x, y, width, height),
                NULL, NULL);
            }
          else
            {
              pika_histogram_calculate (
                histogram, buffer,
                GEGL_RECTANGLE (x, y, width, height),
                NULL, NULL);
            }
        }

      if (projectable)
        pika_projectable_end_render (projectable);

      g_object_unref (buffer);
    }

end:
  if (run_async && ! async)
    {
      async = pika_async_new ();

      pika_async_finish (async, NULL);
    }

  return async;
}


/*  public functions  */


void
pika_drawable_calculate_histogram (PikaDrawable  *drawable,
                                   PikaHistogram *histogram,
                                   gboolean       with_filters)
{
  g_return_if_fail (PIKA_IS_DRAWABLE (drawable));
  g_return_if_fail (pika_item_is_attached (PIKA_ITEM (drawable)));
  g_return_if_fail (histogram != NULL);

  pika_drawable_calculate_histogram_internal (drawable,
                                              histogram, with_filters,
                                              FALSE);
}

PikaAsync *
pika_drawable_calculate_histogram_async (PikaDrawable  *drawable,
                                         PikaHistogram *histogram,
                                         gboolean       with_filters)
{
  g_return_val_if_fail (PIKA_IS_DRAWABLE (drawable), NULL);
  g_return_val_if_fail (pika_item_is_attached (PIKA_ITEM (drawable)), NULL);
  g_return_val_if_fail (histogram != NULL, NULL);

  return pika_drawable_calculate_histogram_internal (drawable,
                                                     histogram, with_filters,
                                                     TRUE);
}
