/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikadrawable-operation.c
 * Copyright (C) 2007 Øyvind Kolås <pippin@gimp.org>
 *                    Sven Neumann <sven@gimp.org>
 *                    Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "core-types.h"

#include "gegl/pika-gegl-utils.h"

#include "operations/pika-operation-config.h"
#include "operations/pikaoperationsettings.h"

#include "pikadrawable.h"
#include "pikadrawable-operation.h"
#include "pikadrawablefilter.h"
#include "pikaprogress.h"
#include "pikasettings.h"


/*  public functions  */

void
pika_drawable_apply_operation (PikaDrawable *drawable,
                               PikaProgress *progress,
                               const gchar  *undo_desc,
                               GeglNode     *operation)
{
  pika_drawable_apply_operation_with_config (drawable,
                                             progress, undo_desc,
                                             operation, NULL);
}

void
pika_drawable_apply_operation_with_config (PikaDrawable *drawable,
                                           PikaProgress *progress,
                                           const gchar  *undo_desc,
                                           GeglNode     *operation,
                                           GObject      *config)
{
  PikaDrawableFilter *filter;

  g_return_if_fail (PIKA_IS_DRAWABLE (drawable));
  g_return_if_fail (pika_item_is_attached (PIKA_ITEM (drawable)));
  g_return_if_fail (progress == NULL || PIKA_IS_PROGRESS (progress));
  g_return_if_fail (undo_desc != NULL);
  g_return_if_fail (GEGL_IS_NODE (operation));
  g_return_if_fail (config == NULL || PIKA_IS_OPERATION_SETTINGS (config));

  if (! pika_item_mask_intersect (PIKA_ITEM (drawable),
                                  NULL, NULL, NULL, NULL))
    {
      return;
    }

  filter = pika_drawable_filter_new (drawable, undo_desc, operation, NULL);

  pika_drawable_filter_set_add_alpha (filter,
                                      pika_gegl_node_has_key (operation,
                                                              "needs-alpha"));

  if (config)
    {
      pika_operation_config_sync_node (config, operation);

      pika_operation_settings_sync_drawable_filter (
        PIKA_OPERATION_SETTINGS (config), filter);
    }

  pika_drawable_filter_apply  (filter, NULL);
  pika_drawable_filter_commit (filter, progress, TRUE);

  g_object_unref (filter);

  if (progress)
    pika_progress_end (progress);
}

void
pika_drawable_apply_operation_by_name (PikaDrawable *drawable,
                                       PikaProgress *progress,
                                       const gchar  *undo_desc,
                                       const gchar  *operation_type,
                                       GObject      *config)
{
  GeglNode *node;

  g_return_if_fail (PIKA_IS_DRAWABLE (drawable));
  g_return_if_fail (pika_item_is_attached (PIKA_ITEM (drawable)));
  g_return_if_fail (progress == NULL || PIKA_IS_PROGRESS (progress));
  g_return_if_fail (undo_desc != NULL);
  g_return_if_fail (operation_type != NULL);
  g_return_if_fail (config == NULL || PIKA_IS_SETTINGS (config));

  node = g_object_new (GEGL_TYPE_NODE,
                       "operation", operation_type,
                       NULL);

  if (config)
    gegl_node_set (node,
                   "config", config,
                   NULL);

  pika_drawable_apply_operation (drawable, progress, undo_desc, node);

  g_object_unref (node);
}
