/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"

#include "core-types.h"

#include "gegl/pika-gegl-utils.h"

#include "pika-memsize.h"
#include "pikaimage.h"
#include "pikadrawable.h"
#include "pikadrawablemodundo.h"


enum
{
  PROP_0,
  PROP_COPY_BUFFER
};


static void     pika_drawable_mod_undo_constructed  (GObject             *object);
static void     pika_drawable_mod_undo_set_property (GObject             *object,
                                                     guint                property_id,
                                                     const GValue        *value,
                                                     GParamSpec          *pspec);
static void     pika_drawable_mod_undo_get_property (GObject             *object,
                                                     guint                property_id,
                                                     GValue              *value,
                                                     GParamSpec          *pspec);

static gint64   pika_drawable_mod_undo_get_memsize  (PikaObject          *object,
                                                     gint64              *gui_size);

static void     pika_drawable_mod_undo_pop          (PikaUndo            *undo,
                                                     PikaUndoMode         undo_mode,
                                                     PikaUndoAccumulator *accum);
static void     pika_drawable_mod_undo_free         (PikaUndo            *undo,
                                                     PikaUndoMode         undo_mode);


G_DEFINE_TYPE (PikaDrawableModUndo, pika_drawable_mod_undo, PIKA_TYPE_ITEM_UNDO)

#define parent_class pika_drawable_mod_undo_parent_class


static void
pika_drawable_mod_undo_class_init (PikaDrawableModUndoClass *klass)
{
  GObjectClass    *object_class      = G_OBJECT_CLASS (klass);
  PikaObjectClass *pika_object_class = PIKA_OBJECT_CLASS (klass);
  PikaUndoClass   *undo_class        = PIKA_UNDO_CLASS (klass);

  object_class->constructed      = pika_drawable_mod_undo_constructed;
  object_class->set_property     = pika_drawable_mod_undo_set_property;
  object_class->get_property     = pika_drawable_mod_undo_get_property;

  pika_object_class->get_memsize = pika_drawable_mod_undo_get_memsize;

  undo_class->pop                = pika_drawable_mod_undo_pop;
  undo_class->free               = pika_drawable_mod_undo_free;

  g_object_class_install_property (object_class, PROP_COPY_BUFFER,
                                   g_param_spec_boolean ("copy-buffer",
                                                         NULL, NULL,
                                                         FALSE,
                                                         PIKA_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY));
}

static void
pika_drawable_mod_undo_init (PikaDrawableModUndo *undo)
{
}

static void
pika_drawable_mod_undo_constructed (GObject *object)
{
  PikaDrawableModUndo *drawable_mod_undo = PIKA_DRAWABLE_MOD_UNDO (object);
  PikaItem            *item;
  PikaDrawable        *drawable;

  G_OBJECT_CLASS (parent_class)->constructed (object);

  pika_assert (PIKA_IS_DRAWABLE (PIKA_ITEM_UNDO (object)->item));

  item     = PIKA_ITEM_UNDO (object)->item;
  drawable = PIKA_DRAWABLE (item);

  if (drawable_mod_undo->copy_buffer)
    {
      drawable_mod_undo->buffer =
        pika_gegl_buffer_dup (pika_drawable_get_buffer (drawable));
    }
  else
    {
      drawable_mod_undo->buffer =
        g_object_ref (pika_drawable_get_buffer (drawable));
    }

  pika_item_get_offset (item,
                        &drawable_mod_undo->offset_x,
                        &drawable_mod_undo->offset_y);
}

static void
pika_drawable_mod_undo_set_property (GObject      *object,
                                     guint         property_id,
                                     const GValue *value,
                                     GParamSpec   *pspec)
{
  PikaDrawableModUndo *drawable_mod_undo = PIKA_DRAWABLE_MOD_UNDO (object);

  switch (property_id)
    {
    case PROP_COPY_BUFFER:
      drawable_mod_undo->copy_buffer = g_value_get_boolean (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_drawable_mod_undo_get_property (GObject    *object,
                                     guint       property_id,
                                     GValue     *value,
                                     GParamSpec *pspec)
{
  PikaDrawableModUndo *drawable_mod_undo = PIKA_DRAWABLE_MOD_UNDO (object);

  switch (property_id)
    {
    case PROP_COPY_BUFFER:
      g_value_set_boolean (value, drawable_mod_undo->copy_buffer);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static gint64
pika_drawable_mod_undo_get_memsize (PikaObject *object,
                                    gint64     *gui_size)
{
  PikaDrawableModUndo *drawable_mod_undo = PIKA_DRAWABLE_MOD_UNDO (object);
  gint64               memsize           = 0;

  memsize += pika_gegl_buffer_get_memsize (drawable_mod_undo->buffer);

  return memsize + PIKA_OBJECT_CLASS (parent_class)->get_memsize (object,
                                                                  gui_size);
}

static void
pika_drawable_mod_undo_pop (PikaUndo            *undo,
                            PikaUndoMode         undo_mode,
                            PikaUndoAccumulator *accum)
{
  PikaDrawableModUndo *drawable_mod_undo = PIKA_DRAWABLE_MOD_UNDO (undo);
  PikaDrawable        *drawable          = PIKA_DRAWABLE (PIKA_ITEM_UNDO (undo)->item);
  GeglBuffer          *buffer;
  gint                 offset_x;
  gint                 offset_y;

  PIKA_UNDO_CLASS (parent_class)->pop (undo, undo_mode, accum);

  buffer   = drawable_mod_undo->buffer;
  offset_x = drawable_mod_undo->offset_x;
  offset_y = drawable_mod_undo->offset_y;

  drawable_mod_undo->buffer = g_object_ref (pika_drawable_get_buffer (drawable));

  pika_item_get_offset (PIKA_ITEM (drawable),
                        &drawable_mod_undo->offset_x,
                        &drawable_mod_undo->offset_y);

  pika_drawable_set_buffer_full (drawable, FALSE, NULL,
                                 buffer,
                                 GEGL_RECTANGLE (offset_x, offset_y, 0, 0),
                                 TRUE);
  g_object_unref (buffer);
}

static void
pika_drawable_mod_undo_free (PikaUndo     *undo,
                             PikaUndoMode  undo_mode)
{
  PikaDrawableModUndo *drawable_mod_undo = PIKA_DRAWABLE_MOD_UNDO (undo);

  g_clear_object (&drawable_mod_undo->buffer);

  PIKA_UNDO_CLASS (parent_class)->free (undo, undo_mode);
}
