/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "core-types.h"

#include "pika-memsize.h"
#include "pikaimage.h"
#include "pikagrouplayer.h"
#include "pikagrouplayerundo.h"


static void     pika_group_layer_undo_constructed (GObject             *object);

static gint64   pika_group_layer_undo_get_memsize (PikaObject          *object,
                                                   gint64              *gui_size);

static void     pika_group_layer_undo_pop         (PikaUndo            *undo,
                                                   PikaUndoMode         undo_mode,
                                                   PikaUndoAccumulator *accum);
static void     pika_group_layer_undo_free        (PikaUndo            *undo,
                                                   PikaUndoMode         undo_mode);


G_DEFINE_TYPE (PikaGroupLayerUndo, pika_group_layer_undo, PIKA_TYPE_ITEM_UNDO)

#define parent_class pika_group_layer_undo_parent_class


static void
pika_group_layer_undo_class_init (PikaGroupLayerUndoClass *klass)
{
  GObjectClass    *object_class      = G_OBJECT_CLASS (klass);
  PikaObjectClass *pika_object_class = PIKA_OBJECT_CLASS (klass);
  PikaUndoClass   *undo_class        = PIKA_UNDO_CLASS (klass);

  object_class->constructed      = pika_group_layer_undo_constructed;

  pika_object_class->get_memsize = pika_group_layer_undo_get_memsize;

  undo_class->pop                = pika_group_layer_undo_pop;
  undo_class->free               = pika_group_layer_undo_free;
}

static void
pika_group_layer_undo_init (PikaGroupLayerUndo *undo)
{
}

static void
pika_group_layer_undo_constructed (GObject *object)
{
  PikaGroupLayerUndo *group_layer_undo = PIKA_GROUP_LAYER_UNDO (object);
  PikaGroupLayer     *group;

  G_OBJECT_CLASS (parent_class)->constructed (object);

  g_return_if_fail (PIKA_IS_GROUP_LAYER (PIKA_ITEM_UNDO (object)->item));

  group = PIKA_GROUP_LAYER (PIKA_ITEM_UNDO (object)->item);

  switch (PIKA_UNDO (object)->undo_type)
    {
    case PIKA_UNDO_GROUP_LAYER_SUSPEND_RESIZE:
    case PIKA_UNDO_GROUP_LAYER_RESUME_RESIZE:
    case PIKA_UNDO_GROUP_LAYER_SUSPEND_MASK:
    case PIKA_UNDO_GROUP_LAYER_START_TRANSFORM:
    case PIKA_UNDO_GROUP_LAYER_END_TRANSFORM:
      break;

    case PIKA_UNDO_GROUP_LAYER_RESUME_MASK:
      group_layer_undo->mask_buffer =
        _pika_group_layer_get_suspended_mask(group,
                                             &group_layer_undo->mask_bounds);

      if (group_layer_undo->mask_buffer)
        g_object_ref (group_layer_undo->mask_buffer);
      break;

    case PIKA_UNDO_GROUP_LAYER_CONVERT:
      group_layer_undo->prev_type = pika_drawable_get_base_type (PIKA_DRAWABLE (group));
      group_layer_undo->prev_precision = pika_drawable_get_precision (PIKA_DRAWABLE (group));
      group_layer_undo->prev_has_alpha = pika_drawable_has_alpha (PIKA_DRAWABLE (group));
      break;

    default:
      g_return_if_reached ();
    }
}

static gint64
pika_group_layer_undo_get_memsize (PikaObject *object,
                                   gint64     *gui_size)
{
  PikaGroupLayerUndo *group_layer_undo = PIKA_GROUP_LAYER_UNDO (object);
  gint64              memsize       = 0;

  memsize += pika_gegl_buffer_get_memsize (group_layer_undo->mask_buffer);

  return memsize + PIKA_OBJECT_CLASS (parent_class)->get_memsize (object,
                                                                  gui_size);
}

static void
pika_group_layer_undo_pop (PikaUndo            *undo,
                           PikaUndoMode         undo_mode,
                           PikaUndoAccumulator *accum)
{
  PikaGroupLayerUndo *group_layer_undo = PIKA_GROUP_LAYER_UNDO (undo);
  PikaGroupLayer     *group;

  group = PIKA_GROUP_LAYER (PIKA_ITEM_UNDO (undo)->item);

  PIKA_UNDO_CLASS (parent_class)->pop (undo, undo_mode, accum);

  switch (undo->undo_type)
    {
    case PIKA_UNDO_GROUP_LAYER_SUSPEND_RESIZE:
    case PIKA_UNDO_GROUP_LAYER_RESUME_RESIZE:
      if ((undo_mode       == PIKA_UNDO_MODE_UNDO &&
           undo->undo_type == PIKA_UNDO_GROUP_LAYER_SUSPEND_RESIZE) ||
          (undo_mode       == PIKA_UNDO_MODE_REDO &&
           undo->undo_type == PIKA_UNDO_GROUP_LAYER_RESUME_RESIZE))
        {
          /*  resume group layer auto-resizing  */

          pika_group_layer_resume_resize (group, FALSE);
        }
      else
        {
          /*  suspend group layer auto-resizing  */

          pika_group_layer_suspend_resize (group, FALSE);

          if (undo->undo_type == PIKA_UNDO_GROUP_LAYER_RESUME_RESIZE &&
              group_layer_undo->mask_buffer)
            {
              PikaLayerMask *mask = pika_layer_get_mask (PIKA_LAYER (group));

              pika_drawable_set_buffer_full (PIKA_DRAWABLE (mask),
                                             FALSE, NULL,
                                             group_layer_undo->mask_buffer,
                                             &group_layer_undo->mask_bounds,
                                             TRUE);
            }
        }
      break;

    case PIKA_UNDO_GROUP_LAYER_SUSPEND_MASK:
    case PIKA_UNDO_GROUP_LAYER_RESUME_MASK:
      if ((undo_mode       == PIKA_UNDO_MODE_UNDO &&
           undo->undo_type == PIKA_UNDO_GROUP_LAYER_SUSPEND_MASK) ||
          (undo_mode       == PIKA_UNDO_MODE_REDO &&
           undo->undo_type == PIKA_UNDO_GROUP_LAYER_RESUME_MASK))
        {
          /*  resume group layer mask auto-resizing  */

          pika_group_layer_resume_mask (group, FALSE);
        }
      else
        {
          /*  suspend group layer mask auto-resizing  */

          pika_group_layer_suspend_mask (group, FALSE);

          if (undo->undo_type == PIKA_UNDO_GROUP_LAYER_RESUME_MASK &&
              group_layer_undo->mask_buffer)
            {
              _pika_group_layer_set_suspended_mask (
                group,
                group_layer_undo->mask_buffer,
                &group_layer_undo->mask_bounds);
            }
        }
      break;

    case PIKA_UNDO_GROUP_LAYER_START_TRANSFORM:
    case PIKA_UNDO_GROUP_LAYER_END_TRANSFORM:
      if ((undo_mode       == PIKA_UNDO_MODE_UNDO &&
           undo->undo_type == PIKA_UNDO_GROUP_LAYER_START_TRANSFORM) ||
          (undo_mode       == PIKA_UNDO_MODE_REDO &&
           undo->undo_type == PIKA_UNDO_GROUP_LAYER_END_TRANSFORM))
        {
          /*  end group layer transform operation  */

          _pika_group_layer_end_transform (group, FALSE);
        }
      else
        {
          /*  start group layer transform operation  */

          _pika_group_layer_start_transform (group, FALSE);
        }
      break;

    case PIKA_UNDO_GROUP_LAYER_CONVERT:
      {
        PikaImageBaseType type;
        PikaPrecision     precision;
        gboolean          has_alpha;

        type      = pika_drawable_get_base_type (PIKA_DRAWABLE (group));
        precision = pika_drawable_get_precision (PIKA_DRAWABLE (group));
        has_alpha = pika_drawable_has_alpha (PIKA_DRAWABLE (group));

        pika_drawable_convert_type (PIKA_DRAWABLE (group),
                                    pika_item_get_image (PIKA_ITEM (group)),
                                    group_layer_undo->prev_type,
                                    group_layer_undo->prev_precision,
                                    group_layer_undo->prev_has_alpha,
                                    NULL, NULL,
                                    0, 0,
                                    FALSE, NULL);

        group_layer_undo->prev_type      = type;
        group_layer_undo->prev_precision = precision;
        group_layer_undo->prev_has_alpha = has_alpha;
      }
      break;

    default:
      g_return_if_reached ();
    }
}

static void
pika_group_layer_undo_free (PikaUndo     *undo,
                            PikaUndoMode  undo_mode)
{
  PikaGroupLayerUndo *group_layer_undo = PIKA_GROUP_LAYER_UNDO (undo);

  g_clear_object (&group_layer_undo->mask_buffer);

  PIKA_UNDO_CLASS (parent_class)->free (undo, undo_mode);
}
