/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"

#include "core-types.h"

#include "pikachannel.h"
#include "pikacontext.h"
#include "pikaimage.h"
#include "pikaimage-item-list.h"
#include "pikaimage-undo.h"
#include "pikaitem.h"
#include "pikalayer.h"
#include "pikaobjectqueue.h"
#include "pikaprogress.h"

#include "vectors/pikavectors.h"

#include "pika-intl.h"


/*  public functions  */

gboolean
pika_image_item_list_bounds (PikaImage *image,
                             GList     *list,
                             gint      *x,
                             gint      *y,
                             gint      *width,
                             gint      *height)
{
  GList    *l;
  gboolean  bounds = FALSE;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), FALSE);
  g_return_val_if_fail (x != 0, FALSE);
  g_return_val_if_fail (y != 0, FALSE);
  g_return_val_if_fail (width != 0, FALSE);
  g_return_val_if_fail (height != 0, FALSE);

  for (l = list; l; l = g_list_next (l))
    {
      PikaItem *item = l->data;
      gint      tmp_x, tmp_y;
      gint      tmp_w, tmp_h;

      if (pika_item_bounds (item, &tmp_x, &tmp_y, &tmp_w, &tmp_h))
        {
          gint off_x, off_y;

          pika_item_get_offset (item, &off_x, &off_y);

          if (bounds)
            {
              pika_rectangle_union (*x, *y, *width, *height,
                                    tmp_x + off_x, tmp_y + off_y,
                                    tmp_w, tmp_h,
                                    x, y, width, height);
            }
          else
            {
              *x      = tmp_x + off_x;
              *y      = tmp_y + off_y;
              *width  = tmp_w;
              *height = tmp_h;

              bounds = TRUE;
            }
        }
    }

  if (! bounds)
    {
      *x      = 0;
      *y      = 0;
      *width  = pika_image_get_width  (image);
      *height = pika_image_get_height (image);
    }

  return bounds;
}

void
pika_image_item_list_translate (PikaImage *image,
                                GList     *list,
                                gint       offset_x,
                                gint       offset_y,
                                gboolean   push_undo)
{
  g_return_if_fail (PIKA_IS_IMAGE (image));

  if (list)
    {
      GList *l;

      if (list->next)
        {
          if (push_undo)
            {
              pika_image_undo_group_start (image, PIKA_UNDO_GROUP_ITEM_DISPLACE,
                                           C_("undo-type", "Translate Items"));
            }

          for (l = list; l; l = g_list_next (l))
            pika_item_start_transform (PIKA_ITEM (l->data), push_undo);
        }

      for (l = list; l; l = g_list_next (l))
        pika_item_translate (PIKA_ITEM (l->data),
                             offset_x, offset_y, push_undo);

      if (list->next)
        {
          for (l = list; l; l = g_list_next (l))
            pika_item_end_transform (PIKA_ITEM (l->data), push_undo);

          if (push_undo)
            pika_image_undo_group_end (image);
        }
    }
}

void
pika_image_item_list_flip (PikaImage           *image,
                           GList               *list,
                           PikaContext         *context,
                           PikaOrientationType  flip_type,
                           gdouble              axis,
                           PikaTransformResize  expected_clip_result)
{
  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_CONTEXT (context));

  if (list)
    {
      GList *l;

      if (list->next)
        {
          pika_image_undo_group_start (image, PIKA_UNDO_GROUP_TRANSFORM,
                                       C_("undo-type", "Flip Items"));

          for (l = list; l; l = g_list_next (l))
            pika_item_start_transform (PIKA_ITEM (l->data), TRUE);
        }

      for (l = list; l; l = g_list_next (l))
        {
          PikaItem *item = l->data;

          pika_item_flip (item, context,
                          flip_type, axis,
                          pika_item_get_clip (item, expected_clip_result) !=
                          PIKA_TRANSFORM_RESIZE_ADJUST);
        }

      if (list->next)
        {
          for (l = list; l; l = g_list_next (l))
            pika_item_end_transform (PIKA_ITEM (l->data), TRUE);

          pika_image_undo_group_end (image);
        }
    }
}

void
pika_image_item_list_rotate (PikaImage        *image,
                             GList            *list,
                             PikaContext      *context,
                             PikaRotationType  rotate_type,
                             gdouble           center_x,
                             gdouble           center_y,
                             gboolean          clip_result)
{
  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_CONTEXT (context));

  if (list)
    {
      GList *l;

      if (list->next)
        {
          pika_image_undo_group_start (image, PIKA_UNDO_GROUP_TRANSFORM,
                                       C_("undo-type", "Rotate Items"));

          for (l = list; l; l = g_list_next (l))
            pika_item_start_transform (PIKA_ITEM (l->data), TRUE);
        }

      for (l = list; l; l = g_list_next (l))
        {
          PikaItem *item = l->data;

          pika_item_rotate (item, context,
                            rotate_type, center_x, center_y,
                            pika_item_get_clip (item, clip_result));
        }

      if (list->next)
        {
          for (l = list; l; l = g_list_next (l))
            pika_item_end_transform (PIKA_ITEM (l->data), TRUE);

          pika_image_undo_group_end (image);
        }
    }
}

void
pika_image_item_list_transform (PikaImage              *image,
                                GList                  *list,
                                PikaContext            *context,
                                const PikaMatrix3      *matrix,
                                PikaTransformDirection  direction,
                                PikaInterpolationType   interpolation_type,
                                PikaTransformResize     clip_result,
                                PikaProgress           *progress)
{
  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_CONTEXT (context));
  g_return_if_fail (progress == NULL || PIKA_IS_PROGRESS (progress));

  if (list)
    {
      PikaObjectQueue *queue = NULL;
      GList           *l;

      if (progress)
        {
          queue    = pika_object_queue_new (progress);
          progress = PIKA_PROGRESS (queue);

          pika_object_queue_push_list (queue, list);
        }

      if (list->next)
        {
          pika_image_undo_group_start (image, PIKA_UNDO_GROUP_TRANSFORM,
                                       C_("undo-type", "Transform Items"));

          for (l = list; l; l = g_list_next (l))
            pika_item_start_transform (PIKA_ITEM (l->data), TRUE);
        }

      for (l = list; l; l = g_list_next (l))
        {
          PikaItem *item = l->data;

          if (queue)
            pika_object_queue_pop (queue);

          pika_item_transform (item, context,
                               matrix, direction,
                               interpolation_type,
                               pika_item_get_clip (item, clip_result),
                               progress);
        }

      if (list->next)
        {
          for (l = list; l; l = g_list_next (l))
            pika_item_end_transform (PIKA_ITEM (l->data), TRUE);

          pika_image_undo_group_end (image);
        }

      g_clear_object (&queue);
    }
}

/**
 * pika_image_item_list_get_list:
 * @image:   An @image.
 * @type:    Which type of items to return.
 * @set:     Set the returned items are part of.
 *
 * This function returns a #GList of #PikaItem<!-- -->s for which the
 * @type and @set criterions match.
 *
 * Returns: The list of items.
 **/
GList *
pika_image_item_list_get_list (PikaImage        *image,
                               PikaItemTypeMask  type,
                               PikaItemSet       set)
{
  GList *all_items;
  GList *list;
  GList *return_list = NULL;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);

  if (type & PIKA_ITEM_TYPE_LAYERS)
    {
      all_items = pika_image_get_layer_list (image);

      for (list = all_items; list; list = g_list_next (list))
        {
          PikaItem *item = list->data;

          if (pika_item_is_in_set (item, set))
            return_list = g_list_prepend (return_list, item);
        }

      g_list_free (all_items);
    }

  if (type & PIKA_ITEM_TYPE_CHANNELS)
    {
      all_items = pika_image_get_channel_list (image);

      for (list = all_items; list; list = g_list_next (list))
        {
          PikaItem *item = list->data;

          if (pika_item_is_in_set (item, set))
            return_list = g_list_prepend (return_list, item);
        }

      g_list_free (all_items);
    }

  if (type & PIKA_ITEM_TYPE_VECTORS)
    {
      all_items = pika_image_get_vectors_list (image);

      for (list = all_items; list; list = g_list_next (list))
        {
          PikaItem *item = list->data;

          if (pika_item_is_in_set (item, set))
            return_list = g_list_prepend (return_list, item);
        }

      g_list_free (all_items);
    }

  return g_list_reverse (return_list);
}

static GList *
pika_image_item_list_remove_children (GList          *list,
                                      const PikaItem *parent)
{
  GList *l = list;

  while (l)
    {
      PikaItem *item = l->data;

      l = g_list_next (l);

      if (pika_viewable_is_ancestor (PIKA_VIEWABLE (parent),
                                     PIKA_VIEWABLE (item)))
        {
          list = g_list_remove (list, item);
        }
    }

  return list;
}

/**
 * pika_image_item_list_filter:
 * @image:
 * @items: the original list of #PikaItem-s.
 *
 * Filter @list by modifying it directly (so the original list should
 * not be used anymore, only its result), removing all children items
 * with ancestors also in @list.
 *
 * Returns: the modified @list where all items which have an ancestor in
 *          @list have been removed.
 */
GList *
pika_image_item_list_filter (GList *list)
{
  GList *l;

  if (! list)
    return NULL;

  for (l = list; l; l = g_list_next (l))
    {
      PikaItem *item = l->data;
      GList    *next;

      next = pika_image_item_list_remove_children (g_list_next (l), item);

      l->next = next;
      if (next)
        next->prev = l;
    }

  return list;
}
