/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <cairo.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"
#include "libpikacolor/pikacolor.h"

#include "core-types.h"

#include "pikaimage.h"
#include "pikaimage-color-profile.h"
#include "pikaimage-metadata.h"
#include "pikaimage-private.h"
#include "pikaimage-undo-push.h"


/* public functions */


PikaMetadata *
pika_image_get_metadata (PikaImage *image)
{
  PikaImagePrivate *private;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);

  private = PIKA_IMAGE_GET_PRIVATE (image);

  return private->metadata;
}

void
pika_image_set_metadata (PikaImage    *image,
                         PikaMetadata *metadata,
                         gboolean      push_undo)
{
  PikaImagePrivate *private;

  g_return_if_fail (PIKA_IS_IMAGE (image));

  private = PIKA_IMAGE_GET_PRIVATE (image);

  if (metadata != private->metadata)
    {
      if (push_undo)
        pika_image_undo_push_image_metadata (image, NULL);

      g_set_object (&private->metadata, metadata);

      if (private->metadata)
        {
          pika_image_metadata_update_pixel_size      (image);
          pika_image_metadata_update_bits_per_sample (image);
          pika_image_metadata_update_resolution      (image);
          pika_image_metadata_update_colorspace      (image);
        }

      g_object_notify (G_OBJECT (image), "metadata");
    }
}

void
pika_image_metadata_update_pixel_size (PikaImage *image)
{
  PikaMetadata *metadata;

  g_return_if_fail (PIKA_IS_IMAGE (image));

  metadata = pika_image_get_metadata (image);

  if (metadata)
    {
      pika_metadata_set_pixel_size (metadata,
                                    pika_image_get_width  (image),
                                    pika_image_get_height (image));
    }
}

void
pika_image_metadata_update_bits_per_sample (PikaImage *image)
{
  PikaMetadata *metadata;

  g_return_if_fail (PIKA_IS_IMAGE (image));

  metadata = pika_image_get_metadata (image);

  if (metadata)
    {
      switch (pika_image_get_component_type (image))
        {
        case PIKA_COMPONENT_TYPE_U8:
          pika_metadata_set_bits_per_sample (metadata, 8);
          break;

        case PIKA_COMPONENT_TYPE_U16:
        case PIKA_COMPONENT_TYPE_HALF:
          pika_metadata_set_bits_per_sample (metadata, 16);
          break;

        case PIKA_COMPONENT_TYPE_U32:
        case PIKA_COMPONENT_TYPE_FLOAT:
          pika_metadata_set_bits_per_sample (metadata, 32);
          break;

        case PIKA_COMPONENT_TYPE_DOUBLE:
          pika_metadata_set_bits_per_sample (metadata, 64);
          break;
        }
    }
}

void
pika_image_metadata_update_resolution (PikaImage *image)
{
  PikaMetadata *metadata;

  g_return_if_fail (PIKA_IS_IMAGE (image));

  metadata = pika_image_get_metadata (image);

  if (metadata)
    {
      gdouble xres, yres;

      pika_image_get_resolution (image, &xres, &yres);
      pika_metadata_set_resolution (metadata, xres, yres,
                                    pika_image_get_unit (image));
    }
}

void
pika_image_metadata_update_colorspace (PikaImage *image)
{
  PikaMetadata *metadata;

  g_return_if_fail (PIKA_IS_IMAGE (image));

  metadata = pika_image_get_metadata (image);

  if (metadata)
    {
      /*  See the discussions in issue #3532 and issue #301  */

      PikaColorProfile       *profile = pika_image_get_color_profile (image);
      PikaMetadataColorspace  space   = PIKA_METADATA_COLORSPACE_UNSPECIFIED;

      if (profile)
        {
          static PikaColorProfile *adobe = NULL;

          if (! adobe)
            adobe = pika_color_profile_new_rgb_adobe ();

          if (pika_color_profile_is_equal (profile, adobe))
            space = PIKA_METADATA_COLORSPACE_ADOBERGB;
        }
      else
        {
          space = PIKA_METADATA_COLORSPACE_SRGB;
        }

      pika_metadata_set_colorspace (metadata, space);
    }
}
