/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "core-types.h"

#include "pika.h"
#include "pikaimage.h"
#include "pikaimage-private.h"
#include "pikaimage-sample-points.h"
#include "pikaimage-undo-push.h"
#include "pikasamplepoint.h"

#include "pika-intl.h"


/*  public functions  */

PikaSamplePoint *
pika_image_add_sample_point_at_pos (PikaImage *image,
                                    gint       x,
                                    gint       y,
                                    gboolean   push_undo)
{
  PikaSamplePoint *sample_point;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);
  g_return_val_if_fail (x >= 0 && x < pika_image_get_width  (image), NULL);
  g_return_val_if_fail (y >= 0 && y < pika_image_get_height (image), NULL);

  sample_point = pika_sample_point_new (image->pika->next_sample_point_id++);

  if (push_undo)
    pika_image_undo_push_sample_point (image, C_("undo-type", "Add Sample Point"),
                                       sample_point);

  pika_image_add_sample_point (image, sample_point, x, y);
  g_object_unref (sample_point);

  return sample_point;
}

void
pika_image_add_sample_point (PikaImage       *image,
                             PikaSamplePoint *sample_point,
                             gint             x,
                             gint             y)
{
  PikaImagePrivate *private;

  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_SAMPLE_POINT (sample_point));

  private = PIKA_IMAGE_GET_PRIVATE (image);

  private->sample_points = g_list_append (private->sample_points, sample_point);

  pika_sample_point_set_position (sample_point, x, y);
  g_object_ref (sample_point);

  pika_image_sample_point_added (image, sample_point);
}

void
pika_image_remove_sample_point (PikaImage       *image,
                                PikaSamplePoint *sample_point,
                                gboolean         push_undo)
{
  PikaImagePrivate *private;

  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_SAMPLE_POINT (sample_point));

  private = PIKA_IMAGE_GET_PRIVATE (image);

  if (push_undo)
    pika_image_undo_push_sample_point (image,
                                       C_("undo-type", "Remove Sample Point"),
                                       sample_point);

  private->sample_points = g_list_remove (private->sample_points, sample_point);
  pika_aux_item_removed (PIKA_AUX_ITEM (sample_point));

  pika_image_sample_point_removed (image, sample_point);

  pika_sample_point_set_position (sample_point,
                                  PIKA_SAMPLE_POINT_POSITION_UNDEFINED,
                                  PIKA_SAMPLE_POINT_POSITION_UNDEFINED);
  g_object_unref (sample_point);
}

void
pika_image_move_sample_point (PikaImage       *image,
                              PikaSamplePoint *sample_point,
                              gint             x,
                              gint             y,
                              gboolean         push_undo)
{
  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_SAMPLE_POINT (sample_point));
  g_return_if_fail (x >= 0);
  g_return_if_fail (y >= 0);
  g_return_if_fail (x < pika_image_get_width  (image));
  g_return_if_fail (y < pika_image_get_height (image));

  if (push_undo)
    pika_image_undo_push_sample_point (image,
                                       C_("undo-type", "Move Sample Point"),
                                       sample_point);

  pika_sample_point_set_position (sample_point, x, y);

  pika_image_sample_point_moved (image, sample_point);
}

void
pika_image_set_sample_point_pick_mode (PikaImage         *image,
                                       PikaSamplePoint   *sample_point,
                                       PikaColorPickMode  pick_mode,
                                       gboolean           push_undo)
{
  g_return_if_fail (PIKA_IS_IMAGE (image));
  g_return_if_fail (PIKA_IS_SAMPLE_POINT (sample_point));

  if (push_undo)
    pika_image_undo_push_sample_point (image,
                                       C_("undo-type",
                                          "Set Sample Point Pick Mode"),
                                       sample_point);

  pika_sample_point_set_pick_mode (sample_point, pick_mode);

  /* well... */
  pika_image_sample_point_moved (image, sample_point);
}

GList *
pika_image_get_sample_points (PikaImage *image)
{
  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);

  return PIKA_IMAGE_GET_PRIVATE (image)->sample_points;
}

PikaSamplePoint *
pika_image_get_sample_point (PikaImage *image,
                             guint32    id)
{
  GList *sample_points;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);

  for (sample_points = PIKA_IMAGE_GET_PRIVATE (image)->sample_points;
       sample_points;
       sample_points = g_list_next (sample_points))
    {
      PikaSamplePoint *sample_point = sample_points->data;

      if (pika_aux_item_get_id (PIKA_AUX_ITEM (sample_point)) == id)
        return sample_point;
    }

  return NULL;
}

PikaSamplePoint *
pika_image_get_next_sample_point (PikaImage *image,
                                  guint32    id,
                                  gboolean  *sample_point_found)
{
  GList *sample_points;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);
  g_return_val_if_fail (sample_point_found != NULL, NULL);

  if (id == 0)
    *sample_point_found = TRUE;
  else
    *sample_point_found = FALSE;

  for (sample_points = PIKA_IMAGE_GET_PRIVATE (image)->sample_points;
       sample_points;
       sample_points = g_list_next (sample_points))
    {
      PikaSamplePoint *sample_point = sample_points->data;

      if (*sample_point_found) /* this is the first guide after the found one */
        return sample_point;

      if (pika_aux_item_get_id (PIKA_AUX_ITEM (sample_point)) == id) /* found it, next one will be returned */
        *sample_point_found = TRUE;
    }

  return NULL;
}
