/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"

#include "core-types.h"

#include "pikaimage.h"
#include "pikalayer.h"
#include "pikalayermask.h"
#include "pikalayermaskundo.h"


enum
{
  PROP_0,
  PROP_LAYER_MASK
};


static void     pika_layer_mask_undo_constructed  (GObject             *object);
static void     pika_layer_mask_undo_set_property (GObject             *object,
                                                   guint                property_id,
                                                   const GValue        *value,
                                                   GParamSpec          *pspec);
static void     pika_layer_mask_undo_get_property (GObject             *object,
                                                   guint                property_id,
                                                   GValue              *value,
                                                   GParamSpec          *pspec);

static gint64   pika_layer_mask_undo_get_memsize  (PikaObject          *object,
                                                   gint64              *gui_size);

static void     pika_layer_mask_undo_pop          (PikaUndo            *undo,
                                                   PikaUndoMode         undo_mode,
                                                   PikaUndoAccumulator *accum);
static void     pika_layer_mask_undo_free         (PikaUndo            *undo,
                                                   PikaUndoMode         undo_mode);


G_DEFINE_TYPE (PikaLayerMaskUndo, pika_layer_mask_undo, PIKA_TYPE_ITEM_UNDO)

#define parent_class pika_layer_mask_undo_parent_class


static void
pika_layer_mask_undo_class_init (PikaLayerMaskUndoClass *klass)
{
  GObjectClass    *object_class      = G_OBJECT_CLASS (klass);
  PikaObjectClass *pika_object_class = PIKA_OBJECT_CLASS (klass);
  PikaUndoClass   *undo_class        = PIKA_UNDO_CLASS (klass);

  object_class->constructed      = pika_layer_mask_undo_constructed;
  object_class->set_property     = pika_layer_mask_undo_set_property;
  object_class->get_property     = pika_layer_mask_undo_get_property;

  pika_object_class->get_memsize = pika_layer_mask_undo_get_memsize;

  undo_class->pop                = pika_layer_mask_undo_pop;
  undo_class->free               = pika_layer_mask_undo_free;

  g_object_class_install_property (object_class, PROP_LAYER_MASK,
                                   g_param_spec_object ("layer-mask", NULL, NULL,
                                                        PIKA_TYPE_LAYER_MASK,
                                                        PIKA_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT_ONLY));
}

static void
pika_layer_mask_undo_init (PikaLayerMaskUndo *undo)
{
}

static void
pika_layer_mask_undo_constructed (GObject *object)
{
  PikaLayerMaskUndo *layer_mask_undo = PIKA_LAYER_MASK_UNDO (object);

  G_OBJECT_CLASS (parent_class)->constructed (object);

  pika_assert (PIKA_IS_LAYER (PIKA_ITEM_UNDO (object)->item));
  pika_assert (PIKA_IS_LAYER_MASK (layer_mask_undo->layer_mask));
}

static void
pika_layer_mask_undo_set_property (GObject      *object,
                                   guint         property_id,
                                   const GValue *value,
                                   GParamSpec   *pspec)
{
  PikaLayerMaskUndo *layer_mask_undo = PIKA_LAYER_MASK_UNDO (object);

  switch (property_id)
    {
    case PROP_LAYER_MASK:
      layer_mask_undo->layer_mask = g_value_dup_object (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_layer_mask_undo_get_property (GObject    *object,
                                   guint       property_id,
                                   GValue     *value,
                                   GParamSpec *pspec)
{
  PikaLayerMaskUndo *layer_mask_undo = PIKA_LAYER_MASK_UNDO (object);

  switch (property_id)
    {
    case PROP_LAYER_MASK:
      g_value_set_object (value, layer_mask_undo->layer_mask);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static gint64
pika_layer_mask_undo_get_memsize (PikaObject *object,
                                  gint64     *gui_size)
{
  PikaLayerMaskUndo *layer_mask_undo = PIKA_LAYER_MASK_UNDO (object);
  PikaLayer         *layer           = PIKA_LAYER (PIKA_ITEM_UNDO (object)->item);
  gint64             memsize         = 0;

  /* don't use !pika_item_is_attached() here */
  if (pika_layer_get_mask (layer) != layer_mask_undo->layer_mask)
    memsize += pika_object_get_memsize (PIKA_OBJECT (layer_mask_undo->layer_mask),
                                        gui_size);

  return memsize + PIKA_OBJECT_CLASS (parent_class)->get_memsize (object,
                                                                  gui_size);
}

static void
pika_layer_mask_undo_pop (PikaUndo            *undo,
                          PikaUndoMode         undo_mode,
                          PikaUndoAccumulator *accum)
{
  PikaLayerMaskUndo *layer_mask_undo = PIKA_LAYER_MASK_UNDO (undo);
  PikaLayer         *layer           = PIKA_LAYER (PIKA_ITEM_UNDO (undo)->item);

  PIKA_UNDO_CLASS (parent_class)->pop (undo, undo_mode, accum);

  if ((undo_mode       == PIKA_UNDO_MODE_UNDO &&
       undo->undo_type == PIKA_UNDO_LAYER_MASK_ADD) ||
      (undo_mode       == PIKA_UNDO_MODE_REDO &&
       undo->undo_type == PIKA_UNDO_LAYER_MASK_REMOVE))
    {
      /*  remove layer mask  */

      pika_layer_apply_mask (layer, PIKA_MASK_DISCARD, FALSE);
    }
  else
    {
      /*  restore layer mask  */

      pika_layer_add_mask (layer, layer_mask_undo->layer_mask, FALSE, NULL);
    }
}

static void
pika_layer_mask_undo_free (PikaUndo     *undo,
                           PikaUndoMode  undo_mode)
{
  PikaLayerMaskUndo *layer_mask_undo = PIKA_LAYER_MASK_UNDO (undo);

  g_clear_object (&layer_mask_undo->layer_mask);

  PIKA_UNDO_CLASS (parent_class)->free (undo, undo_mode);
}
