/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"

#include "core-types.h"

#include "pikaimage.h"
#include "pikalayer.h"
#include "pikalayerundo.h"


enum
{
  PROP_0,
  PROP_PREV_PARENT,
  PROP_PREV_POSITION,
  PROP_PREV_LAYERS
};


static void     pika_layer_undo_constructed  (GObject             *object);
static void     pika_layer_undo_finalize     (GObject             *object);
static void     pika_layer_undo_set_property (GObject             *object,
                                              guint                property_id,
                                              const GValue        *value,
                                              GParamSpec          *pspec);
static void     pika_layer_undo_get_property (GObject             *object,
                                              guint                property_id,
                                              GValue              *value,
                                              GParamSpec          *pspec);

static gint64   pika_layer_undo_get_memsize  (PikaObject          *object,
                                              gint64              *gui_size);

static void     pika_layer_undo_pop          (PikaUndo            *undo,
                                              PikaUndoMode         undo_mode,
                                              PikaUndoAccumulator *accum);


G_DEFINE_TYPE (PikaLayerUndo, pika_layer_undo, PIKA_TYPE_ITEM_UNDO)

#define parent_class pika_layer_undo_parent_class


static void
pika_layer_undo_class_init (PikaLayerUndoClass *klass)
{
  GObjectClass    *object_class      = G_OBJECT_CLASS (klass);
  PikaObjectClass *pika_object_class = PIKA_OBJECT_CLASS (klass);
  PikaUndoClass   *undo_class        = PIKA_UNDO_CLASS (klass);

  object_class->constructed      = pika_layer_undo_constructed;
  object_class->finalize         = pika_layer_undo_finalize;
  object_class->set_property     = pika_layer_undo_set_property;
  object_class->get_property     = pika_layer_undo_get_property;

  pika_object_class->get_memsize = pika_layer_undo_get_memsize;

  undo_class->pop                = pika_layer_undo_pop;

  g_object_class_install_property (object_class, PROP_PREV_PARENT,
                                   g_param_spec_object ("prev-parent",
                                                        NULL, NULL,
                                                        PIKA_TYPE_LAYER,
                                                        PIKA_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT_ONLY));

  g_object_class_install_property (object_class, PROP_PREV_POSITION,
                                   g_param_spec_int ("prev-position", NULL, NULL,
                                                     0, G_MAXINT, 0,
                                                     PIKA_PARAM_READWRITE |
                                                     G_PARAM_CONSTRUCT_ONLY));

  g_object_class_install_property (object_class, PROP_PREV_LAYERS,
                                   g_param_spec_pointer ("prev-layers", NULL, NULL,
                                                         PIKA_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY));
}

static void
pika_layer_undo_init (PikaLayerUndo *undo)
{
  undo->prev_layers = NULL;
}

static void
pika_layer_undo_constructed (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->constructed (object);

  pika_assert (PIKA_IS_LAYER (PIKA_ITEM_UNDO (object)->item));
}

static void
pika_layer_undo_finalize (GObject *object)
{
  PikaLayerUndo *undo = PIKA_LAYER_UNDO (object);

  g_clear_pointer (&undo->prev_layers, g_list_free);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}
static void
pika_layer_undo_set_property (GObject      *object,
                              guint         property_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
  PikaLayerUndo *layer_undo = PIKA_LAYER_UNDO (object);

  switch (property_id)
    {
    case PROP_PREV_PARENT:
      layer_undo->prev_parent = g_value_get_object (value);
      break;
    case PROP_PREV_POSITION:
      layer_undo->prev_position = g_value_get_int (value);
      break;
    case PROP_PREV_LAYERS:
      layer_undo->prev_layers = g_list_copy (g_value_get_pointer (value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_layer_undo_get_property (GObject    *object,
                              guint       property_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
  PikaLayerUndo *layer_undo = PIKA_LAYER_UNDO (object);

  switch (property_id)
    {
    case PROP_PREV_PARENT:
      g_value_set_object (value, layer_undo->prev_parent);
      break;
    case PROP_PREV_POSITION:
      g_value_set_int (value, layer_undo->prev_position);
      break;
    case PROP_PREV_LAYERS:
      g_value_set_pointer (value, layer_undo->prev_layers);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static gint64
pika_layer_undo_get_memsize (PikaObject *object,
                             gint64     *gui_size)
{
  PikaItemUndo *item_undo = PIKA_ITEM_UNDO (object);
  gint64        memsize   = 0;

  if (! pika_item_is_attached (item_undo->item))
    memsize += pika_object_get_memsize (PIKA_OBJECT (item_undo->item),
                                        gui_size);

  return memsize + PIKA_OBJECT_CLASS (parent_class)->get_memsize (object,
                                                                  gui_size);
}

static void
pika_layer_undo_pop (PikaUndo            *undo,
                     PikaUndoMode         undo_mode,
                     PikaUndoAccumulator *accum)
{
  PikaLayerUndo *layer_undo = PIKA_LAYER_UNDO (undo);
  PikaLayer     *layer      = PIKA_LAYER (PIKA_ITEM_UNDO (undo)->item);

  PIKA_UNDO_CLASS (parent_class)->pop (undo, undo_mode, accum);

  if ((undo_mode       == PIKA_UNDO_MODE_UNDO &&
       undo->undo_type == PIKA_UNDO_LAYER_ADD) ||
      (undo_mode       == PIKA_UNDO_MODE_REDO &&
       undo->undo_type == PIKA_UNDO_LAYER_REMOVE))
    {
      /*  remove layer  */

      /*  record the current parent and position  */
      layer_undo->prev_parent   = pika_layer_get_parent (layer);
      layer_undo->prev_position = pika_item_get_index (PIKA_ITEM (layer));

      pika_image_remove_layer (undo->image, layer, FALSE,
                               layer_undo->prev_layers);
    }
  else
    {
      /*  restore layer  */

      /*  record the active layer  */
      g_clear_pointer (&layer_undo->prev_layers, g_list_free);
      layer_undo->prev_layers = g_list_copy (pika_image_get_selected_layers (undo->image));

      pika_image_add_layer (undo->image, layer,
                            layer_undo->prev_parent,
                            layer_undo->prev_position, FALSE);
    }
}
