/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "core-types.h"

#include "paint/pikapaintoptions.h"

#include "pika.h"
#include "pikapaintinfo.h"


static void    pika_paint_info_dispose         (GObject       *object);
static void    pika_paint_info_finalize        (GObject       *object);
static gchar * pika_paint_info_get_description (PikaViewable  *viewable,
                                                gchar        **tooltip);


G_DEFINE_TYPE (PikaPaintInfo, pika_paint_info, PIKA_TYPE_VIEWABLE)

#define parent_class pika_paint_info_parent_class


static void
pika_paint_info_class_init (PikaPaintInfoClass *klass)
{
  GObjectClass      *object_class   = G_OBJECT_CLASS (klass);
  PikaViewableClass *viewable_class = PIKA_VIEWABLE_CLASS (klass);

  object_class->dispose           = pika_paint_info_dispose;
  object_class->finalize          = pika_paint_info_finalize;

  viewable_class->get_description = pika_paint_info_get_description;
}

static void
pika_paint_info_init (PikaPaintInfo *paint_info)
{
  paint_info->pika          = NULL;
  paint_info->paint_type    = G_TYPE_NONE;
  paint_info->blurb         = NULL;
  paint_info->paint_options = NULL;
}

static void
pika_paint_info_dispose (GObject *object)
{
  PikaPaintInfo *paint_info = PIKA_PAINT_INFO (object);

  if (paint_info->paint_options)
    {
      g_object_run_dispose (G_OBJECT (paint_info->paint_options));
      g_clear_object (&paint_info->paint_options);
    }

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
pika_paint_info_finalize (GObject *object)
{
  PikaPaintInfo *paint_info = PIKA_PAINT_INFO (object);

  g_clear_pointer (&paint_info->blurb, g_free);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gchar *
pika_paint_info_get_description (PikaViewable  *viewable,
                                 gchar        **tooltip)
{
  PikaPaintInfo *paint_info = PIKA_PAINT_INFO (viewable);

  return g_strdup (paint_info->blurb);
}

PikaPaintInfo *
pika_paint_info_new (Pika        *pika,
                     GType        paint_type,
                     GType        paint_options_type,
                     const gchar *identifier,
                     const gchar *blurb,
                     const gchar *icon_name)
{
  PikaPaintInfo *paint_info;

  g_return_val_if_fail (PIKA_IS_PIKA (pika), NULL);
  g_return_val_if_fail (identifier != NULL, NULL);
  g_return_val_if_fail (blurb != NULL, NULL);
  g_return_val_if_fail (icon_name != NULL, NULL);

  paint_info = g_object_new (PIKA_TYPE_PAINT_INFO,
                             "name",      identifier,
                             "icon-name", icon_name,
                             NULL);

  paint_info->pika               = pika;
  paint_info->paint_type         = paint_type;
  paint_info->paint_options_type = paint_options_type;
  paint_info->blurb              = g_strdup (blurb);

  paint_info->paint_options      = pika_paint_options_new (paint_info);

  return paint_info;
}

void
pika_paint_info_set_standard (Pika          *pika,
                              PikaPaintInfo *paint_info)
{
  g_return_if_fail (PIKA_IS_PIKA (pika));
  g_return_if_fail (! paint_info || PIKA_IS_PAINT_INFO (paint_info));

  g_set_object (&pika->standard_paint_info, paint_info);
}

PikaPaintInfo *
pika_paint_info_get_standard (Pika *pika)
{
  g_return_val_if_fail (PIKA_IS_PIKA (pika), NULL);

  return pika->standard_paint_info;
}
