/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <string.h>

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"

#include "core-types.h"

#include "pika.h"
#include "pikadatafactory.h"
#include "pikafilteredcontainer.h"
#include "pikapaintinfo.h"
#include "pikatoolinfo.h"
#include "pikatooloptions.h"
#include "pikatoolpreset.h"


static void    pika_tool_info_dispose         (GObject       *object);
static void    pika_tool_info_finalize        (GObject       *object);

static gchar * pika_tool_info_get_description (PikaViewable  *viewable,
                                               gchar        **tooltip);


G_DEFINE_TYPE (PikaToolInfo, pika_tool_info, PIKA_TYPE_TOOL_ITEM)

#define parent_class pika_tool_info_parent_class


static void
pika_tool_info_class_init (PikaToolInfoClass *klass)
{
  GObjectClass      *object_class   = G_OBJECT_CLASS (klass);
  PikaViewableClass *viewable_class = PIKA_VIEWABLE_CLASS (klass);

  object_class->dispose           = pika_tool_info_dispose;
  object_class->finalize          = pika_tool_info_finalize;

  viewable_class->get_description = pika_tool_info_get_description;
}

static void
pika_tool_info_init (PikaToolInfo *tool_info)
{
}

static void
pika_tool_info_dispose (GObject *object)
{
  PikaToolInfo *tool_info = PIKA_TOOL_INFO (object);

  if (tool_info->tool_options)
    {
      g_object_run_dispose (G_OBJECT (tool_info->tool_options));
      g_clear_object (&tool_info->tool_options);
    }

  g_clear_object (&tool_info->presets);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
pika_tool_info_finalize (GObject *object)
{
  PikaToolInfo *tool_info = PIKA_TOOL_INFO (object);

  g_clear_pointer (&tool_info->label,       g_free);
  g_clear_pointer (&tool_info->tooltip,     g_free);
  g_clear_pointer (&tool_info->menu_label,  g_free);
  g_clear_pointer (&tool_info->menu_accel,  g_free);
  g_clear_pointer (&tool_info->help_domain, g_free);
  g_clear_pointer (&tool_info->help_id,     g_free);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gchar *
pika_tool_info_get_description (PikaViewable  *viewable,
                                gchar        **tooltip)
{
  PikaToolInfo *tool_info = PIKA_TOOL_INFO (viewable);

  if (tooltip)
    *tooltip = g_strdup (tool_info->tooltip);

  return g_strdup (tool_info->label);
}

static gboolean
pika_tool_info_filter_preset (PikaObject *object,
                              gpointer    user_data)
{
  PikaToolPreset *preset    = PIKA_TOOL_PRESET (object);
  PikaToolInfo   *tool_info = user_data;

  return preset->tool_options->tool_info == tool_info;
}

PikaToolInfo *
pika_tool_info_new (Pika                *pika,
                    GType                tool_type,
                    GType                tool_options_type,
                    PikaContextPropMask  context_props,
                    const gchar         *identifier,
                    const gchar         *label,
                    const gchar         *tooltip,
                    const gchar         *menu_label,
                    const gchar         *menu_accel,
                    const gchar         *help_domain,
                    const gchar         *help_id,
                    const gchar         *paint_core_name,
                    const gchar         *icon_name)
{
  PikaPaintInfo *paint_info;
  PikaToolInfo  *tool_info;

  g_return_val_if_fail (PIKA_IS_PIKA (pika), NULL);
  g_return_val_if_fail (identifier != NULL, NULL);
  g_return_val_if_fail (label != NULL, NULL);
  g_return_val_if_fail (tooltip != NULL, NULL);
  g_return_val_if_fail (help_id != NULL, NULL);
  g_return_val_if_fail (paint_core_name != NULL, NULL);
  g_return_val_if_fail (icon_name != NULL, NULL);

  paint_info = (PikaPaintInfo *)
    pika_container_get_child_by_name (pika->paint_info_list, paint_core_name);

  g_return_val_if_fail (PIKA_IS_PAINT_INFO (paint_info), NULL);

  tool_info = g_object_new (PIKA_TYPE_TOOL_INFO,
                            "name",      identifier,
                            "icon-name", icon_name,
                            NULL);

  tool_info->pika              = pika;
  tool_info->tool_type         = tool_type;
  tool_info->tool_options_type = tool_options_type;
  tool_info->context_props     = context_props;

  tool_info->label             = g_strdup (label);
  tool_info->tooltip           = g_strdup (tooltip);

  tool_info->menu_label        = g_strdup (menu_label);
  tool_info->menu_accel        = g_strdup (menu_accel);

  tool_info->help_domain       = g_strdup (help_domain);
  tool_info->help_id           = g_strdup (help_id);

  tool_info->paint_info        = paint_info;

  if (tool_info->tool_options_type == paint_info->paint_options_type)
    {
      tool_info->tool_options = g_object_ref (PIKA_TOOL_OPTIONS (paint_info->paint_options));
    }
  else
    {
      tool_info->tool_options = g_object_new (tool_info->tool_options_type,
                                              "pika", pika,
                                              "name", identifier,
                                              NULL);
    }

  g_object_set (tool_info->tool_options,
                "tool",      tool_info,
                "tool-info", tool_info, NULL);

  pika_tool_options_set_gui_mode (tool_info->tool_options, TRUE);

  if (tool_info->tool_options_type != PIKA_TYPE_TOOL_OPTIONS)
    {
      PikaContainer *presets;

      presets = pika_data_factory_get_container (pika->tool_preset_factory);

      tool_info->presets =
        pika_filtered_container_new (presets,
                                     pika_tool_info_filter_preset,
                                     tool_info);
    }

  return tool_info;
}

void
pika_tool_info_set_standard (Pika         *pika,
                             PikaToolInfo *tool_info)
{
  g_return_if_fail (PIKA_IS_PIKA (pika));
  g_return_if_fail (! tool_info || PIKA_IS_TOOL_INFO (tool_info));

  g_set_object (&pika->standard_tool_info, tool_info);
}

PikaToolInfo *
pika_tool_info_get_standard (Pika *pika)
{
  g_return_val_if_fail (PIKA_IS_PIKA (pika), NULL);

  return pika->standard_tool_info;
}

gchar *
pika_tool_info_get_action_name (PikaToolInfo *tool_info)
{
  const gchar *identifier;
  gchar       *tmp;
  gchar       *name;

  g_return_val_if_fail (PIKA_IS_TOOL_INFO (tool_info), NULL);

  identifier = pika_object_get_name (PIKA_OBJECT (tool_info));

  g_return_val_if_fail (g_str_has_prefix (identifier, "pika-"), NULL);
  g_return_val_if_fail (g_str_has_suffix (identifier, "-tool"), NULL);

  tmp = g_strndup (identifier + strlen ("pika-"),
                    strlen (identifier) - strlen ("pika--tool"));

  name = g_strdup_printf ("tools-%s", tmp);

  g_free (tmp);

  return name;
}

GFile *
pika_tool_info_get_options_file (PikaToolInfo *tool_info,
                                 const gchar  *suffix)
{
  gchar *basename;
  GFile *file;

  g_return_val_if_fail (PIKA_IS_TOOL_INFO (tool_info), NULL);

  /* also works for a NULL suffix */
  basename = g_strconcat (pika_object_get_name (tool_info), suffix, NULL);

  file = pika_directory_file ("tool-options", basename, NULL);
  g_free (basename);

  return file;
}
