/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995-1999 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <string.h>

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikabase/pikabase.h"
#include "libpikaconfig/pikaconfig.h"
#include "libpikawidgets/pikawidgets.h"

#include "dialogs-types.h"

#include "config/pikaguiconfig.h"

#include "core/pika.h"
#include "core/pikacontext.h"
#include "core/pikaimage.h"
#include "core/pikaimage-new.h"
#include "core/pikatemplate.h"

#include "widgets/pikacontainercombobox.h"
#include "widgets/pikahelp-ids.h"
#include "widgets/pikamessagebox.h"
#include "widgets/pikamessagedialog.h"
#include "widgets/pikatemplateeditor.h"
#include "widgets/pikawidgets-utils.h"

#include "image-new-dialog.h"

#include "pika-intl.h"


#define RESPONSE_RESET 1

typedef struct
{
  GtkWidget    *dialog;
  GtkWidget    *confirm_dialog;

  GtkWidget    *combo;
  GtkWidget    *editor;

  PikaContext  *context;
  PikaTemplate *template;
} ImageNewDialog;


/*  local function prototypes  */

static void   image_new_dialog_free      (ImageNewDialog *private);
static void   image_new_dialog_response  (GtkWidget      *widget,
                                          gint            response_id,
                                          ImageNewDialog *private);
static void   image_new_template_changed (PikaContext    *context,
                                          PikaTemplate   *template,
                                          ImageNewDialog *private);
static void   image_new_confirm_dialog   (ImageNewDialog *private);
static void   image_new_create_image     (ImageNewDialog *private);


/*  public functions  */

GtkWidget *
image_new_dialog_new (PikaContext *context)
{
  ImageNewDialog *private;
  GtkWidget      *dialog;
  GtkWidget      *main_vbox;
  GtkWidget      *hbox;
  GtkWidget      *label;
  PikaSizeEntry  *entry;

  g_return_val_if_fail (PIKA_IS_CONTEXT (context), NULL);

  private = g_slice_new0 (ImageNewDialog);

  private->context  = pika_context_new (context->pika, "image-new-dialog",
                                        context);
  private->template = g_object_new (PIKA_TYPE_TEMPLATE, NULL);

  private->dialog = dialog =
    pika_dialog_new (_("Create a New Image"),
                     "pika-image-new",
                     NULL, 0,
                     pika_standard_help_func, PIKA_HELP_FILE_NEW,

                     _("_Reset"),  RESPONSE_RESET,
                     _("_Cancel"), GTK_RESPONSE_CANCEL,
                     _("_OK"),     GTK_RESPONSE_OK,

                     NULL);
  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

  pika_dialog_set_alternative_button_order (GTK_DIALOG (dialog),
                                           RESPONSE_RESET,
                                           GTK_RESPONSE_OK,
                                           GTK_RESPONSE_CANCEL,
                                           -1);

  gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

  g_object_set_data_full (G_OBJECT (dialog),
                          "pika-image-new-dialog", private,
                          (GDestroyNotify) image_new_dialog_free);

  g_signal_connect (dialog, "response",
                    G_CALLBACK (image_new_dialog_response),
                    private);

  main_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 12);
  gtk_container_set_border_width (GTK_CONTAINER (main_vbox), 12);
  gtk_box_pack_start (GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))),
                      main_vbox, TRUE, TRUE, 0);
  gtk_widget_show (main_vbox);

  /*  The template combo  */
  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
  gtk_box_pack_start (GTK_BOX (main_vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show (hbox);

  label = gtk_label_new_with_mnemonic (_("_Template:"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  private->combo = g_object_new (PIKA_TYPE_CONTAINER_COMBO_BOX,
                                 "container",         context->pika->templates,
                                 "context",           private->context,
                                 "view-size",         16,
                                 "view-border-width", 0,
                                 "ellipsize",         PANGO_ELLIPSIZE_NONE,
                                 "focus-on-click",    FALSE,
                                 NULL);
  gtk_box_pack_start (GTK_BOX (hbox), private->combo, TRUE, TRUE, 0);
  gtk_widget_show (private->combo);

  gtk_label_set_mnemonic_widget (GTK_LABEL (label), private->combo);

  g_signal_connect (private->context, "template-changed",
                    G_CALLBACK (image_new_template_changed),
                    private);

  /*  Template editor  */
  private->editor = pika_template_editor_new (private->template, context->pika,
                                              FALSE);
  gtk_box_pack_start (GTK_BOX (main_vbox), private->editor, FALSE, FALSE, 0);
  gtk_widget_show (private->editor);

  entry = PIKA_SIZE_ENTRY (pika_template_editor_get_size_se (PIKA_TEMPLATE_EDITOR (private->editor)));
  pika_size_entry_set_activates_default (entry, TRUE);
  pika_size_entry_grab_focus (entry);

  image_new_template_changed (private->context,
                              pika_context_get_template (private->context),
                              private);

  return dialog;
}

void
image_new_dialog_set (GtkWidget    *dialog,
                      PikaImage    *image,
                      PikaTemplate *template)
{
  ImageNewDialog *private;

  g_return_if_fail (PIKA_IS_DIALOG (dialog));
  g_return_if_fail (image == NULL || PIKA_IS_IMAGE (image));
  g_return_if_fail (template == NULL || PIKA_IS_TEMPLATE (template));

  private = g_object_get_data (G_OBJECT (dialog), "pika-image-new-dialog");

  g_return_if_fail (private != NULL);

  pika_context_set_template (private->context, template);

  if (! template)
    {
      template = pika_image_new_get_last_template (private->context->pika,
                                                   image);

      image_new_template_changed (private->context, template, private);

      g_object_unref (template);
    }
}


/*  private functions  */

static void
image_new_dialog_free (ImageNewDialog *private)
{
  g_object_unref (private->context);
  g_object_unref (private->template);

  g_slice_free (ImageNewDialog, private);
}

static void
image_new_dialog_response (GtkWidget      *dialog,
                           gint            response_id,
                           ImageNewDialog *private)
{
  switch (response_id)
    {
    case RESPONSE_RESET:
      pika_config_sync (G_OBJECT (private->context->pika->config->default_image),
                        G_OBJECT (private->template), 0);
      pika_context_set_template (private->context, NULL);
      break;

    case GTK_RESPONSE_OK:
      if (pika_template_get_initial_size (private->template) >
          PIKA_GUI_CONFIG (private->context->pika->config)->max_new_image_size)
        image_new_confirm_dialog (private);
      else
        image_new_create_image (private);
      break;

    default:
      gtk_widget_destroy (dialog);
      break;
    }
}

static void
image_new_template_changed (PikaContext    *context,
                            PikaTemplate   *template,
                            ImageNewDialog *private)
{
  PikaTemplateEditor *editor;
  GtkWidget          *chain;
  gdouble             xres, yres;
  gchar              *comment;

  if (! template)
    return;

  editor = PIKA_TEMPLATE_EDITOR (private->editor);
  chain  = pika_template_editor_get_resolution_chain (editor);

  xres = pika_template_get_resolution_x (template);
  yres = pika_template_get_resolution_y (template);

  pika_chain_button_set_active (PIKA_CHAIN_BUTTON (chain),
                                ABS (xres - yres) < PIKA_MIN_RESOLUTION);

  comment = (gchar *) pika_template_get_comment (template);

  if (! comment || ! strlen (comment))
    comment = g_strdup (pika_template_get_comment (private->template));
  else
    comment = NULL;

  /*  make sure the resolution values are copied first (see bug #546924)  */
  pika_config_sync (G_OBJECT (template), G_OBJECT (private->template),
                    PIKA_TEMPLATE_PARAM_COPY_FIRST);
  pika_config_sync (G_OBJECT (template), G_OBJECT (private->template), 0);

  if (comment)
    {
      g_object_set (private->template,
                    "comment", comment,
                    NULL);

      g_free (comment);
    }
}


/*  the confirm dialog  */

static void
image_new_confirm_response (GtkWidget      *dialog,
                            gint            response_id,
                            ImageNewDialog *private)
{
  gtk_widget_destroy (dialog);

  private->confirm_dialog = NULL;

  if (response_id == GTK_RESPONSE_OK)
    image_new_create_image (private);
  else
    gtk_widget_set_sensitive (private->dialog, TRUE);
}

static void
image_new_confirm_dialog (ImageNewDialog *private)
{
  PikaGuiConfig *config;
  GtkWidget     *dialog;
  gchar         *size;

  if (private->confirm_dialog)
    {
      gtk_window_present (GTK_WINDOW (private->confirm_dialog));
      return;
    }

  private->confirm_dialog =
    dialog = pika_message_dialog_new (_("Confirm Image Size"),
                                      PIKA_ICON_DIALOG_WARNING,
                                      private->dialog,
                                      GTK_DIALOG_DESTROY_WITH_PARENT,
                                      pika_standard_help_func, NULL,

                                      _("_Cancel"), GTK_RESPONSE_CANCEL,
                                      _("_OK"),     GTK_RESPONSE_OK,

                                      NULL);

  pika_dialog_set_alternative_button_order (GTK_DIALOG (private->confirm_dialog),
                                           GTK_RESPONSE_OK,
                                           GTK_RESPONSE_CANCEL,
                                           -1);

  g_signal_connect (dialog, "response",
                    G_CALLBACK (image_new_confirm_response),
                    private);

  size = g_format_size (pika_template_get_initial_size (private->template));
  pika_message_box_set_primary_text (PIKA_MESSAGE_DIALOG (dialog)->box,
                                     _("You are trying to create an image "
                                       "with a size of %s."), size);
  g_free (size);

  config = PIKA_GUI_CONFIG (private->context->pika->config);
  size = g_format_size (config->max_new_image_size);
  pika_message_box_set_text (PIKA_MESSAGE_DIALOG (dialog)->box,
                              _("An image of the chosen size will use more "
                                "memory than what is configured as "
                                "\"Maximum new image size\" in the Preferences "
                                "dialog (currently %s)."), size);
  g_free (size);

  gtk_widget_set_sensitive (private->dialog, FALSE);

  gtk_widget_show (dialog);
}

static void
image_new_create_image (ImageNewDialog *private)
{
  PikaTemplate *template = g_object_ref (private->template);
  Pika         *pika     = private->context->pika;
  PikaImage    *image;

  gtk_widget_hide (private->dialog);

  image = pika_image_new_from_template (pika, template,
                                        pika_get_user_context (pika));
  pika_create_display (pika, image, pika_template_get_unit (template), 1.0,
                       G_OBJECT (pika_widget_get_monitor (private->dialog)));
  g_object_unref (image);

  gtk_widget_destroy (private->dialog);

  pika_image_new_set_last_template (pika, template);

  g_object_unref (template);
}
