/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * Copyright (C) 2003  Henrik Brix Andersen <brix@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikaconfig/pikaconfig.h"
#include "libpikawidgets/pikawidgets.h"

#include "dialogs-types.h"

#include "core/pika.h"
#include "core/pikadrawable.h"
#include "core/pikaimage.h"
#include "core/pikapaintinfo.h"
#include "core/pikastrokeoptions.h"
#include "core/pikatoolinfo.h"

#include "widgets/pikacontainercombobox.h"
#include "widgets/pikacontainerview.h"
#include "widgets/pikaviewabledialog.h"
#include "widgets/pikastrokeeditor.h"

#include "stroke-dialog.h"

#include "pika-intl.h"


#define RESPONSE_RESET 1


typedef struct _StrokeDialog StrokeDialog;

struct _StrokeDialog
{
  GList              *items;
  GList              *drawables;
  PikaContext        *context;
  PikaStrokeOptions  *options;
  PikaStrokeCallback  callback;
  gpointer            user_data;

  GtkWidget          *tool_combo;
};


/*  local function prototypes  */

static void  stroke_dialog_free        (StrokeDialog *private);
static void  stroke_dialog_response    (GtkWidget    *dialog,
                                        gint          response_id,
                                        StrokeDialog *private);
static void  stroke_dialog_expand_tabs (GtkWidget    *widget,
                                        gpointer      data);


/*  public function  */

GtkWidget *
stroke_dialog_new (GList              *items,
                   GList              *drawables,
                   PikaContext        *context,
                   const gchar        *title,
                   const gchar        *icon_name,
                   const gchar        *help_id,
                   GtkWidget          *parent,
                   PikaStrokeOptions  *options,
                   PikaStrokeCallback  callback,
                   gpointer            user_data)
{
  StrokeDialog *private;
  PikaImage    *image;
  GtkWidget    *dialog;
  GtkWidget    *main_vbox;
  GtkWidget    *switcher;
  GtkWidget    *stack;
  GtkWidget    *frame;

  g_return_val_if_fail (items, NULL);
  g_return_val_if_fail (drawables, NULL);
  g_return_val_if_fail (PIKA_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (icon_name != NULL, NULL);
  g_return_val_if_fail (help_id != NULL, NULL);
  g_return_val_if_fail (parent == NULL || GTK_IS_WIDGET (parent), NULL);
  g_return_val_if_fail (callback != NULL, NULL);

  image = pika_item_get_image (items->data);

  private = g_slice_new0 (StrokeDialog);

  private->items     = g_list_copy (items);
  private->drawables = g_list_copy (drawables);
  private->context   = context;
  private->options   = pika_stroke_options_new (context->pika, context, TRUE);
  private->callback  = callback;
  private->user_data = user_data;

  pika_config_sync (G_OBJECT (options),
                    G_OBJECT (private->options), 0);

  dialog = pika_viewable_dialog_new (g_list_copy (items), context,
                                     title, "pika-stroke-options",
                                     icon_name,
                                     _("Choose Stroke Style"),
                                     parent,
                                     pika_standard_help_func,
                                     help_id,

                                     _("_Reset"),  RESPONSE_RESET,
                                     _("_Cancel"), GTK_RESPONSE_CANCEL,
                                     _("_Stroke"), GTK_RESPONSE_OK,

                                     NULL);

  pika_dialog_set_alternative_button_order (GTK_DIALOG (dialog),
                                           RESPONSE_RESET,
                                           GTK_RESPONSE_OK,
                                           GTK_RESPONSE_CANCEL,
                                           -1);

  gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

  g_object_weak_ref (G_OBJECT (dialog),
                     (GWeakNotify) stroke_dialog_free, private);

  g_signal_connect (dialog, "response",
                    G_CALLBACK (stroke_dialog_response),
                    private);

  main_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 12);
  gtk_container_set_border_width (GTK_CONTAINER (main_vbox), 12);
  gtk_box_pack_start (GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (dialog))),
                      main_vbox, TRUE, TRUE, 0);
  gtk_widget_show (main_vbox);


  /* switcher */

  switcher = gtk_stack_switcher_new ();
  gtk_box_pack_start (GTK_BOX (main_vbox), switcher, TRUE, TRUE, 0);
  gtk_widget_show (switcher);

  stack = gtk_stack_new ();
  gtk_stack_switcher_set_stack (GTK_STACK_SWITCHER (switcher),
                                GTK_STACK (stack));
  gtk_box_pack_start (GTK_BOX (main_vbox), stack, TRUE, TRUE, 0);
  gtk_widget_show (stack);

  /*  the stroke frame  */

  frame = pika_frame_new (NULL);
  gtk_stack_add_titled (GTK_STACK (stack),
                        frame,
                        "stroke-tool",
                        "Line");
  gtk_widget_show (frame);

  {
    GtkWidget *stroke_editor;
    gdouble    xres;
    gdouble    yres;

    pika_image_get_resolution (image, &xres, &yres);

    stroke_editor = pika_stroke_editor_new (private->options, yres, FALSE,
                                            FALSE);
    gtk_container_add (GTK_CONTAINER (frame), stroke_editor);
    gtk_widget_show (stroke_editor);

  }


  /*  the paint tool frame  */

  frame = pika_frame_new (NULL);
  gtk_stack_add_titled (GTK_STACK (stack),
                        frame,
                        "paint-tool",
                        "Paint tool");
  gtk_widget_show (frame);

  {
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *label;
    GtkWidget *combo;
    GtkWidget *button;

    vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
    gtk_container_add (GTK_CONTAINER (frame), vbox);
    gtk_widget_show (vbox);

    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show (hbox);

    label = gtk_label_new_with_mnemonic (_("P_aint tool:"));
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    combo = pika_container_combo_box_new (image->pika->paint_info_list,
                                          PIKA_CONTEXT (private->options),
                                          16, 0);
    gtk_box_pack_start (GTK_BOX (hbox), combo, TRUE, TRUE, 0);
    gtk_widget_show (combo);

    private->tool_combo = combo;

    button = pika_prop_check_button_new (G_OBJECT (private->options),
                                         "emulate-brush-dynamics",
                                         _("_Emulate brush dynamics"));
    gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  }

  /* Setting hexpand property of all tabs to true */
  gtk_container_foreach (GTK_CONTAINER (switcher),
                         stroke_dialog_expand_tabs,
                         NULL);

  return dialog;
}


/*  private functions  */

static void
stroke_dialog_free (StrokeDialog *private)
{
  g_object_unref (private->options);
  g_list_free (private->drawables);
  g_list_free (private->items);

  g_slice_free (StrokeDialog, private);
}

static void
stroke_dialog_response (GtkWidget    *dialog,
                        gint          response_id,
                        StrokeDialog *private)
{
  switch (response_id)
    {
    case RESPONSE_RESET:
      {
        PikaToolInfo *tool_info = pika_context_get_tool (private->context);

        pika_config_reset (PIKA_CONFIG (private->options));

        pika_container_view_select_item (PIKA_CONTAINER_VIEW (private->tool_combo),
                                         PIKA_VIEWABLE (tool_info->paint_info));

      }
      break;

    case GTK_RESPONSE_OK:
      private->callback (dialog,
                         private->items,
                         private->drawables,
                         private->context,
                         private->options,
                         private->user_data);
      break;

    default:
      gtk_widget_destroy (dialog);
      break;
    }
}

static void
stroke_dialog_expand_tabs (GtkWidget *widget, gpointer data)
{
  gtk_widget_set_hexpand (widget, TRUE);
}
