/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikawidgets/pikawidgets.h"

#include "dialogs-types.h"

#include "core/pikacontext.h"
#include "core/pikaimage.h"

#include "vectors/pikavectors.h"

#include "item-options-dialog.h"
#include "vectors-options-dialog.h"

#include "pika-intl.h"


typedef struct _VectorsOptionsDialog VectorsOptionsDialog;

struct _VectorsOptionsDialog
{
  PikaVectorsOptionsCallback  callback;
  gpointer                    user_data;
};


/*  local function prototypes  */

static void  vectors_options_dialog_free     (VectorsOptionsDialog *private);
static void  vectors_options_dialog_callback (GtkWidget            *dialog,
                                              PikaImage            *image,
                                              PikaItem             *item,
                                              PikaContext          *context,
                                              const gchar          *item_name,
                                              gboolean              item_visible,
                                              PikaColorTag          item_color_tag,
                                              gboolean              item_lock_content,
                                              gboolean              item_lock_position,
                                              gboolean              item_lock_visibility,
                                              gpointer              user_data);


/*  public functions  */

GtkWidget *
vectors_options_dialog_new (PikaImage                  *image,
                            PikaVectors                *vectors,
                            PikaContext                *context,
                            GtkWidget                  *parent,
                            const gchar                *title,
                            const gchar                *role,
                            const gchar                *icon_name,
                            const gchar                *desc,
                            const gchar                *help_id,
                            const gchar                *vectors_name,
                            gboolean                    vectors_visible,
                            PikaColorTag                vectors_color_tag,
                            gboolean                    vectors_lock_content,
                            gboolean                    vectors_lock_position,
                            gboolean                    vectors_lock_visibility,
                            PikaVectorsOptionsCallback  callback,
                            gpointer                    user_data)
{
  VectorsOptionsDialog *private;
  GtkWidget            *dialog;

  g_return_val_if_fail (PIKA_IS_IMAGE (image), NULL);
  g_return_val_if_fail (vectors == NULL || PIKA_IS_VECTORS (vectors), NULL);
  g_return_val_if_fail (PIKA_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (GTK_IS_WIDGET (parent), NULL);
  g_return_val_if_fail (title != NULL, NULL);
  g_return_val_if_fail (role != NULL, NULL);
  g_return_val_if_fail (icon_name != NULL, NULL);
  g_return_val_if_fail (desc != NULL, NULL);
  g_return_val_if_fail (help_id != NULL, NULL);
  g_return_val_if_fail (callback != NULL, NULL);

  private = g_slice_new0 (VectorsOptionsDialog);

  private->callback  = callback;
  private->user_data = user_data;

  dialog = item_options_dialog_new (image, PIKA_ITEM (vectors), context,
                                    parent, title, role,
                                    icon_name, desc, help_id,
                                    _("Path _name:"),
                                    PIKA_ICON_LOCK_CONTENT,
                                    _("Lock path _strokes"),
                                    _("Lock path _position"),
                                    _("Lock path _visibility"),
                                    vectors_name,
                                    vectors_visible,
                                    vectors_color_tag,
                                    vectors_lock_content,
                                    vectors_lock_position,
                                    vectors_lock_visibility,
                                    vectors_options_dialog_callback,
                                    private);

  g_object_weak_ref (G_OBJECT (dialog),
                     (GWeakNotify) vectors_options_dialog_free, private);

  return dialog;
}


/*  private functions  */

static void
vectors_options_dialog_free (VectorsOptionsDialog *private)
{
  g_slice_free (VectorsOptionsDialog, private);
}

static void
vectors_options_dialog_callback (GtkWidget    *dialog,
                                 PikaImage    *image,
                                 PikaItem     *item,
                                 PikaContext  *context,
                                 const gchar  *item_name,
                                 gboolean      item_visible,
                                 PikaColorTag  item_color_tag,
                                 gboolean      item_lock_content,
                                 gboolean      item_lock_position,
                                 gboolean      item_lock_visibility,
                                 gpointer      user_data)
{
  VectorsOptionsDialog *private = user_data;

  private->callback (dialog,
                     image,
                     PIKA_VECTORS (item),
                     context,
                     item_name,
                     item_visible,
                     item_color_tag,
                     item_lock_content,
                     item_lock_position,
                     item_lock_visibility,
                     private->user_data);
}
