/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikacanvaspen.c
 * Copyright (C) 2010 Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikabase/pikabase.h"
#include "libpikacolor/pikacolor.h"
#include "libpikamath/pikamath.h"

#include "display-types.h"

#include "core/pikacontext.h"
#include "core/pikaparamspecs.h"

#include "pikacanvas-style.h"
#include "pikacanvaspen.h"
#include "pikadisplayshell.h"


enum
{
  PROP_0,
  PROP_COLOR,
  PROP_WIDTH
};


typedef struct _PikaCanvasPenPrivate PikaCanvasPenPrivate;

struct _PikaCanvasPenPrivate
{
  PikaRGB color;
  gint    width;
};

#define GET_PRIVATE(pen) \
        ((PikaCanvasPenPrivate *) pika_canvas_pen_get_instance_private ((PikaCanvasPen *) (pen)))


/*  local function prototypes  */

static void             pika_canvas_pen_set_property (GObject        *object,
                                                      guint           property_id,
                                                      const GValue   *value,
                                                      GParamSpec     *pspec);
static void             pika_canvas_pen_get_property (GObject        *object,
                                                      guint           property_id,
                                                      GValue         *value,
                                                      GParamSpec     *pspec);
static cairo_region_t * pika_canvas_pen_get_extents  (PikaCanvasItem *item);
static void             pika_canvas_pen_stroke       (PikaCanvasItem *item,
                                                      cairo_t        *cr);


G_DEFINE_TYPE_WITH_PRIVATE (PikaCanvasPen, pika_canvas_pen,
                            PIKA_TYPE_CANVAS_POLYGON)

#define parent_class pika_canvas_pen_parent_class


static void
pika_canvas_pen_class_init (PikaCanvasPenClass *klass)
{
  GObjectClass        *object_class = G_OBJECT_CLASS (klass);
  PikaCanvasItemClass *item_class   = PIKA_CANVAS_ITEM_CLASS (klass);

  object_class->set_property = pika_canvas_pen_set_property;
  object_class->get_property = pika_canvas_pen_get_property;

  item_class->get_extents    = pika_canvas_pen_get_extents;
  item_class->stroke         = pika_canvas_pen_stroke;

  g_object_class_install_property (object_class, PROP_COLOR,
                                   pika_param_spec_rgb ("color", NULL, NULL,
                                                        FALSE, NULL,
                                                        PIKA_PARAM_READWRITE));

  g_object_class_install_property (object_class, PROP_WIDTH,
                                   g_param_spec_int ("width", NULL, NULL,
                                                     1, G_MAXINT, 1,
                                                     PIKA_PARAM_READWRITE));
}

static void
pika_canvas_pen_init (PikaCanvasPen *pen)
{
}

static void
pika_canvas_pen_set_property (GObject      *object,
                              guint         property_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
  PikaCanvasPenPrivate *private = GET_PRIVATE (object);

  switch (property_id)
    {
    case PROP_COLOR:
      pika_value_get_rgb (value, &private->color);
      break;
    case PROP_WIDTH:
      private->width = g_value_get_int (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_canvas_pen_get_property (GObject    *object,
                              guint       property_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
  PikaCanvasPenPrivate *private = GET_PRIVATE (object);

  switch (property_id)
    {
    case PROP_COLOR:
      pika_value_set_rgb (value, &private->color);
      break;
    case PROP_WIDTH:
      g_value_set_int (value, private->width);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static cairo_region_t *
pika_canvas_pen_get_extents (PikaCanvasItem *item)
{
  PikaCanvasPenPrivate *private = GET_PRIVATE (item);
  cairo_region_t       *region;

  region = PIKA_CANVAS_ITEM_CLASS (parent_class)->get_extents (item);

  if (region)
    {
      cairo_rectangle_int_t rectangle;

      cairo_region_get_extents (region, &rectangle);

      rectangle.x      -= ceil (private->width / 2.0);
      rectangle.y      -= ceil (private->width / 2.0);
      rectangle.width  += private->width + 1;
      rectangle.height += private->width + 1;

      cairo_region_union_rectangle (region, &rectangle);
    }

  return region;
}

static void
pika_canvas_pen_stroke (PikaCanvasItem *item,
                        cairo_t        *cr)
{
  PikaCanvasPenPrivate *private = GET_PRIVATE (item);

  pika_canvas_set_pen_style (pika_canvas_item_get_canvas (item), cr,
                             &private->color, private->width);
  cairo_stroke (cr);
}

PikaCanvasItem *
pika_canvas_pen_new (PikaDisplayShell  *shell,
                     const PikaVector2 *points,
                     gint               n_points,
                     PikaContext       *context,
                     PikaActiveColor    color,
                     gint               width)
{
  PikaCanvasItem *item;
  PikaArray      *array;
  PikaRGB         rgb;

  g_return_val_if_fail (PIKA_IS_DISPLAY_SHELL (shell), NULL);
  g_return_val_if_fail (points != NULL && n_points > 1, NULL);
  g_return_val_if_fail (PIKA_IS_CONTEXT (context), NULL);

  array = pika_array_new ((const guint8 *) points,
                          n_points * sizeof (PikaVector2), TRUE);

  switch (color)
    {
    case PIKA_ACTIVE_COLOR_FOREGROUND:
      pika_context_get_foreground (context, &rgb);
      break;

    case PIKA_ACTIVE_COLOR_BACKGROUND:
      pika_context_get_background (context, &rgb);
      break;
    }

  item = g_object_new (PIKA_TYPE_CANVAS_PEN,
                       "shell",  shell,
                       "points", array,
                       "color",  &rgb,
                       "width",  width,
                       NULL);

  pika_array_free (array);

  return item;
}
