/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "display-types.h"

#include "config/pikaguiconfig.h"

#include "core/pikaimage.h"

#include "widgets/pikacursor.h"
#include "widgets/pikadialogfactory.h"
#include "widgets/pikadockcontainer.h"
#include "widgets/pikasessioninfo.h"

#include "pikacanvascursor.h"
#include "pikadisplay.h"
#include "pikacursorview.h"
#include "pikadisplayshell.h"
#include "pikadisplayshell-cursor.h"
#include "pikadisplayshell-transform.h"
#include "pikastatusbar.h"


static void  pika_display_shell_real_set_cursor (PikaDisplayShell   *shell,
                                                 PikaCursorType      cursor_type,
                                                 PikaToolCursorType  tool_cursor,
                                                 PikaCursorModifier  modifier,
                                                 gboolean            always_install);


/*  public functions  */

void
pika_display_shell_set_cursor (PikaDisplayShell   *shell,
                               PikaCursorType      cursor_type,
                               PikaToolCursorType  tool_cursor,
                               PikaCursorModifier  modifier)
{
  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  if (! shell->using_override_cursor)
    {
      pika_display_shell_real_set_cursor (shell,
                                          cursor_type,
                                          tool_cursor,
                                          modifier,
                                          FALSE);
    }
}

void
pika_display_shell_unset_cursor (PikaDisplayShell *shell)
{
  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  if (! shell->using_override_cursor)
    {
      pika_display_shell_real_set_cursor (shell,
                                          (PikaCursorType) -1, 0, 0, FALSE);
    }
}

void
pika_display_shell_set_override_cursor (PikaDisplayShell *shell,
                                        PikaCursorType    cursor_type)
{
  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  if (! shell->using_override_cursor ||
      (shell->using_override_cursor &&
       shell->override_cursor != cursor_type))
    {
      shell->override_cursor       = cursor_type;
      shell->using_override_cursor = TRUE;

      pika_cursor_set (shell->canvas,
                       shell->cursor_handedness,
                       cursor_type,
                       PIKA_TOOL_CURSOR_NONE,
                       PIKA_CURSOR_MODIFIER_NONE);
    }
}

void
pika_display_shell_unset_override_cursor (PikaDisplayShell *shell)
{
  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  if (shell->using_override_cursor)
    {
      shell->using_override_cursor = FALSE;

      pika_display_shell_real_set_cursor (shell,
                                          shell->current_cursor,
                                          shell->tool_cursor,
                                          shell->cursor_modifier,
                                          TRUE);
    }
}

void
pika_display_shell_update_software_cursor (PikaDisplayShell    *shell,
                                           PikaCursorPrecision  precision,
                                           gint                 display_x,
                                           gint                 display_y,
                                           gdouble              image_x,
                                           gdouble              image_y)
{
  PikaImageWindow   *image_window;
  PikaDialogFactory *factory;
  PikaStatusbar     *statusbar;
  GtkWidget         *widget;
  PikaImage         *image;

  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  image = pika_display_get_image (shell->display);

  if (shell->draw_cursor &&
      shell->proximity   &&
      display_x >= 0     &&
      display_y >= 0)
    {
      pika_canvas_item_begin_change (shell->cursor);

      pika_canvas_cursor_set (shell->cursor,
                              display_x,
                              display_y);
      pika_canvas_item_set_visible (shell->cursor, TRUE);

      pika_canvas_item_end_change (shell->cursor);
    }
  else
    {
      pika_canvas_item_set_visible (shell->cursor, FALSE);
    }

  /*  use the passed image_coords for the statusbar because they are
   *  possibly snapped...
   */
  statusbar = pika_display_shell_get_statusbar (shell);

  pika_statusbar_update_cursor (statusbar, precision, image_x, image_y);

  image_window = pika_display_shell_get_window (shell);
  factory = pika_dock_container_get_dialog_factory (PIKA_DOCK_CONTAINER (image_window));

  widget = pika_dialog_factory_find_widget (factory, "pika-cursor-view");

  if (widget)
    {
      GtkWidget *cursor_view = gtk_bin_get_child (GTK_BIN (widget));

      if (cursor_view)
        {
          gint t_x = -1;
          gint t_y = -1;

          /*  ...but use the unsnapped display_coords for the info window  */
          if (display_x >= 0 && display_y >= 0)
            pika_display_shell_untransform_xy (shell, display_x, display_y,
                                               &t_x, &t_y, FALSE);

          pika_cursor_view_update_cursor (PIKA_CURSOR_VIEW (cursor_view),
                                          image, shell->unit,
                                          t_x, t_y);
        }
    }
}

void
pika_display_shell_clear_software_cursor (PikaDisplayShell *shell)
{
  PikaImageWindow   *image_window;
  PikaDialogFactory *factory;
  PikaStatusbar     *statusbar;
  GtkWidget         *widget;

  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  pika_canvas_item_set_visible (shell->cursor, FALSE);

  statusbar = pika_display_shell_get_statusbar (shell);

  pika_statusbar_clear_cursor (statusbar);

  image_window = pika_display_shell_get_window (shell);
  factory = pika_dock_container_get_dialog_factory (PIKA_DOCK_CONTAINER (image_window));

  widget = pika_dialog_factory_find_widget (factory, "pika-cursor-view");

  if (widget)
    {
      GtkWidget *cursor_view = gtk_bin_get_child (GTK_BIN (widget));

      if (cursor_view)
        pika_cursor_view_clear_cursor (PIKA_CURSOR_VIEW (cursor_view));
    }
}


/*  private functions  */

static void
pika_display_shell_real_set_cursor (PikaDisplayShell   *shell,
                                    PikaCursorType      cursor_type,
                                    PikaToolCursorType  tool_cursor,
                                    PikaCursorModifier  modifier,
                                    gboolean            always_install)
{
  PikaHandedness cursor_handedness;

  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));

  if (cursor_type == (PikaCursorType) -1)
    {
      shell->current_cursor = cursor_type;

      if (gtk_widget_is_drawable (shell->canvas))
        gdk_window_set_cursor (gtk_widget_get_window (shell->canvas), NULL);

      return;
    }

  if (cursor_type != PIKA_CURSOR_NONE &&
      cursor_type != PIKA_CURSOR_BAD  &&
      cursor_type != PIKA_CURSOR_SINGLE_DOT)
    {
      switch (shell->display->config->cursor_mode)
        {
        case PIKA_CURSOR_MODE_TOOL_ICON:
          break;

        case PIKA_CURSOR_MODE_TOOL_CROSSHAIR:
          if (cursor_type < PIKA_CURSOR_CORNER_TOP ||
              cursor_type > PIKA_CURSOR_SIDE_TOP_LEFT)
            {
              /* the corner and side cursors count as crosshair, so leave
               * them and override everything else
               */
              cursor_type = PIKA_CURSOR_CROSSHAIR_SMALL;
            }
          break;

        case PIKA_CURSOR_MODE_CROSSHAIR:
          cursor_type = PIKA_CURSOR_CROSSHAIR;
          tool_cursor = PIKA_TOOL_CURSOR_NONE;

          if (modifier != PIKA_CURSOR_MODIFIER_BAD)
            {
              /* the bad modifier is always shown */
              modifier = PIKA_CURSOR_MODIFIER_NONE;
            }
          break;
        }
    }

  cursor_type = pika_cursor_rotate (cursor_type, shell->rotate_angle);

  cursor_handedness = PIKA_GUI_CONFIG (shell->display->config)->cursor_handedness;

  if (shell->cursor_handedness != cursor_handedness ||
      shell->current_cursor    != cursor_type       ||
      shell->tool_cursor       != tool_cursor       ||
      shell->cursor_modifier   != modifier          ||
      always_install)
    {
      shell->cursor_handedness = cursor_handedness;
      shell->current_cursor    = cursor_type;
      shell->tool_cursor       = tool_cursor;
      shell->cursor_modifier   = modifier;

      pika_cursor_set (shell->canvas,
                       cursor_handedness,
                       cursor_type, tool_cursor, modifier);
    }
}
