/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikabase/pikabase.h"
#include "libpikacolor/pikacolor.h"
#include "libpikaconfig/pikaconfig.h"
#include "libpikamath/pikamath.h"
#include "libpikawidgets/pikawidgets.h"

#include "display-types.h"

#include "config/pikadisplayoptions.h"
#include "config/pikaguiconfig.h"

#include "core/pika.h"
#include "core/pika-cairo.h"
#include "core/pikaguide.h"
#include "core/pikaimage.h"
#include "core/pikaimage-grid.h"
#include "core/pikaimage-guides.h"
#include "core/pikaimage-quick-mask.h"
#include "core/pikaimage-sample-points.h"
#include "core/pikaitem.h"
#include "core/pikaitemstack.h"
#include "core/pikasamplepoint.h"
#include "core/pikatreehandler.h"

#include "vectors/pikavectors.h"

#include "widgets/pikawidgets-utils.h"

#include "pikacanvascanvasboundary.h"
#include "pikacanvasguide.h"
#include "pikacanvaslayerboundary.h"
#include "pikacanvaspath.h"
#include "pikacanvasproxygroup.h"
#include "pikacanvassamplepoint.h"
#include "pikadisplay.h"
#include "pikadisplayshell.h"
#include "pikadisplayshell-appearance.h"
#include "pikadisplayshell-callbacks.h"
#include "pikadisplayshell-expose.h"
#include "pikadisplayshell-handlers.h"
#include "pikadisplayshell-profile.h"
#include "pikadisplayshell-render.h"
#include "pikadisplayshell-rulers.h"
#include "pikadisplayshell-scale.h"
#include "pikadisplayshell-scroll.h"
#include "pikadisplayshell-selection.h"
#include "pikadisplayshell-title.h"
#include "pikaimagewindow.h"
#include "pikastatusbar.h"

#include "pika-intl.h"


/*  local function prototypes  */

static void   pika_display_shell_clean_dirty_handler        (PikaImage        *image,
                                                             PikaDirtyMask     dirty_mask,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_undo_event_handler         (PikaImage        *image,
                                                             PikaUndoEvent     event,
                                                             PikaUndo         *undo,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_grid_notify_handler        (PikaGrid         *grid,
                                                             GParamSpec       *pspec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_name_changed_handler       (PikaImage        *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_selection_invalidate_handler
                                                            (PikaImage        *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_component_visibility_changed_handler
                                                            (PikaImage        *image,
                                                             PikaChannelType   channel,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_size_changed_detailed_handler
                                                            (PikaImage        *image,
                                                             gint              previous_origin_x,
                                                             gint              previous_origin_y,
                                                             gint              previous_width,
                                                             gint              previous_height,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_resolution_changed_handler (PikaImage        *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_quick_mask_changed_handler (PikaImage        *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_guide_add_handler          (PikaImage        *image,
                                                             PikaGuide        *guide,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_guide_remove_handler       (PikaImage        *image,
                                                             PikaGuide        *guide,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_guide_move_handler         (PikaImage        *image,
                                                             PikaGuide        *guide,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_sample_point_add_handler   (PikaImage        *image,
                                                             PikaSamplePoint  *sample_point,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_sample_point_remove_handler(PikaImage        *image,
                                                             PikaSamplePoint  *sample_point,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_sample_point_move_handler  (PikaImage        *image,
                                                             PikaSamplePoint  *sample_point,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_mode_changed_handler       (PikaImage        *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_precision_changed_handler  (PikaImage        *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_profile_changed_handler    (PikaColorManaged *image,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_saved_handler              (PikaImage        *image,
                                                             GFile            *file,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_exported_handler           (PikaImage        *image,
                                                             GFile            *file,
                                                             PikaDisplayShell *shell);

static void   pika_display_shell_active_vectors_handler     (PikaImage        *image,
                                                             PikaDisplayShell *shell);

static void   pika_display_shell_vectors_freeze_handler     (PikaVectors      *vectors,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_vectors_thaw_handler       (PikaVectors      *vectors,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_vectors_visible_handler    (PikaVectors      *vectors,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_vectors_add_handler        (PikaContainer    *container,
                                                             PikaVectors      *vectors,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_vectors_remove_handler     (PikaContainer    *container,
                                                             PikaVectors      *vectors,
                                                             PikaDisplayShell *shell);

static void   pika_display_shell_check_notify_handler       (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_title_notify_handler       (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_nav_size_notify_handler    (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_monitor_res_notify_handler (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_padding_notify_handler     (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_ants_speed_notify_handler  (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void   pika_display_shell_quality_notify_handler     (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void  pika_display_shell_color_config_notify_handler (GObject          *config,
                                                             GParamSpec       *param_spec,
                                                             PikaDisplayShell *shell);
static void  pika_display_shell_display_changed_handler     (PikaContext      *context,
                                                             PikaDisplay      *display,
                                                             PikaDisplayShell *shell);


/*  public functions  */

void
pika_display_shell_connect (PikaDisplayShell *shell)
{
  PikaImage         *image;
  PikaContainer     *vectors;
  PikaDisplayConfig *config;
  PikaColorConfig   *color_config;
  PikaContext       *user_context;
  GList             *list;

  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));
  g_return_if_fail (PIKA_IS_DISPLAY (shell->display));

  image = pika_display_get_image (shell->display);

  g_return_if_fail (PIKA_IS_IMAGE (image));

  vectors = pika_image_get_vectors (image);

  config       = shell->display->config;
  color_config = PIKA_CORE_CONFIG (config)->color_management;

  user_context = pika_get_user_context (shell->display->pika);

  g_signal_connect (image, "clean",
                    G_CALLBACK (pika_display_shell_clean_dirty_handler),
                    shell);
  g_signal_connect (image, "dirty",
                    G_CALLBACK (pika_display_shell_clean_dirty_handler),
                    shell);
  g_signal_connect (image, "undo-event",
                    G_CALLBACK (pika_display_shell_undo_event_handler),
                    shell);

  g_signal_connect (pika_image_get_grid (image), "notify",
                    G_CALLBACK (pika_display_shell_grid_notify_handler),
                    shell);
  g_object_set (shell->grid, "grid", pika_image_get_grid (image), NULL);

  g_signal_connect (image, "name-changed",
                    G_CALLBACK (pika_display_shell_name_changed_handler),
                    shell);
  g_signal_connect (image, "selection-invalidate",
                    G_CALLBACK (pika_display_shell_selection_invalidate_handler),
                    shell);
  g_signal_connect (image, "component-visibility-changed",
                    G_CALLBACK (pika_display_shell_component_visibility_changed_handler),
                    shell);
  g_signal_connect (image, "size-changed-detailed",
                    G_CALLBACK (pika_display_shell_size_changed_detailed_handler),
                    shell);
  g_signal_connect (image, "resolution-changed",
                    G_CALLBACK (pika_display_shell_resolution_changed_handler),
                    shell);
  g_signal_connect (image, "quick-mask-changed",
                    G_CALLBACK (pika_display_shell_quick_mask_changed_handler),
                    shell);

  g_signal_connect (image, "guide-added",
                    G_CALLBACK (pika_display_shell_guide_add_handler),
                    shell);
  g_signal_connect (image, "guide-removed",
                    G_CALLBACK (pika_display_shell_guide_remove_handler),
                    shell);
  g_signal_connect (image, "guide-moved",
                    G_CALLBACK (pika_display_shell_guide_move_handler),
                    shell);
  for (list = pika_image_get_guides (image);
       list;
       list = g_list_next (list))
    {
      pika_display_shell_guide_add_handler (image, list->data, shell);
    }

  g_signal_connect (image, "sample-point-added",
                    G_CALLBACK (pika_display_shell_sample_point_add_handler),
                    shell);
  g_signal_connect (image, "sample-point-removed",
                    G_CALLBACK (pika_display_shell_sample_point_remove_handler),
                    shell);
  g_signal_connect (image, "sample-point-moved",
                    G_CALLBACK (pika_display_shell_sample_point_move_handler),
                    shell);
  for (list = pika_image_get_sample_points (image);
       list;
       list = g_list_next (list))
    {
      pika_display_shell_sample_point_add_handler (image, list->data, shell);
    }

  g_signal_connect (image, "mode-changed",
                    G_CALLBACK (pika_display_shell_mode_changed_handler),
                    shell);
  g_signal_connect (image, "precision-changed",
                    G_CALLBACK (pika_display_shell_precision_changed_handler),
                    shell);
  g_signal_connect (image, "profile-changed",
                    G_CALLBACK (pika_display_shell_profile_changed_handler),
                    shell);
  g_signal_connect_swapped (image, "simulation-profile-changed",
                            G_CALLBACK (pika_display_shell_profile_update),
                            shell);
  g_signal_connect_swapped (image, "simulation-intent-changed",
                            G_CALLBACK (pika_display_shell_profile_update),
                            shell);
  g_signal_connect_swapped (image, "simulation-bpc-changed",
                            G_CALLBACK (pika_display_shell_profile_update),
                            shell);

  g_signal_connect (image, "saved",
                    G_CALLBACK (pika_display_shell_saved_handler),
                    shell);
  g_signal_connect (image, "exported",
                    G_CALLBACK (pika_display_shell_exported_handler),
                    shell);

  g_signal_connect (image, "selected-vectors-changed",
                    G_CALLBACK (pika_display_shell_active_vectors_handler),
                    shell);

  shell->vectors_freeze_handler =
    pika_tree_handler_connect (vectors, "freeze",
                               G_CALLBACK (pika_display_shell_vectors_freeze_handler),
                               shell);
  shell->vectors_thaw_handler =
    pika_tree_handler_connect (vectors, "thaw",
                               G_CALLBACK (pika_display_shell_vectors_thaw_handler),
                               shell);
  shell->vectors_visible_handler =
    pika_tree_handler_connect (vectors, "visibility-changed",
                               G_CALLBACK (pika_display_shell_vectors_visible_handler),
                               shell);

  g_signal_connect (vectors, "add",
                    G_CALLBACK (pika_display_shell_vectors_add_handler),
                    shell);
  g_signal_connect (vectors, "remove",
                    G_CALLBACK (pika_display_shell_vectors_remove_handler),
                    shell);
  for (list = pika_item_stack_get_item_iter (PIKA_ITEM_STACK (vectors));
       list;
       list = g_list_next (list))
    {
      pika_display_shell_vectors_add_handler (vectors, list->data, shell);
    }

  g_signal_connect (config,
                    "notify::transparency-size",
                    G_CALLBACK (pika_display_shell_check_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::transparency-type",
                    G_CALLBACK (pika_display_shell_check_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::transparency-custom-color1",
                    G_CALLBACK (pika_display_shell_check_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::transparency-custom-color2",
                    G_CALLBACK (pika_display_shell_check_notify_handler),
                    shell);

  g_signal_connect (config,
                    "notify::image-title-format",
                    G_CALLBACK (pika_display_shell_title_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::image-status-format",
                    G_CALLBACK (pika_display_shell_title_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::navigation-preview-size",
                    G_CALLBACK (pika_display_shell_nav_size_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::monitor-resolution-from-windowing-system",
                    G_CALLBACK (pika_display_shell_monitor_res_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::monitor-xresolution",
                    G_CALLBACK (pika_display_shell_monitor_res_notify_handler),
                    shell);
  g_signal_connect (config,
                    "notify::monitor-yresolution",
                    G_CALLBACK (pika_display_shell_monitor_res_notify_handler),
                    shell);

  g_signal_connect (config->default_view,
                    "notify::padding-mode",
                    G_CALLBACK (pika_display_shell_padding_notify_handler),
                    shell);
  g_signal_connect (config->default_view,
                    "notify::padding-color",
                    G_CALLBACK (pika_display_shell_padding_notify_handler),
                    shell);
  g_signal_connect (config->default_fullscreen_view,
                    "notify::padding-mode",
                    G_CALLBACK (pika_display_shell_padding_notify_handler),
                    shell);
  g_signal_connect (config->default_fullscreen_view,
                    "notify::padding-color",
                    G_CALLBACK (pika_display_shell_padding_notify_handler),
                    shell);

  g_signal_connect (config,
                    "notify::marching-ants-speed",
                    G_CALLBACK (pika_display_shell_ants_speed_notify_handler),
                    shell);

  g_signal_connect (config,
                    "notify::zoom-quality",
                    G_CALLBACK (pika_display_shell_quality_notify_handler),
                    shell);

  g_signal_connect (color_config, "notify",
                    G_CALLBACK (pika_display_shell_color_config_notify_handler),
                    shell);

  g_signal_connect (user_context, "display-changed",
                    G_CALLBACK (pika_display_shell_display_changed_handler),
                    shell);

  pika_display_shell_active_vectors_handler     (image, shell);
  pika_display_shell_quick_mask_changed_handler (image, shell);
  pika_display_shell_profile_changed_handler    (PIKA_COLOR_MANAGED (image),
                                                 shell);
  pika_display_shell_color_config_notify_handler (G_OBJECT (color_config),
                                                  NULL, /* sync all */
                                                  shell);

  pika_canvas_layer_boundary_set_layers (PIKA_CANVAS_LAYER_BOUNDARY (shell->layer_boundary),
                                         pika_image_get_selected_layers (image));

  pika_canvas_canvas_boundary_set_image (PIKA_CANVAS_CANVAS_BOUNDARY (shell->canvas_boundary),
                                         image);

  if (shell->show_all)
    {
      pika_image_inc_show_all_count (image);

      pika_image_flush (image);
    }
}

void
pika_display_shell_disconnect (PikaDisplayShell *shell)
{
  PikaImage         *image;
  PikaContainer     *vectors;
  PikaDisplayConfig *config;
  PikaColorConfig   *color_config;
  PikaContext       *user_context;
  GList             *list;

  g_return_if_fail (PIKA_IS_DISPLAY_SHELL (shell));
  g_return_if_fail (PIKA_IS_DISPLAY (shell->display));

  image = pika_display_get_image (shell->display);

  g_return_if_fail (PIKA_IS_IMAGE (image));

  vectors = pika_image_get_vectors (image);

  config       = shell->display->config;
  color_config = PIKA_CORE_CONFIG (config)->color_management;

  user_context = pika_get_user_context (shell->display->pika);

  pika_canvas_layer_boundary_set_layers (PIKA_CANVAS_LAYER_BOUNDARY (shell->layer_boundary),
                                         NULL);

  pika_canvas_canvas_boundary_set_image (PIKA_CANVAS_CANVAS_BOUNDARY (shell->canvas_boundary),
                                         NULL);

  g_signal_handlers_disconnect_by_func (user_context,
                                        pika_display_shell_display_changed_handler,
                                        shell);

  g_signal_handlers_disconnect_by_func (color_config,
                                        pika_display_shell_color_config_notify_handler,
                                        shell);
  shell->color_config_set = FALSE;

  g_signal_handlers_disconnect_by_func (config,
                                        pika_display_shell_quality_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config,
                                        pika_display_shell_ants_speed_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config->default_fullscreen_view,
                                        pika_display_shell_padding_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config->default_view,
                                        pika_display_shell_padding_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config,
                                        pika_display_shell_monitor_res_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config,
                                        pika_display_shell_nav_size_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config,
                                        pika_display_shell_title_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (config,
                                        pika_display_shell_check_notify_handler,
                                        shell);

  g_signal_handlers_disconnect_by_func (vectors,
                                        pika_display_shell_vectors_remove_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (vectors,
                                        pika_display_shell_vectors_add_handler,
                                        shell);

  pika_tree_handler_disconnect (shell->vectors_visible_handler);
  shell->vectors_visible_handler = NULL;

  pika_tree_handler_disconnect (shell->vectors_thaw_handler);
  shell->vectors_thaw_handler = NULL;

  pika_tree_handler_disconnect (shell->vectors_freeze_handler);
  shell->vectors_freeze_handler = NULL;

  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_active_vectors_handler,
                                        shell);

  for (list = pika_item_stack_get_item_iter (PIKA_ITEM_STACK (vectors));
       list;
       list = g_list_next (list))
    {
      pika_canvas_proxy_group_remove_item (PIKA_CANVAS_PROXY_GROUP (shell->vectors),
                                           list->data);
    }

  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_exported_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_saved_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_profile_changed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_profile_update,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_precision_changed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_mode_changed_handler,
                                        shell);

  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_guide_add_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_guide_remove_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_guide_move_handler,
                                        shell);
  for (list = pika_image_get_guides (image);
       list;
       list = g_list_next (list))
    {
      pika_canvas_proxy_group_remove_item (PIKA_CANVAS_PROXY_GROUP (shell->guides),
                                           list->data);
    }

  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_sample_point_add_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_sample_point_remove_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_sample_point_move_handler,
                                        shell);
  for (list = pika_image_get_sample_points (image);
       list;
       list = g_list_next (list))
    {
      pika_canvas_proxy_group_remove_item (PIKA_CANVAS_PROXY_GROUP (shell->sample_points),
                                           list->data);
    }

  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_quick_mask_changed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_resolution_changed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_component_visibility_changed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_size_changed_detailed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_selection_invalidate_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_name_changed_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (pika_image_get_grid (image),
                                        pika_display_shell_grid_notify_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_undo_event_handler,
                                        shell);
  g_signal_handlers_disconnect_by_func (image,
                                        pika_display_shell_clean_dirty_handler,
                                        shell);

  if (shell->show_all)
    {
      pika_image_dec_show_all_count (image);

      pika_image_flush (image);
    }
}


/*  private functions  */

static void
pika_display_shell_clean_dirty_handler (PikaImage        *image,
                                        PikaDirtyMask     dirty_mask,
                                        PikaDisplayShell *shell)
{
  pika_display_shell_title_update (shell);
}

static void
pika_display_shell_undo_event_handler (PikaImage        *image,
                                       PikaUndoEvent     event,
                                       PikaUndo         *undo,
                                       PikaDisplayShell *shell)
{
  pika_display_shell_title_update (shell);
}

static void
pika_display_shell_grid_notify_handler (PikaGrid         *grid,
                                        GParamSpec       *pspec,
                                        PikaDisplayShell *shell)
{
  g_object_set (shell->grid, "grid", grid, NULL);
}

static void
pika_display_shell_name_changed_handler (PikaImage        *image,
                                         PikaDisplayShell *shell)
{
  pika_display_shell_title_update (shell);
}

static void
pika_display_shell_selection_invalidate_handler (PikaImage        *image,
                                                 PikaDisplayShell *shell)
{
  pika_display_shell_selection_undraw (shell);
}

static void
pika_display_shell_resolution_changed_handler (PikaImage        *image,
                                               PikaDisplayShell *shell)
{
  pika_display_shell_scale_update (shell);

  if (shell->dot_for_dot)
    {
      if (shell->unit != PIKA_UNIT_PIXEL)
        {
          pika_display_shell_rulers_update (shell);
        }

      pika_display_shell_scaled (shell);
    }
  else
    {
      /* A resolution change has the same effect as a size change from
       * a display shell point of view. Force a redraw of the display
       * so that we don't get any display garbage.
       */

      PikaDisplayConfig *config = shell->display->config;
      gboolean           resize_window;

      /* Resize windows only in multi-window mode */
      resize_window = (config->resize_windows_on_resize &&
                       ! PIKA_GUI_CONFIG (config)->single_window_mode);

      pika_display_shell_scale_resize (shell, resize_window, FALSE);
    }
}

static void
pika_display_shell_quick_mask_changed_handler (PikaImage        *image,
                                               PikaDisplayShell *shell)
{
  GtkImage *gtk_image;
  gboolean  quick_mask_state;

  gtk_image = GTK_IMAGE (gtk_bin_get_child (GTK_BIN (shell->quick_mask_button)));

  g_signal_handlers_block_by_func (shell->quick_mask_button,
                                   pika_display_shell_quick_mask_toggled,
                                   shell);

  quick_mask_state = pika_image_get_quick_mask_state (image);

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (shell->quick_mask_button),
                                quick_mask_state);

  if (quick_mask_state)
    gtk_image_set_from_icon_name (gtk_image, PIKA_ICON_QUICK_MASK_ON,
                                  GTK_ICON_SIZE_MENU);
  else
    gtk_image_set_from_icon_name (gtk_image, PIKA_ICON_QUICK_MASK_OFF,
                                  GTK_ICON_SIZE_MENU);

  g_signal_handlers_unblock_by_func (shell->quick_mask_button,
                                     pika_display_shell_quick_mask_toggled,
                                     shell);
}

static void
pika_display_shell_guide_add_handler (PikaImage        *image,
                                      PikaGuide        *guide,
                                      PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->guides);
  PikaCanvasItem       *item;
  PikaGuideStyle        style;

  style = pika_guide_get_style (guide);
  item = pika_canvas_guide_new (shell,
                                pika_guide_get_orientation (guide),
                                pika_guide_get_position (guide),
                                style);

  pika_canvas_proxy_group_add_item (group, guide, item);
  g_object_unref (item);
}

static void
pika_display_shell_guide_remove_handler (PikaImage        *image,
                                         PikaGuide        *guide,
                                         PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->guides);

  pika_canvas_proxy_group_remove_item (group, guide);
}

static void
pika_display_shell_guide_move_handler (PikaImage        *image,
                                       PikaGuide        *guide,
                                       PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->guides);
  PikaCanvasItem       *item;

  item = pika_canvas_proxy_group_get_item (group, guide);

  pika_canvas_guide_set (item,
                         pika_guide_get_orientation (guide),
                         pika_guide_get_position (guide));
}

static void
pika_display_shell_sample_point_add_handler (PikaImage        *image,
                                             PikaSamplePoint  *sample_point,
                                             PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->sample_points);
  PikaCanvasItem       *item;
  GList                *list;
  gint                  x;
  gint                  y;
  gint                  i;

  pika_sample_point_get_position (sample_point, &x, &y);

  item = pika_canvas_sample_point_new (shell, x, y, 0, TRUE);

  pika_canvas_proxy_group_add_item (group, sample_point, item);
  g_object_unref (item);

  for (list = pika_image_get_sample_points (image), i = 1;
       list;
       list = g_list_next (list), i++)
    {
      PikaSamplePoint *sample_point = list->data;

      item = pika_canvas_proxy_group_get_item (group, sample_point);

      if (item)
        g_object_set (item,
                      "index", i,
                      NULL);
    }
}

static void
pika_display_shell_sample_point_remove_handler (PikaImage        *image,
                                                PikaSamplePoint  *sample_point,
                                                PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->sample_points);
  GList                *list;
  gint                  i;

  pika_canvas_proxy_group_remove_item (group, sample_point);

  for (list = pika_image_get_sample_points (image), i = 1;
       list;
       list = g_list_next (list), i++)
    {
      PikaSamplePoint *sample_point = list->data;
      PikaCanvasItem  *item;

      item = pika_canvas_proxy_group_get_item (group, sample_point);

      if (item)
        g_object_set (item,
                      "index", i,
                      NULL);
    }
}

static void
pika_display_shell_sample_point_move_handler (PikaImage        *image,
                                              PikaSamplePoint  *sample_point,
                                              PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->sample_points);
  PikaCanvasItem       *item;
  gint                  x;
  gint                  y;

  item = pika_canvas_proxy_group_get_item (group, sample_point);

  pika_sample_point_get_position (sample_point, &x, &y);

  pika_canvas_sample_point_set (item, x, y);
}

static void
pika_display_shell_component_visibility_changed_handler (PikaImage        *image,
                                                         PikaChannelType   channel,
                                                         PikaDisplayShell *shell)
{
  if (channel == PIKA_CHANNEL_ALPHA && shell->show_all)
    pika_display_shell_expose_full (shell);
}

static void
pika_display_shell_size_changed_detailed_handler (PikaImage        *image,
                                                  gint              previous_origin_x,
                                                  gint              previous_origin_y,
                                                  gint              previous_width,
                                                  gint              previous_height,
                                                  PikaDisplayShell *shell)
{
  PikaDisplayConfig *config = shell->display->config;
  gboolean           resize_window;

  /* Resize windows only in multi-window mode */
  resize_window = (config->resize_windows_on_resize &&
                   ! PIKA_GUI_CONFIG (config)->single_window_mode);

  if (resize_window)
    {
      PikaImageWindow *window = pika_display_shell_get_window (shell);

      if (window && pika_image_window_get_active_shell (window) == shell)
        {
          /* If the window is resized just center the image in it when it
           * has change size
           */
          pika_image_window_shrink_wrap (window, FALSE);
        }
    }
  else
    {
      PikaImage *image      = pika_display_get_image (shell->display);
      gint       new_width  = pika_image_get_width  (image);
      gint       new_height = pika_image_get_height (image);
      gint       scaled_previous_origin_x;
      gint       scaled_previous_origin_y;
      gboolean   horizontally;
      gboolean   vertically;

      scaled_previous_origin_x = SCALEX (shell, previous_origin_x);
      scaled_previous_origin_y = SCALEY (shell, previous_origin_y);

      horizontally = (SCALEX (shell, previous_width)  >  shell->disp_width  &&
                      SCALEX (shell, new_width)       <= shell->disp_width);
      vertically   = (SCALEY (shell, previous_height) >  shell->disp_height &&
                      SCALEY (shell, new_height)      <= shell->disp_height);

      pika_display_shell_scroll_set_offset (shell,
                                            shell->offset_x + scaled_previous_origin_x,
                                            shell->offset_y + scaled_previous_origin_y);

      if (! pika_display_shell_get_infinite_canvas (shell))
        {
          pika_display_shell_scroll_center_image (shell,
                                                  horizontally, vertically);
        }

      /* The above calls might not lead to a call to
       * pika_display_shell_scroll_clamp_and_update() and
       * pika_display_shell_expose_full() in all cases because when
       * scaling the old and new scroll offset might be the same.
       *
       * We need them to be called in all cases, so simply call them
       * explicitly here at the end
       */
      pika_display_shell_scroll_clamp_and_update (shell);

      pika_display_shell_expose_full (shell);
      pika_display_shell_render_invalidate_full (shell);
    }
}

static void
pika_display_shell_mode_changed_handler (PikaImage        *image,
                                         PikaDisplayShell *shell)
{
  pika_display_shell_profile_update (shell);
}

static void
pika_display_shell_precision_changed_handler (PikaImage        *image,
                                              PikaDisplayShell *shell)
{
  pika_display_shell_profile_update (shell);
}

static void
pika_display_shell_profile_changed_handler (PikaColorManaged *image,
                                            PikaDisplayShell *shell)
{
  pika_color_managed_profile_changed (PIKA_COLOR_MANAGED (shell));
}

static void
pika_display_shell_saved_handler (PikaImage        *image,
                                  GFile            *file,
                                  PikaDisplayShell *shell)
{
  PikaStatusbar *statusbar = pika_display_shell_get_statusbar (shell);

  pika_statusbar_push_temp (statusbar, PIKA_MESSAGE_INFO,
                            PIKA_ICON_DOCUMENT_SAVE,
                            _("Image saved to '%s'"),
                            pika_file_get_utf8_name (file));
}

static void
pika_display_shell_exported_handler (PikaImage        *image,
                                     GFile            *file,
                                     PikaDisplayShell *shell)
{
  PikaStatusbar *statusbar = pika_display_shell_get_statusbar (shell);

  pika_statusbar_push_temp (statusbar, PIKA_MESSAGE_INFO,
                            PIKA_ICON_DOCUMENT_SAVE,
                            _("Image exported to '%s'"),
                            pika_file_get_utf8_name (file));
}

static void
pika_display_shell_active_vectors_handler (PikaImage        *image,
                                           PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group    = PIKA_CANVAS_PROXY_GROUP (shell->vectors);
  GList                *selected = pika_image_get_selected_vectors (image);
  GList                *list;

  for (list = pika_image_get_vectors_iter (image);
       list;
       list = g_list_next (list))
    {
      PikaVectors    *vectors = list->data;
      PikaCanvasItem *item;

      item = pika_canvas_proxy_group_get_item (group, vectors);

      pika_canvas_item_set_highlight (item,
                                      g_list_find (selected, vectors) != NULL);
    }
}

static void
pika_display_shell_vectors_freeze_handler (PikaVectors      *vectors,
                                           PikaDisplayShell *shell)
{
  /* do nothing */
}

static void
pika_display_shell_vectors_thaw_handler (PikaVectors      *vectors,
                                         PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->vectors);
  PikaCanvasItem       *item;

  item = pika_canvas_proxy_group_get_item (group, vectors);

  pika_canvas_path_set (item, pika_vectors_get_bezier (vectors));
}

static void
pika_display_shell_vectors_visible_handler (PikaVectors      *vectors,
                                            PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->vectors);
  PikaCanvasItem       *item;

  item = pika_canvas_proxy_group_get_item (group, vectors);

  pika_canvas_item_set_visible (item,
                                pika_item_get_visible (PIKA_ITEM (vectors)));
}

static void
pika_display_shell_vectors_add_handler (PikaContainer    *container,
                                        PikaVectors      *vectors,
                                        PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->vectors);
  PikaCanvasItem       *item;

  item = pika_canvas_path_new (shell,
                               pika_vectors_get_bezier (vectors),
                               0, 0,
                               FALSE,
                               PIKA_PATH_STYLE_VECTORS);
  pika_canvas_item_set_visible (item,
                                pika_item_get_visible (PIKA_ITEM (vectors)));

  pika_canvas_proxy_group_add_item (group, vectors, item);
  g_object_unref (item);
}

static void
pika_display_shell_vectors_remove_handler (PikaContainer    *container,
                                           PikaVectors      *vectors,
                                           PikaDisplayShell *shell)
{
  PikaCanvasProxyGroup *group = PIKA_CANVAS_PROXY_GROUP (shell->vectors);

  pika_canvas_proxy_group_remove_item (group, vectors);
}

static void
pika_display_shell_check_notify_handler (GObject          *config,
                                         GParamSpec       *param_spec,
                                         PikaDisplayShell *shell)
{
  PikaCanvasPaddingMode padding_mode;
  PikaRGB               padding_color;

  g_clear_pointer (&shell->checkerboard, cairo_pattern_destroy);

  pika_display_shell_get_padding (shell, &padding_mode, &padding_color);

  switch (padding_mode)
    {
    case PIKA_CANVAS_PADDING_MODE_LIGHT_CHECK:
    case PIKA_CANVAS_PADDING_MODE_DARK_CHECK:
      pika_display_shell_set_padding (shell, padding_mode, &padding_color);
      break;

    default:
      break;
    }

  pika_display_shell_expose_full (shell);
}

static void
pika_display_shell_title_notify_handler (GObject          *config,
                                         GParamSpec       *param_spec,
                                         PikaDisplayShell *shell)
{
  pika_display_shell_title_update (shell);
}

static void
pika_display_shell_nav_size_notify_handler (GObject          *config,
                                            GParamSpec       *param_spec,
                                            PikaDisplayShell *shell)
{
  g_clear_pointer (&shell->nav_popup, gtk_widget_destroy);
}

static void
pika_display_shell_monitor_res_notify_handler (GObject          *config,
                                               GParamSpec       *param_spec,
                                               PikaDisplayShell *shell)
{
  if (PIKA_DISPLAY_CONFIG (config)->monitor_res_from_gdk)
    {
      pika_get_monitor_resolution (pika_widget_get_monitor (GTK_WIDGET (shell)),
                                   &shell->monitor_xres,
                                   &shell->monitor_yres);
    }
  else
    {
      shell->monitor_xres = PIKA_DISPLAY_CONFIG (config)->monitor_xres;
      shell->monitor_yres = PIKA_DISPLAY_CONFIG (config)->monitor_yres;
    }

  pika_display_shell_scale_update (shell);

  if (! shell->dot_for_dot)
    {
      pika_display_shell_scroll_clamp_and_update (shell);

      pika_display_shell_scaled (shell);

      pika_display_shell_expose_full (shell);
      pika_display_shell_render_invalidate_full (shell);
    }
}

static void
pika_display_shell_padding_notify_handler (GObject          *config,
                                           GParamSpec       *param_spec,
                                           PikaDisplayShell *shell)
{
  PikaDisplayConfig     *display_config;
  PikaImageWindow       *window;
  gboolean               fullscreen;
  PikaCanvasPaddingMode  padding_mode;
  PikaRGB                padding_color;

  display_config = shell->display->config;

  window = pika_display_shell_get_window (shell);

  if (window)
    fullscreen = pika_image_window_get_fullscreen (window);
  else
    fullscreen = FALSE;

  /*  if the user did not set the padding mode for this display explicitly  */
  if (! shell->fullscreen_options->padding_mode_set)
    {
      padding_mode  = display_config->default_fullscreen_view->padding_mode;
      padding_color = display_config->default_fullscreen_view->padding_color;

      if (fullscreen)
        {
          pika_display_shell_set_padding (shell, padding_mode, &padding_color);
        }
      else
        {
          shell->fullscreen_options->padding_mode  = padding_mode;
          shell->fullscreen_options->padding_color = padding_color;
        }
    }

  /*  if the user did not set the padding mode for this display explicitly  */
  if (! shell->options->padding_mode_set)
    {
      padding_mode  = display_config->default_view->padding_mode;
      padding_color = display_config->default_view->padding_color;

      if (fullscreen)
        {
          shell->options->padding_mode  = padding_mode;
          shell->options->padding_color = padding_color;
        }
      else
        {
          pika_display_shell_set_padding (shell, padding_mode, &padding_color);
        }
    }
}

static void
pika_display_shell_ants_speed_notify_handler (GObject          *config,
                                              GParamSpec       *param_spec,
                                              PikaDisplayShell *shell)
{
  pika_display_shell_selection_pause (shell);
  pika_display_shell_selection_resume (shell);
}

static void
pika_display_shell_quality_notify_handler (GObject          *config,
                                           GParamSpec       *param_spec,
                                           PikaDisplayShell *shell)
{
  pika_display_shell_expose_full (shell);
  pika_display_shell_render_invalidate_full (shell);
}

static void
pika_display_shell_color_config_notify_handler (GObject          *config,
                                                GParamSpec       *param_spec,
                                                PikaDisplayShell *shell)
{
  if (param_spec)
    {
      gboolean copy = TRUE;

      if (! strcmp (param_spec->name, "mode")                                 ||
          ! strcmp (param_spec->name, "display-rendering-intent")             ||
          ! strcmp (param_spec->name, "display-use-black-point-compensation") ||
          ! strcmp (param_spec->name, "simulation-gamut-check"))
        {
          if (shell->color_config_set)
            copy = FALSE;
        }

      if (copy)
        {
          GValue value = G_VALUE_INIT;

          g_value_init (&value, param_spec->value_type);

          g_object_get_property (config,
                                 param_spec->name, &value);
          g_object_set_property (G_OBJECT (shell->color_config),
                                 param_spec->name, &value);

          g_value_unset (&value);
        }
    }
  else
    {
      pika_config_copy (PIKA_CONFIG (config),
                        PIKA_CONFIG (shell->color_config),
                        0);
      shell->color_config_set = FALSE;
    }
}

static void
pika_display_shell_display_changed_handler (PikaContext      *context,
                                            PikaDisplay      *display,
                                            PikaDisplayShell *shell)
{
  if (shell->display == display)
    pika_display_shell_update_priority_rect (shell);
}
