/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikaoperationcurves.c
 * Copyright (C) 2007 Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <cairo.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikamath/pikamath.h"

#include "operations-types.h"

#include "core/pikacurve.h"
#include "core/pikacurve-map.h"

#include "pikacurvesconfig.h"
#include "pikaoperationcurves.h"

#include "pika-intl.h"


static gboolean pika_operation_curves_process (GeglOperation       *operation,
                                               void                *in_buf,
                                               void                *out_buf,
                                               glong                samples,
                                               const GeglRectangle *roi,
                                               gint                 level);


G_DEFINE_TYPE (PikaOperationCurves, pika_operation_curves,
               PIKA_TYPE_OPERATION_POINT_FILTER)

#define parent_class pika_operation_curves_parent_class


static void
pika_operation_curves_class_init (PikaOperationCurvesClass *klass)
{
  GObjectClass                  *object_class    = G_OBJECT_CLASS (klass);
  GeglOperationClass            *operation_class = GEGL_OPERATION_CLASS (klass);
  GeglOperationPointFilterClass *point_class     = GEGL_OPERATION_POINT_FILTER_CLASS (klass);

  object_class->set_property   = pika_operation_point_filter_set_property;
  object_class->get_property   = pika_operation_point_filter_get_property;

  gegl_operation_class_set_keys (operation_class,
                                 "name",        "pika:curves",
                                 "categories",  "color",
                                 "description", _("Adjust color curves"),
                                 NULL);

  point_class->process = pika_operation_curves_process;

  g_object_class_install_property (object_class,
                                   PIKA_OPERATION_POINT_FILTER_PROP_TRC,
                                   g_param_spec_enum ("trc",
                                                      "Linear/Percptual",
                                                      "What TRC to operate on",
                                                      PIKA_TYPE_TRC_TYPE,
                                                      PIKA_TRC_NON_LINEAR,
                                                      G_PARAM_READWRITE));

  g_object_class_install_property (object_class,
                                   PIKA_OPERATION_POINT_FILTER_PROP_CONFIG,
                                   g_param_spec_object ("config",
                                                        "Config",
                                                        "The config object",
                                                        PIKA_TYPE_CURVES_CONFIG,
                                                        G_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT));
}

static void
pika_operation_curves_init (PikaOperationCurves *self)
{
}

static gboolean
pika_operation_curves_process (GeglOperation       *operation,
                               void                *in_buf,
                               void                *out_buf,
                               glong                samples,
                               const GeglRectangle *roi,
                               gint                 level)
{
  PikaOperationPointFilter *point  = PIKA_OPERATION_POINT_FILTER (operation);
  PikaCurvesConfig         *config = PIKA_CURVES_CONFIG (point->config);
  gfloat                   *src    = in_buf;
  gfloat                   *dest   = out_buf;

  if (! config)
    return FALSE;

  pika_curve_map_pixels (config->curve[0],
                         config->curve[1],
                         config->curve[2],
                         config->curve[3],
                         config->curve[4], src, dest, samples);

  return TRUE;
}
