/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"
#include "libpikaconfig/pikaconfig.h"

#include "paint-types.h"

#include "core/pika.h"
#include "core/pikadrawable.h"
#include "core/pikamybrush.h"
#include "core/pikapaintinfo.h"

#include "pikamybrushoptions.h"

#include "pika-intl.h"


enum
{
  PROP_0,
  PROP_RADIUS,
  PROP_OPAQUE,
  PROP_HARDNESS,
  PROP_ERASER,
  PROP_NO_ERASING
};


static void   pika_mybrush_options_config_iface_init (PikaConfigInterface *config_iface);

static void   pika_mybrush_options_set_property     (GObject      *object,
                                                     guint         property_id,
                                                     const GValue *value,
                                                     GParamSpec   *pspec);
static void   pika_mybrush_options_get_property     (GObject      *object,
                                                     guint         property_id,
                                                     GValue       *value,
                                                     GParamSpec   *pspec);

static void    pika_mybrush_options_mybrush_changed (PikaContext  *context,
                                                     PikaMybrush  *brush);

static void    pika_mybrush_options_reset           (PikaConfig   *config);


G_DEFINE_TYPE_WITH_CODE (PikaMybrushOptions, pika_mybrush_options,
                         PIKA_TYPE_PAINT_OPTIONS,
                         G_IMPLEMENT_INTERFACE (PIKA_TYPE_CONFIG,
                                                pika_mybrush_options_config_iface_init))

static PikaConfigInterface *parent_config_iface = NULL;


static void
pika_mybrush_options_class_init (PikaMybrushOptionsClass *klass)
{
  GObjectClass     *object_class  = G_OBJECT_CLASS (klass);
  PikaContextClass *context_class = PIKA_CONTEXT_CLASS (klass);

  object_class->set_property     = pika_mybrush_options_set_property;
  object_class->get_property     = pika_mybrush_options_get_property;

  context_class->mybrush_changed = pika_mybrush_options_mybrush_changed;

  PIKA_CONFIG_PROP_DOUBLE (object_class, PROP_RADIUS,
                           "radius",
                           _("Radius"),
                           NULL,
                           -2.0, 6.0, 1.0,
                           PIKA_PARAM_STATIC_STRINGS);

  PIKA_CONFIG_PROP_DOUBLE (object_class, PROP_OPAQUE,
                           "opaque",
                           _("Base Opacity"),
                           NULL,
                           0.0, 2.0, 1.0,
                           PIKA_PARAM_STATIC_STRINGS);

  PIKA_CONFIG_PROP_DOUBLE (object_class, PROP_HARDNESS,
                           "hardness",
                           _("Hardness"),
                           NULL,
                           0.0, 1.0, 1.0,
                           PIKA_PARAM_STATIC_STRINGS);

  PIKA_CONFIG_PROP_BOOLEAN (object_class, PROP_ERASER,
                            "eraser",
                            _("Erase with this brush"),
                            NULL,
                            FALSE,
                            PIKA_PARAM_STATIC_STRINGS);

  PIKA_CONFIG_PROP_BOOLEAN (object_class, PROP_NO_ERASING,
                            "no-erasing",
                            _("No erasing effect"),
                            _("Never decrease alpha of existing pixels"),
                            FALSE,
                            PIKA_PARAM_STATIC_STRINGS);
}

static void
pika_mybrush_options_config_iface_init (PikaConfigInterface *config_iface)
{
  parent_config_iface = g_type_interface_peek_parent (config_iface);

  config_iface->reset = pika_mybrush_options_reset;
}

static void
pika_mybrush_options_init (PikaMybrushOptions *options)
{
}

static void
pika_mybrush_options_set_property (GObject      *object,
                                   guint         property_id,
                                   const GValue *value,
                                   GParamSpec   *pspec)
{
  PikaMybrushOptions *options = PIKA_MYBRUSH_OPTIONS (object);

  switch (property_id)
    {
    case PROP_RADIUS:
      options->radius = g_value_get_double (value);
      break;
    case PROP_HARDNESS:
      options->hardness = g_value_get_double (value);
      break;
    case PROP_OPAQUE:
      options->opaque = g_value_get_double (value);
      break;
    case PROP_ERASER:
      options->eraser = g_value_get_boolean (value);
      break;
    case PROP_NO_ERASING:
      options->no_erasing = g_value_get_boolean (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_mybrush_options_get_property (GObject    *object,
                                   guint       property_id,
                                   GValue     *value,
                                   GParamSpec *pspec)
{
  PikaMybrushOptions *options = PIKA_MYBRUSH_OPTIONS (object);

  switch (property_id)
    {
    case PROP_RADIUS:
      g_value_set_double (value, options->radius);
      break;
    case PROP_OPAQUE:
      g_value_set_double (value, options->opaque);
      break;
    case PROP_HARDNESS:
      g_value_set_double (value, options->hardness);
      break;
    case PROP_ERASER:
      g_value_set_boolean (value, options->eraser);
      break;
    case PROP_NO_ERASING:
      g_value_set_boolean (value, options->no_erasing);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_mybrush_options_mybrush_changed (PikaContext *context,
                                      PikaMybrush *brush)
{
  if (brush)
    g_object_set (context,
                  "radius",   pika_mybrush_get_radius (brush),
                  "opaque",   pika_mybrush_get_opaque (brush),
                  "hardness", pika_mybrush_get_hardness (brush),
                  "eraser",   pika_mybrush_get_is_eraser (brush),
                  NULL);
}

static void
pika_mybrush_options_reset (PikaConfig *config)
{
  PikaContext *context = PIKA_CONTEXT (config);
  PikaMybrush *brush   = pika_context_get_mybrush (context);

  parent_config_iface->reset (config);

  pika_mybrush_options_mybrush_changed (context, brush);
}
