/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"

#include "paint-types.h"

#include "pikapaintcore.h"
#include "pikapaintcoreundo.h"


enum
{
  PROP_0,
  PROP_PAINT_CORE
};


static void   pika_paint_core_undo_constructed  (GObject             *object);
static void   pika_paint_core_undo_set_property (GObject             *object,
                                                 guint                property_id,
                                                 const GValue        *value,
                                                 GParamSpec          *pspec);
static void   pika_paint_core_undo_get_property (GObject             *object,
                                                 guint                property_id,
                                                 GValue              *value,
                                                 GParamSpec          *pspec);

static void   pika_paint_core_undo_pop          (PikaUndo            *undo,
                                                 PikaUndoMode         undo_mode,
                                                 PikaUndoAccumulator *accum);
static void   pika_paint_core_undo_free         (PikaUndo            *undo,
                                                 PikaUndoMode         undo_mode);


G_DEFINE_TYPE (PikaPaintCoreUndo, pika_paint_core_undo, PIKA_TYPE_UNDO)

#define parent_class pika_paint_core_undo_parent_class


static void
pika_paint_core_undo_class_init (PikaPaintCoreUndoClass *klass)
{
  GObjectClass  *object_class = G_OBJECT_CLASS (klass);
  PikaUndoClass *undo_class   = PIKA_UNDO_CLASS (klass);

  object_class->constructed  = pika_paint_core_undo_constructed;
  object_class->set_property = pika_paint_core_undo_set_property;
  object_class->get_property = pika_paint_core_undo_get_property;

  undo_class->pop            = pika_paint_core_undo_pop;
  undo_class->free           = pika_paint_core_undo_free;

  g_object_class_install_property (object_class, PROP_PAINT_CORE,
                                   g_param_spec_object ("paint-core", NULL, NULL,
                                                        PIKA_TYPE_PAINT_CORE,
                                                        PIKA_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT_ONLY));
}

static void
pika_paint_core_undo_init (PikaPaintCoreUndo *undo)
{
}

static void
pika_paint_core_undo_constructed (GObject *object)
{
  PikaPaintCoreUndo *paint_core_undo = PIKA_PAINT_CORE_UNDO (object);

  G_OBJECT_CLASS (parent_class)->constructed (object);

  pika_assert (PIKA_IS_PAINT_CORE (paint_core_undo->paint_core));

  paint_core_undo->last_coords = paint_core_undo->paint_core->start_coords;
}

static void
pika_paint_core_undo_set_property (GObject      *object,
                                   guint         property_id,
                                   const GValue *value,
                                   GParamSpec   *pspec)
{
  PikaPaintCoreUndo *paint_core_undo = PIKA_PAINT_CORE_UNDO (object);

  switch (property_id)
    {
    case PROP_PAINT_CORE:
      g_set_weak_pointer (&paint_core_undo->paint_core,
                          g_value_get_object (value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_paint_core_undo_get_property (GObject    *object,
                                   guint       property_id,
                                   GValue     *value,
                                   GParamSpec *pspec)
{
  PikaPaintCoreUndo *paint_core_undo = PIKA_PAINT_CORE_UNDO (object);

  switch (property_id)
    {
    case PROP_PAINT_CORE:
      g_value_set_object (value, paint_core_undo->paint_core);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_paint_core_undo_pop (PikaUndo              *undo,
                          PikaUndoMode           undo_mode,
                          PikaUndoAccumulator   *accum)
{
  PikaPaintCoreUndo *paint_core_undo = PIKA_PAINT_CORE_UNDO (undo);

  PIKA_UNDO_CLASS (parent_class)->pop (undo, undo_mode, accum);

  /*  only pop if the core still exists  */
  if (paint_core_undo->paint_core)
    {
      PikaCoords tmp_coords;

      tmp_coords = paint_core_undo->paint_core->last_coords;
      paint_core_undo->paint_core->last_coords = paint_core_undo->last_coords;
      paint_core_undo->last_coords = tmp_coords;
    }
}

static void
pika_paint_core_undo_free (PikaUndo     *undo,
                           PikaUndoMode  undo_mode)
{
  PikaPaintCoreUndo *paint_core_undo = PIKA_PAINT_CORE_UNDO (undo);

  g_clear_weak_pointer (&paint_core_undo->paint_core);

  PIKA_UNDO_CLASS (parent_class)->free (undo, undo_mode);
}
