/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "glib-object.h"

#include "pika-debug.h"
#include "pika-log.h"


static const GDebugKey log_keys[] =
{
  { "tool-events",        PIKA_LOG_TOOL_EVENTS        },
  { "tool-focus",         PIKA_LOG_TOOL_FOCUS         },
  { "dnd",                PIKA_LOG_DND                },
  { "help",               PIKA_LOG_HELP               },
  { "dialog-factory",     PIKA_LOG_DIALOG_FACTORY     },
  { "menus",              PIKA_LOG_MENUS              },
  { "save-dialog",        PIKA_LOG_SAVE_DIALOG        },
  { "image-scale",        PIKA_LOG_IMAGE_SCALE        },
  { "shadow-tiles",       PIKA_LOG_SHADOW_TILES       },
  { "scale",              PIKA_LOG_SCALE              },
  { "wm",                 PIKA_LOG_WM                 },
  { "floating-selection", PIKA_LOG_FLOATING_SELECTION },
  { "shm",                PIKA_LOG_SHM                },
  { "text-editing",       PIKA_LOG_TEXT_EDITING       },
  { "key-events",         PIKA_LOG_KEY_EVENTS         },
  { "auto-tab-style",     PIKA_LOG_AUTO_TAB_STYLE     },
  { "instances",          PIKA_LOG_INSTANCES          },
  { "rectangle-tool",     PIKA_LOG_RECTANGLE_TOOL     },
  { "brush-cache",        PIKA_LOG_BRUSH_CACHE        },
  { "projection",         PIKA_LOG_PROJECTION         },
  { "xcf",                PIKA_LOG_XCF                }
};

static const gchar * const log_domains[] =
{
  "Pika",
  "Pika-Actions",
  "Pika-Base",
  "Pika-Composite",
  "Pika-Config",
  "Pika-Core",
  "Pika-Dialogs",
  "Pika-Display",
  "Pika-File",
  "Pika-GEGL",
  "Pika-GUI",
  "Pika-Menus",
  "Pika-Operations",
  "Pika-PDB",
  "Pika-Paint",
  "Pika-Paint-Funcs",
  "Pika-Plug-In",
  "Pika-Text",
  "Pika-Tools",
  "Pika-Vectors",
  "Pika-Widgets",
  "Pika-XCF",
  "LibPikaBase",
  "LibPikaColor",
  "LibPikaConfig",
  "LibPikaMath",
  "LibPikaModule",
  "LibPikaThumb",
  "LibPikaWidgets",
  "GEGL",
  NULL
};


PikaLogFlags pika_log_flags = 0;


void
pika_log_init (void)
{
  const gchar *env_log_val = g_getenv ("PIKA_LOG");

  if (! env_log_val)
    env_log_val = g_getenv ("PIKA_DEBUG");

  if (env_log_val)
    g_setenv ("G_MESSAGES_DEBUG", env_log_val, TRUE);

  if (env_log_val)
    {
      /*  g_parse_debug_string() has special treatment of the string 'help',
       *  but we want to use it for the PIKA_LOG_HELP domain. "list-all"
       *  is a replacement for "help" in PIKA.
       */
      if (g_ascii_strcasecmp (env_log_val, "list-all") == 0)
        pika_log_flags = g_parse_debug_string ("help",
                                               log_keys,
                                               G_N_ELEMENTS (log_keys));
      else if (g_ascii_strcasecmp (env_log_val, "help") == 0)
        pika_log_flags = PIKA_LOG_HELP;
      else
        pika_log_flags = g_parse_debug_string (env_log_val,
                                               log_keys,
                                               G_N_ELEMENTS (log_keys));

      if (pika_log_flags & PIKA_LOG_INSTANCES)
        {
          pika_debug_enable_instances ();
        }
      else if (! pika_log_flags)
        {
          /* If the environment variable was set but no log flags are
           * set as a result, let's assume one is not sure how to use
           * the log flags and output the list of keys as a helper.
           */
          pika_log_flags = g_parse_debug_string ("help",
                                                 log_keys,
                                                 G_N_ELEMENTS (log_keys));
        }
    }
}

void
pika_log (PikaLogFlags  flags,
          const gchar  *function,
          gint          line,
          const gchar  *format,
          ...)
{
  va_list args;

  va_start (args, format);
  pika_logv (flags, function, line, format, args);
  va_end (args);
}

void
pika_logv (PikaLogFlags  flags,
           const gchar  *function,
           gint          line,
           const gchar  *format,
           va_list       args)
{
  const gchar *domain = "unknown";
  gchar       *message;
  gint         i;

  for (i = 0; i < G_N_ELEMENTS (log_keys); i++)
    if (log_keys[i].value == flags)
      {
        domain = log_keys[i].key;
        break;
      }

  if (format)
    message = g_strdup_vprintf (format, args);
  else
    message = g_strdup ("called");

  g_log (domain, G_LOG_LEVEL_DEBUG,
         "%s(%d): %s", function, line, message);

  g_free (message);
}

PikaLogHandler
pika_log_set_handler (gboolean       global,
                      GLogLevelFlags log_levels,
                      GLogFunc       log_func,
                      gpointer       user_data)
{
  PikaLogHandler handler;
  gint           n;
  gint           i;

  g_return_val_if_fail (log_func != NULL, NULL);

  n = G_N_ELEMENTS (log_domains) - (global ? 1 : 0);

  handler = g_new (guint, n + 1);

  handler[0] = n;

  for (i = 0; i < n; i++)
    {
      handler[i + 1] = g_log_set_handler (log_domains[i], log_levels,
                                          log_func, user_data);
    }

  return handler;
}

void
pika_log_remove_handler (PikaLogHandler handler)
{
  gint n;
  gint i;

  g_return_if_fail (handler != NULL);

  n = handler[0];

  for (i = 0; i < n; i++)
    g_log_remove_handler (log_domains[i], handler[i + 1]);

  g_free (handler);
}
