/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gegl.h>

#include "libpikabase/pikabase.h"
#include "libpikaconfig/pikaconfig.h"

#include "text-types.h"

#include "gegl/pika-babl.h"

#include "core/pika-memsize.h"
#include "core/pikaitem.h"
#include "core/pikaitemundo.h"

#include "pikatext.h"
#include "pikatextlayer.h"
#include "pikatextundo.h"


enum
{
  PROP_0,
  PROP_PARAM
};


static void     pika_text_undo_constructed  (GObject             *object);
static void     pika_text_undo_set_property (GObject             *object,
                                             guint                property_id,
                                             const GValue        *value,
                                             GParamSpec          *pspec);
static void     pika_text_undo_get_property (GObject             *object,
                                             guint                property_id,
                                             GValue              *value,
                                             GParamSpec          *pspec);

static gint64   pika_text_undo_get_memsize  (PikaObject          *object,
                                             gint64              *gui_size);

static void     pika_text_undo_pop          (PikaUndo            *undo,
                                             PikaUndoMode         undo_mode,
                                             PikaUndoAccumulator *accum);
static void     pika_text_undo_free         (PikaUndo            *undo,
                                             PikaUndoMode         undo_mode);


G_DEFINE_TYPE (PikaTextUndo, pika_text_undo, PIKA_TYPE_ITEM_UNDO)

#define parent_class pika_text_undo_parent_class


static void
pika_text_undo_class_init (PikaTextUndoClass *klass)
{
  GObjectClass    *object_class      = G_OBJECT_CLASS (klass);
  PikaObjectClass *pika_object_class = PIKA_OBJECT_CLASS (klass);
  PikaUndoClass   *undo_class        = PIKA_UNDO_CLASS (klass);

  object_class->constructed      = pika_text_undo_constructed;
  object_class->set_property     = pika_text_undo_set_property;
  object_class->get_property     = pika_text_undo_get_property;

  pika_object_class->get_memsize = pika_text_undo_get_memsize;

  undo_class->pop                = pika_text_undo_pop;
  undo_class->free               = pika_text_undo_free;

  g_object_class_install_property (object_class, PROP_PARAM,
                                   g_param_spec_param ("param", NULL, NULL,
                                                       G_TYPE_PARAM,
                                                       PIKA_PARAM_READWRITE |
                                                       G_PARAM_CONSTRUCT_ONLY));
}

static void
pika_text_undo_init (PikaTextUndo *undo)
{
}

static void
pika_text_undo_constructed (GObject *object)
{
  PikaTextUndo  *text_undo = PIKA_TEXT_UNDO (object);
  PikaTextLayer *layer;

  G_OBJECT_CLASS (parent_class)->constructed (object);

  pika_assert (PIKA_IS_TEXT_LAYER (PIKA_ITEM_UNDO (text_undo)->item));

  layer = PIKA_TEXT_LAYER (PIKA_ITEM_UNDO (text_undo)->item);

  switch (PIKA_UNDO (object)->undo_type)
    {
    case PIKA_UNDO_TEXT_LAYER:
      if (text_undo->pspec)
        {
          pika_assert (text_undo->pspec->owner_type == PIKA_TYPE_TEXT);

          text_undo->value = g_slice_new0 (GValue);

          g_value_init (text_undo->value, text_undo->pspec->value_type);
          g_object_get_property (G_OBJECT (layer->text),
                                 text_undo->pspec->name, text_undo->value);
        }
      else if (layer->text)
        {
          text_undo->text = pika_config_duplicate (PIKA_CONFIG (layer->text));
        }
      break;

    case PIKA_UNDO_TEXT_LAYER_MODIFIED:
      text_undo->modified = layer->modified;
      break;

    case PIKA_UNDO_TEXT_LAYER_CONVERT:
      text_undo->format = pika_drawable_get_format (PIKA_DRAWABLE (layer));
      break;

    default:
      pika_assert_not_reached ();
    }
}

static void
pika_text_undo_set_property (GObject      *object,
                             guint         property_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
  PikaTextUndo *text_undo = PIKA_TEXT_UNDO (object);

  switch (property_id)
    {
    case PROP_PARAM:
      text_undo->pspec = g_value_get_param (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_text_undo_get_property (GObject    *object,
                             guint       property_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
  PikaTextUndo *text_undo = PIKA_TEXT_UNDO (object);

  switch (property_id)
    {
    case PROP_PARAM:
      g_value_set_param (value, (GParamSpec *) text_undo->pspec);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static gint64
pika_text_undo_get_memsize (PikaObject *object,
                            gint64     *gui_size)
{
  PikaTextUndo *undo    = PIKA_TEXT_UNDO (object);
  gint64        memsize = 0;

  memsize += pika_g_value_get_memsize (undo->value);
  memsize += pika_object_get_memsize (PIKA_OBJECT (undo->text), NULL);

  return memsize + PIKA_OBJECT_CLASS (parent_class)->get_memsize (object,
                                                                  gui_size);
}

static void
pika_text_undo_pop (PikaUndo            *undo,
                    PikaUndoMode         undo_mode,
                    PikaUndoAccumulator *accum)
{
  PikaTextUndo  *text_undo = PIKA_TEXT_UNDO (undo);
  PikaTextLayer *layer     = PIKA_TEXT_LAYER (PIKA_ITEM_UNDO (undo)->item);

  PIKA_UNDO_CLASS (parent_class)->pop (undo, undo_mode, accum);

  switch (undo->undo_type)
    {
    case PIKA_UNDO_TEXT_LAYER:
      if (text_undo->pspec)
        {
          GValue *value;

          g_return_if_fail (layer->text != NULL);

          value = g_slice_new0 (GValue);
          g_value_init (value, text_undo->pspec->value_type);

          g_object_get_property (G_OBJECT (layer->text),
                                 text_undo->pspec->name, value);

          g_object_set_property (G_OBJECT (layer->text),
                                 text_undo->pspec->name, text_undo->value);

          g_value_unset (text_undo->value);
          g_slice_free (GValue, text_undo->value);

          text_undo->value = value;
        }
      else
        {
          PikaText *text;

          text = (layer->text ?
                  pika_config_duplicate (PIKA_CONFIG (layer->text)) : NULL);

          if (layer->text && text_undo->text)
            pika_config_sync (G_OBJECT (text_undo->text),
                              G_OBJECT (layer->text), 0);
          else
            pika_text_layer_set_text (layer, text_undo->text);

          if (text_undo->text)
            g_object_unref (text_undo->text);

          text_undo->text = text;
        }
      break;

    case PIKA_UNDO_TEXT_LAYER_MODIFIED:
      {
        gboolean modified;

#if 0
        g_print ("setting layer->modified from %s to %s\n",
                 layer->modified ? "TRUE" : "FALSE",
                 text_undo->modified ? "TRUE" : "FALSE");
#endif

        modified = layer->modified;
        g_object_set (layer, "modified", text_undo->modified, NULL);
        text_undo->modified = modified;

        pika_viewable_invalidate_preview (PIKA_VIEWABLE (layer));
      }
      break;

    case PIKA_UNDO_TEXT_LAYER_CONVERT:
      {
        const Babl *format;

        format = pika_drawable_get_format (PIKA_DRAWABLE (layer));
        pika_drawable_convert_type (PIKA_DRAWABLE (layer),
                                    pika_item_get_image (PIKA_ITEM (layer)),
                                    pika_babl_format_get_base_type (text_undo->format),
                                    pika_babl_format_get_precision (text_undo->format),
                                    babl_format_has_alpha (text_undo->format),
                                    NULL, NULL,
                                    GEGL_DITHER_NONE, GEGL_DITHER_NONE,
                                    FALSE, NULL);
        text_undo->format = format;
      }
      break;

    default:
      pika_assert_not_reached ();
    }
}

static void
pika_text_undo_free (PikaUndo     *undo,
                     PikaUndoMode  undo_mode)
{
  PikaTextUndo *text_undo = PIKA_TEXT_UNDO (undo);

  g_clear_object (&text_undo->text);

  if (text_undo->pspec)
    {
      g_value_unset (text_undo->value);
      g_slice_free (GValue, text_undo->value);

      text_undo->value = NULL;
      text_undo->pspec = NULL;
    }

  PIKA_UNDO_CLASS (parent_class)->free (undo, undo_mode);
}
