/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikaconfig/pikaconfig.h"
#include "libpikawidgets/pikawidgets.h"

#include "tools-types.h"

#include "core/pika-gui.h"
#include "core/pikaasync.h"
#include "core/pikacancelable.h"
#include "core/pikadrawable.h"
#include "core/pikadrawable-histogram.h"
#include "core/pikahistogram.h"
#include "core/pikaimage.h"
#include "core/pikatoolinfo.h"
#include "core/pikatriviallycancelablewaitable.h"
#include "core/pikawaitable.h"

#include "widgets/pikahelp-ids.h"
#include "widgets/pikahistogrambox.h"
#include "widgets/pikahistogramview.h"

#include "display/pikadisplay.h"

#include "pikahistogramoptions.h"
#include "pikathresholdtool.h"
#include "pikatooloptions-gui.h"

#include "pika-intl.h"


/*  local function prototypes  */

static void       pika_threshold_tool_finalize        (GObject           *object);

static gboolean   pika_threshold_tool_initialize      (PikaTool          *tool,
                                                       PikaDisplay       *display,
                                                       GError           **error);

static gchar    * pika_threshold_tool_get_operation   (PikaFilterTool    *filter_tool,
                                                       gchar            **description);
static void       pika_threshold_tool_dialog          (PikaFilterTool    *filter_tool);
static void       pika_threshold_tool_config_notify   (PikaFilterTool    *filter_tool,
                                                       PikaConfig        *config,
                                                       const GParamSpec  *pspec);

static gboolean   pika_threshold_tool_channel_sensitive
                                                      (gint               value,
                                                       gpointer           data);
static void       pika_threshold_tool_histogram_range (PikaHistogramView *view,
                                                       gint               start,
                                                       gint               end,
                                                       PikaThresholdTool *t_tool);
static void       pika_threshold_tool_auto_clicked    (GtkWidget         *button,
                                                       PikaThresholdTool *t_tool);


G_DEFINE_TYPE (PikaThresholdTool, pika_threshold_tool,
               PIKA_TYPE_FILTER_TOOL)

#define parent_class pika_threshold_tool_parent_class


void
pika_threshold_tool_register (PikaToolRegisterCallback  callback,
                              gpointer                  data)
{
  (* callback) (PIKA_TYPE_THRESHOLD_TOOL,
                PIKA_TYPE_HISTOGRAM_OPTIONS,
                NULL,
                0,
                "pika-threshold-tool",
                _("Threshold"),
                _("Reduce image to two colors using a threshold"),
                N_("_Threshold..."), NULL,
                NULL, PIKA_HELP_TOOL_THRESHOLD,
                PIKA_ICON_TOOL_THRESHOLD,
                data);
}

static void
pika_threshold_tool_class_init (PikaThresholdToolClass *klass)
{
  GObjectClass        *object_class      = G_OBJECT_CLASS (klass);
  PikaToolClass       *tool_class        = PIKA_TOOL_CLASS (klass);
  PikaFilterToolClass *filter_tool_class = PIKA_FILTER_TOOL_CLASS (klass);

  object_class->finalize           = pika_threshold_tool_finalize;

  tool_class->initialize           = pika_threshold_tool_initialize;

  filter_tool_class->get_operation = pika_threshold_tool_get_operation;
  filter_tool_class->dialog        = pika_threshold_tool_dialog;
  filter_tool_class->config_notify = pika_threshold_tool_config_notify;
}

static void
pika_threshold_tool_init (PikaThresholdTool *t_tool)
{
  t_tool->histogram = pika_histogram_new (FALSE);
}

static void
pika_threshold_tool_finalize (GObject *object)
{
  PikaThresholdTool *t_tool = PIKA_THRESHOLD_TOOL (object);

  g_clear_object (&t_tool->histogram);
  g_clear_object (&t_tool->histogram_async);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gboolean
pika_threshold_tool_initialize (PikaTool     *tool,
                                PikaDisplay  *display,
                                GError      **error)
{
  PikaThresholdTool *t_tool      = PIKA_THRESHOLD_TOOL (tool);
  PikaFilterTool    *filter_tool = PIKA_FILTER_TOOL (tool);
  PikaImage         *image       = pika_display_get_image (display);
  GList             *drawables;
  PikaDrawable      *drawable;
  gdouble            low;
  gdouble            high;
  gint               n_bins;

  if (! PIKA_TOOL_CLASS (parent_class)->initialize (tool, display, error))
    {
      return FALSE;
    }

  drawables = pika_image_get_selected_drawables (image);
  if (g_list_length (drawables) != 1)
    {
      if (g_list_length (drawables) > 1)
        pika_tool_message_literal (tool, display,
                                   _("Cannot modify multiple drawables. Select only one."));
      else
        pika_tool_message_literal (tool, display, _("No selected drawables."));

      g_list_free (drawables);
      return FALSE;
    }

  drawable = drawables->data;
  g_list_free (drawables);

  g_clear_object (&t_tool->histogram_async);

  g_object_get (filter_tool->config,
                "low",  &low,
                "high", &high,
                NULL);

  /* this is a hack to make sure that
   * 'pika_histogram_n_bins (t_tool->histogram)' returns the correct value for
   * 'drawable' before the asynchronous calculation of its histogram is
   * finished.
   */
  {
    GeglBuffer *temp;

    temp = gegl_buffer_new (GEGL_RECTANGLE (0, 0, 1, 1),
                            pika_drawable_get_format (drawable));

    pika_histogram_calculate (t_tool->histogram,
                              temp, GEGL_RECTANGLE (0, 0, 1, 1),
                              NULL, NULL);

    g_object_unref (temp);
  }

  n_bins = pika_histogram_n_bins (t_tool->histogram);

  t_tool->histogram_async = pika_drawable_calculate_histogram_async (
    drawable, t_tool->histogram, FALSE);
  pika_histogram_view_set_histogram (t_tool->histogram_box->view,
                                     t_tool->histogram);

  pika_histogram_view_set_range (t_tool->histogram_box->view,
                                 low  * (n_bins - 0.0001),
                                 high * (n_bins - 0.0001));

  return TRUE;
}

static gchar *
pika_threshold_tool_get_operation (PikaFilterTool  *filter_tool,
                                   gchar          **description)
{
  *description = g_strdup (_("Apply Threshold"));

  return g_strdup ("pika:threshold");
}


/**********************/
/*  Threshold dialog  */
/**********************/

static void
pika_threshold_tool_dialog (PikaFilterTool *filter_tool)
{
  PikaThresholdTool    *t_tool       = PIKA_THRESHOLD_TOOL (filter_tool);
  PikaToolOptions      *tool_options = PIKA_TOOL_GET_OPTIONS (filter_tool);
  GtkWidget            *main_vbox;
  GtkWidget            *main_frame;
  GtkWidget            *frame_vbox;
  GtkWidget            *hbox;
  GtkWidget            *label;
  GtkWidget            *hbox2;
  GtkWidget            *box;
  GtkWidget            *button;
  PikaHistogramChannel  channel;

  main_vbox = pika_filter_tool_dialog_get_vbox (filter_tool);

  main_frame = pika_frame_new (NULL);
  gtk_box_pack_start (GTK_BOX (main_vbox), main_frame, TRUE, TRUE, 0);
  gtk_widget_show (main_frame);

  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
  gtk_frame_set_label_widget (GTK_FRAME (main_frame), hbox);
  gtk_widget_show (hbox);

  label = gtk_label_new_with_mnemonic (_("Cha_nnel:"));
  pika_label_set_attributes (GTK_LABEL (label),
                             PANGO_ATTR_WEIGHT, PANGO_WEIGHT_BOLD,
                             -1);
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  t_tool->channel_menu = pika_prop_enum_combo_box_new (filter_tool->config,
                                                       "channel", -1, -1);
  pika_enum_combo_box_set_icon_prefix (PIKA_ENUM_COMBO_BOX (t_tool->channel_menu),
                                       "pika-channel");
  pika_int_combo_box_set_sensitivity (PIKA_INT_COMBO_BOX (t_tool->channel_menu),
                                      pika_threshold_tool_channel_sensitive,
                                      filter_tool, NULL);
  gtk_box_pack_start (GTK_BOX (hbox), t_tool->channel_menu, FALSE, FALSE, 0);

  gtk_label_set_mnemonic_widget (GTK_LABEL (label), t_tool->channel_menu);

  hbox2 = pika_prop_enum_icon_box_new (G_OBJECT (tool_options),
                                       "histogram-scale", "pika-histogram",
                                       0, 0);
  gtk_box_pack_end (GTK_BOX (hbox), hbox2, FALSE, FALSE, 0);

  frame_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 4);
  gtk_container_add (GTK_CONTAINER (main_frame), frame_vbox);
  gtk_widget_show (frame_vbox);

  box = pika_histogram_box_new ();
  gtk_box_pack_start (GTK_BOX (frame_vbox), box, TRUE, TRUE, 0);
  gtk_widget_show (box);

  t_tool->histogram_box = PIKA_HISTOGRAM_BOX (box);

  g_object_get (filter_tool->config,
                "channel", &channel,
                NULL);

  pika_histogram_view_set_channel (t_tool->histogram_box->view, channel);

  g_signal_connect (t_tool->histogram_box->view, "range-changed",
                    G_CALLBACK (pika_threshold_tool_histogram_range),
                    t_tool);

  g_object_bind_property (G_OBJECT (tool_options),                "histogram-scale",
                          G_OBJECT (t_tool->histogram_box->view), "histogram-scale",
                          G_BINDING_SYNC_CREATE |
                          G_BINDING_BIDIRECTIONAL);

  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
  gtk_box_pack_start (GTK_BOX (frame_vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show (hbox);

  button = gtk_button_new_with_mnemonic (_("_Auto"));
  gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  pika_help_set_help_data (button, _("Automatically adjust to optimal "
                                     "binarization threshold"), NULL);
  gtk_widget_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (pika_threshold_tool_auto_clicked),
                    t_tool);
}

static void
pika_threshold_tool_config_notify (PikaFilterTool   *filter_tool,
                                   PikaConfig       *config,
                                   const GParamSpec *pspec)
{
  PikaThresholdTool *t_tool = PIKA_THRESHOLD_TOOL (filter_tool);

  PIKA_FILTER_TOOL_CLASS (parent_class)->config_notify (filter_tool,
                                                        config, pspec);

  if (! t_tool->histogram_box)
    return;

  if (! strcmp (pspec->name, "channel"))
    {
      PikaHistogramChannel channel;

      g_object_get (config,
                    "channel", &channel,
                    NULL);

      pika_histogram_view_set_channel (t_tool->histogram_box->view,
                                       channel);
    }
  else if (! strcmp (pspec->name, "low") ||
           ! strcmp (pspec->name, "high"))
    {
      gdouble low;
      gdouble high;
      gint    n_bins;

      g_object_get (config,
                    "low",  &low,
                    "high", &high,
                    NULL);

      n_bins = pika_histogram_n_bins (t_tool->histogram);

      pika_histogram_view_set_range (t_tool->histogram_box->view,
                                     low  * (n_bins - 0.0001),
                                     high * (n_bins - 0.0001));
    }
}

static gboolean
pika_threshold_tool_channel_sensitive (gint     value,
                                       gpointer data)
{
  GList                *drawables = PIKA_TOOL (data)->drawables;
  PikaDrawable         *drawable;
  PikaHistogramChannel  channel   = value;

  if (!drawables)
    return FALSE;

  g_return_val_if_fail (g_list_length (drawables) == 1, FALSE);
  drawable = drawables->data;

  switch (channel)
    {
    case PIKA_HISTOGRAM_VALUE:
      return TRUE;

    case PIKA_HISTOGRAM_RED:
    case PIKA_HISTOGRAM_GREEN:
    case PIKA_HISTOGRAM_BLUE:
      return pika_drawable_is_rgb (drawable);

    case PIKA_HISTOGRAM_ALPHA:
      return pika_drawable_has_alpha (drawable);

    case PIKA_HISTOGRAM_RGB:
      return pika_drawable_is_rgb (drawable);

    case PIKA_HISTOGRAM_LUMINANCE:
      return pika_drawable_is_rgb (drawable);
    }

  return FALSE;
}

static void
pika_threshold_tool_histogram_range (PikaHistogramView *widget,
                                     gint               start,
                                     gint               end,
                                     PikaThresholdTool *t_tool)
{
  PikaFilterTool *filter_tool = PIKA_FILTER_TOOL (t_tool);
  gint            n_bins      = pika_histogram_n_bins (t_tool->histogram);
  gdouble         low         = (gdouble) start / (n_bins - 1);
  gdouble         high        = (gdouble) end   / (n_bins - 1);
  gdouble         config_low;
  gdouble         config_high;

  g_object_get (filter_tool->config,
                "low",  &config_low,
                "high", &config_high,
                NULL);

  if (low  != config_low ||
      high != config_high)
    {
      g_object_set (filter_tool->config,
                    "low",  low,
                    "high", high,
                    NULL);
    }
}

static void
pika_threshold_tool_auto_clicked (GtkWidget         *button,
                                  PikaThresholdTool *t_tool)
{
  PikaTool     *tool = PIKA_TOOL (t_tool);
  PikaWaitable *waitable;

  waitable = pika_trivially_cancelable_waitable_new (
    PIKA_WAITABLE (t_tool->histogram_async));

  pika_wait (tool->tool_info->pika, waitable, _("Calculating histogram..."));

  g_object_unref (waitable);

  if (pika_async_is_synced   (t_tool->histogram_async) &&
      pika_async_is_finished (t_tool->histogram_async))
    {
      PikaHistogramChannel channel;
      gint                 n_bins;
      gdouble              low;

      g_object_get (PIKA_FILTER_TOOL (t_tool)->config,
                    "channel", &channel,
                    NULL);

      n_bins = pika_histogram_n_bins (t_tool->histogram);

      low = pika_histogram_get_threshold (t_tool->histogram,
                                          channel,
                                          0, n_bins - 1);

      pika_histogram_view_set_range (t_tool->histogram_box->view,
                                     low, n_bins - 1);
    }
}
