/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikabufferview.c
 * Copyright (C) 2001-2004 Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikawidgets/pikawidgets.h"

#include "widgets-types.h"

#include "core/pika.h"
#include "core/pikabuffer.h"
#include "core/pikacontainer.h"
#include "core/pikacontext.h"
#include "core/pikaimage.h"

#include "pikacontainertreeview.h"
#include "pikacontainerview.h"
#include "pikabufferview.h"
#include "pikadnd.h"
#include "pikadocked.h"
#include "pikaeditor.h"
#include "pikahelp-ids.h"
#include "pikaview.h"
#include "pikaviewrendererbuffer.h"
#include "pikauimanager.h"
#include "pikawidgets-utils.h"

#include "pika-intl.h"


static void   pika_buffer_view_docked_iface_init (PikaDockedInterface *iface);

static void   pika_buffer_view_set_context       (PikaDocked          *docked,
                                                  PikaContext         *context);

static void   pika_buffer_view_activate_item     (PikaContainerEditor *editor,
                                                  PikaViewable        *viewable);

static void   pika_buffer_view_clipboard_changed (Pika                *pika,
                                                  PikaBufferView      *buffer_view);
static void   pika_buffer_view_view_notify       (PikaContainerView   *view,
                                                  GParamSpec          *pspec,
                                                  PikaBufferView      *buffer_view);


G_DEFINE_TYPE_WITH_CODE (PikaBufferView, pika_buffer_view,
                         PIKA_TYPE_CONTAINER_EDITOR,
                         G_IMPLEMENT_INTERFACE (PIKA_TYPE_DOCKED,
                                                pika_buffer_view_docked_iface_init))

#define parent_class pika_buffer_view_parent_class

static PikaDockedInterface *parent_docked_iface = NULL;


static void
pika_buffer_view_class_init (PikaBufferViewClass *klass)
{
  PikaContainerEditorClass *editor_class = PIKA_CONTAINER_EDITOR_CLASS (klass);

  editor_class->activate_item = pika_buffer_view_activate_item;
}

static void
pika_buffer_view_docked_iface_init (PikaDockedInterface *iface)
{
  parent_docked_iface = g_type_interface_peek_parent (iface);

  if (! parent_docked_iface)
    parent_docked_iface = g_type_default_interface_peek (PIKA_TYPE_DOCKED);

  iface->set_context = pika_buffer_view_set_context;
}

static void
pika_buffer_view_init (PikaBufferView *view)
{
}

static void
pika_buffer_view_set_context (PikaDocked  *docked,
                              PikaContext *context)
{
  PikaBufferView *view = PIKA_BUFFER_VIEW (docked);

  parent_docked_iface->set_context (docked, context);

  pika_view_renderer_set_context (PIKA_VIEW (view->clipboard_view)->renderer,
                                  context);
}


/*  public functions  */

GtkWidget *
pika_buffer_view_new (PikaViewType     view_type,
                      PikaContainer   *container,
                      PikaContext     *context,
                      gint             view_size,
                      gint             view_border_width,
                      PikaMenuFactory *menu_factory)
{
  PikaBufferView      *buffer_view;
  PikaContainerEditor *editor;
  GtkWidget           *frame;
  GtkWidget           *hbox;

  g_return_val_if_fail (PIKA_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (PIKA_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (view_size > 0 &&
                        view_size <= PIKA_VIEWABLE_MAX_PREVIEW_SIZE, FALSE);
  g_return_val_if_fail (view_border_width >= 0 &&
                        view_border_width <= PIKA_VIEW_MAX_BORDER_WIDTH,
                        FALSE);

  buffer_view = g_object_new (PIKA_TYPE_BUFFER_VIEW,
                              "view-type",         view_type,
                              "container",         container,
                              "context",           context,
                              "view-size",         view_size,
                              "view-border-width", view_border_width,
                              "menu-factory",      menu_factory,
                              "menu-identifier",   "<Buffers>",
                              "ui-path",           "/buffers-popup",
                              NULL);

  editor = PIKA_CONTAINER_EDITOR (buffer_view);

  if (PIKA_IS_CONTAINER_TREE_VIEW (editor->view))
    {
      PikaContainerTreeView *tree_view;

      tree_view = PIKA_CONTAINER_TREE_VIEW (editor->view);

      pika_container_tree_view_connect_name_edited (tree_view,
                                                    G_CALLBACK (pika_container_tree_view_name_edited),
                                                    tree_view);
    }

  frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
  gtk_box_pack_start (GTK_BOX (editor), frame, FALSE, FALSE, 0);
  gtk_box_reorder_child (GTK_BOX (editor), frame, 0);
  gtk_widget_show (frame);

  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 2);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 2);
  gtk_container_add (GTK_CONTAINER (frame), hbox);
  gtk_widget_show (hbox);

  /* FIXME: enable preview of a clipboard image, not just buffer */
  buffer_view->clipboard_view =
    pika_view_new_full_by_types (NULL,
                                 PIKA_TYPE_VIEW,
                                 PIKA_TYPE_BUFFER,
                                 view_size, view_size, view_border_width,
                                 FALSE, FALSE, TRUE);
  gtk_box_pack_start (GTK_BOX (hbox), buffer_view->clipboard_view,
                      FALSE, FALSE, 0);
  gtk_widget_show (buffer_view->clipboard_view);

  g_signal_connect_object (editor->view, "notify::view-size",
                           G_CALLBACK (pika_buffer_view_view_notify),
                           buffer_view, 0);
  g_signal_connect_object (editor->view, "notify::view-border-width",
                           G_CALLBACK (pika_buffer_view_view_notify),
                           buffer_view, 0);

  buffer_view->clipboard_label = gtk_label_new (_("(None)"));
  gtk_box_pack_start (GTK_BOX (hbox), buffer_view->clipboard_label,
                      FALSE, FALSE, 0);
  gtk_widget_show (buffer_view->clipboard_label);

  g_signal_connect_object (context->pika, "clipboard-changed",
                           G_CALLBACK (pika_buffer_view_clipboard_changed),
                           G_OBJECT (buffer_view), 0);

  pika_buffer_view_clipboard_changed (context->pika, buffer_view);

  buffer_view->paste_button =
    pika_editor_add_action_button (PIKA_EDITOR (editor->view), "buffers",
                                   "buffers-paste",
                                   "buffers-paste-in-place",
                                   pika_get_extend_selection_mask (),
                                   NULL);

  buffer_view->paste_into_button =
    pika_editor_add_action_button (PIKA_EDITOR (editor->view), "buffers",
                                   "buffers-paste-into",
                                   "buffers-paste-into-in-place",
                                   pika_get_extend_selection_mask (),
                                   NULL);

  buffer_view->paste_as_new_layer_button =
    pika_editor_add_action_button (PIKA_EDITOR (editor->view), "buffers",
                                   "buffers-paste-as-new-layer",
                                   "buffers-paste-as-new-layer-in-place",
                                   pika_get_extend_selection_mask (),
                                   NULL);

  buffer_view->paste_as_new_image_button =
    pika_editor_add_action_button (PIKA_EDITOR (editor->view), "buffers",
                                   "buffers-paste-as-new-image", NULL);

  buffer_view->delete_button =
    pika_editor_add_action_button (PIKA_EDITOR (editor->view), "buffers",
                                   "buffers-delete", NULL);

  pika_container_view_enable_dnd (editor->view,
                                  GTK_BUTTON (buffer_view->paste_button),
                                  PIKA_TYPE_BUFFER);
  pika_container_view_enable_dnd (editor->view,
                                  GTK_BUTTON (buffer_view->paste_into_button),
                                  PIKA_TYPE_BUFFER);
  pika_container_view_enable_dnd (editor->view,
                                  GTK_BUTTON (buffer_view->paste_as_new_layer_button),
                                  PIKA_TYPE_BUFFER);
  pika_container_view_enable_dnd (editor->view,
                                  GTK_BUTTON (buffer_view->paste_as_new_image_button),
                                  PIKA_TYPE_BUFFER);
  pika_container_view_enable_dnd (editor->view,
                                  GTK_BUTTON (buffer_view->delete_button),
                                  PIKA_TYPE_BUFFER);

  pika_ui_manager_update (pika_editor_get_ui_manager (PIKA_EDITOR (editor->view)),
                          editor);

  return GTK_WIDGET (buffer_view);
}


/*  private functions  */

static void
pika_buffer_view_activate_item (PikaContainerEditor *editor,
                                PikaViewable        *viewable)
{
  PikaBufferView *view = PIKA_BUFFER_VIEW (editor);
  PikaContainer  *container;

  if (PIKA_CONTAINER_EDITOR_CLASS (parent_class)->activate_item)
    PIKA_CONTAINER_EDITOR_CLASS (parent_class)->activate_item (editor, viewable);

  container = pika_container_view_get_container (editor->view);

  if (viewable && pika_container_have (container, PIKA_OBJECT (viewable)))
    {
      gtk_button_clicked (GTK_BUTTON (view->paste_button));
    }
}

static void
pika_buffer_view_clipboard_changed (Pika           *pika,
                                    PikaBufferView *buffer_view)
{
  PikaBuffer *buffer = pika_get_clipboard_buffer (pika);

  pika_view_set_viewable (PIKA_VIEW (buffer_view->clipboard_view),
                          PIKA_VIEWABLE (buffer));

  if (buffer)
    {
      gchar *desc = pika_viewable_get_description (PIKA_VIEWABLE (buffer),
                                                   NULL);
      gtk_label_set_text (GTK_LABEL (buffer_view->clipboard_label), desc);
      g_free (desc);
    }
  else
    {
      gtk_label_set_text (GTK_LABEL (buffer_view->clipboard_label), _("(None)"));
    }
}

static void
pika_buffer_view_view_notify (PikaContainerView *container_view,
                              GParamSpec        *pspec,
                              PikaBufferView    *buffer_view)
{
  PikaView *view = PIKA_VIEW (buffer_view->clipboard_view);
  gint      view_size;
  gint      view_border_width;

  view_size = pika_container_view_get_view_size (container_view,
                                                 &view_border_width);

  pika_view_renderer_set_size_full (view->renderer,
                                    view_size, view_size, view_border_width);
}
