/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikadocked.c
 * Copyright (C) 2003  Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <string.h>

#include <gegl.h>
#include <gtk/gtk.h>

#include "widgets-types.h"

#include "core/pikacontext.h"

#include "pikadocked.h"
#include "pikasessioninfo-aux.h"


enum
{
  TITLE_CHANGED,
  LAST_SIGNAL
};


/*  local function prototypes  */

static void    pika_docked_iface_set_aux_info (PikaDocked *docked,
                                               GList      *aux_info);
static GList * pika_docked_iface_get_aux_info (PikaDocked *docked);


G_DEFINE_INTERFACE (PikaDocked, pika_docked, GTK_TYPE_WIDGET)


static guint docked_signals[LAST_SIGNAL] = { 0 };


/*  private functions  */


static void
pika_docked_default_init (PikaDockedInterface *iface)
{
  docked_signals[TITLE_CHANGED] =
    g_signal_new ("title-changed",
                  PIKA_TYPE_DOCKED,
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (PikaDockedInterface, title_changed),
                  NULL, NULL, NULL,
                  G_TYPE_NONE, 0);

  iface->get_aux_info = pika_docked_iface_get_aux_info;
  iface->set_aux_info = pika_docked_iface_set_aux_info;
}

#define AUX_INFO_SHOW_BUTTON_BAR "show-button-bar"

static void
pika_docked_iface_set_aux_info (PikaDocked *docked,
                                GList      *aux_info)
{
  GList *list;

  for (list = aux_info; list; list = g_list_next (list))
    {
      PikaSessionInfoAux *aux = list->data;

      if (strcmp (aux->name, AUX_INFO_SHOW_BUTTON_BAR) == 0)
        {
          gboolean show = g_ascii_strcasecmp (aux->value, "false");

          pika_docked_set_show_button_bar (docked, show);
        }
    }
}

static GList *
pika_docked_iface_get_aux_info (PikaDocked *docked)
{
  if (pika_docked_has_button_bar (docked))
    {
      gboolean show = pika_docked_get_show_button_bar (docked);

      return g_list_append (NULL,
                            pika_session_info_aux_new (AUX_INFO_SHOW_BUTTON_BAR,
                                                       show ? "true" : "false"));
    }

  return NULL;
}


/*  public functions  */


void
pika_docked_title_changed (PikaDocked *docked)
{
  g_return_if_fail (PIKA_IS_DOCKED (docked));

  g_signal_emit (docked, docked_signals[TITLE_CHANGED], 0);
}

void
pika_docked_set_aux_info (PikaDocked *docked,
                          GList      *aux_info)
{
  PikaDockedInterface *docked_iface;

  g_return_if_fail (PIKA_IS_DOCKED (docked));

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->set_aux_info)
    docked_iface->set_aux_info (docked, aux_info);
}

GList *
pika_docked_get_aux_info (PikaDocked *docked)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), NULL);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->get_aux_info)
    return docked_iface->get_aux_info (docked);

  return NULL;
}

GtkWidget *
pika_docked_get_preview (PikaDocked  *docked,
                         PikaContext *context,
                         GtkIconSize  size)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), NULL);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->get_preview)
    return docked_iface->get_preview (docked, context, size);

  return NULL;
}

gboolean
pika_docked_get_prefer_icon (PikaDocked *docked)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), FALSE);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->get_prefer_icon)
    return docked_iface->get_prefer_icon (docked);

  return FALSE;
}

PikaUIManager *
pika_docked_get_menu (PikaDocked     *docked,
                      const gchar   **ui_path,
                      gpointer       *popup_data)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), NULL);
  g_return_val_if_fail (ui_path != NULL, NULL);
  g_return_val_if_fail (popup_data != NULL, NULL);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->get_menu)
    return docked_iface->get_menu (docked, ui_path, popup_data);

  return NULL;
}

gchar *
pika_docked_get_title (PikaDocked *docked)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), NULL);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->get_title)
    return docked_iface->get_title (docked);

  return NULL;
}

void
pika_docked_set_context (PikaDocked  *docked,
                         PikaContext *context)
{
  PikaDockedInterface *docked_iface;

  g_return_if_fail (PIKA_IS_DOCKED (docked));
  g_return_if_fail (context == NULL || PIKA_IS_CONTEXT (context));

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->set_context)
    docked_iface->set_context (docked, context);
}

gboolean
pika_docked_has_button_bar (PikaDocked *docked)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), FALSE);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->has_button_bar)
    return docked_iface->has_button_bar (docked);

  return FALSE;
}

void
pika_docked_set_show_button_bar (PikaDocked *docked,
                                 gboolean    show)
{
  PikaDockedInterface *docked_iface;

  g_return_if_fail (PIKA_IS_DOCKED (docked));

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->set_show_button_bar)
    docked_iface->set_show_button_bar (docked, show ? TRUE : FALSE);
}

gboolean
pika_docked_get_show_button_bar (PikaDocked *docked)
{
  PikaDockedInterface *docked_iface;

  g_return_val_if_fail (PIKA_IS_DOCKED (docked), FALSE);

  docked_iface = PIKA_DOCKED_GET_IFACE (docked);

  if (docked_iface->get_show_button_bar)
    return docked_iface->get_show_button_bar (docked);

  return FALSE;
}
