/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "widgets-types.h"

#include "core/pikacontext.h"
#include "core/pikaimage.h"

#include "pikadocked.h"
#include "pikaimageeditor.h"
#include "pikauimanager.h"


static void   pika_image_editor_docked_iface_init (PikaDockedInterface *iface);

static void   pika_image_editor_set_context    (PikaDocked       *docked,
                                                PikaContext      *context);

static void   pika_image_editor_dispose        (GObject          *object);

static void   pika_image_editor_real_set_image (PikaImageEditor  *editor,
                                                PikaImage        *image);
static void   pika_image_editor_image_flush    (PikaImage        *image,
                                                gboolean          invalidate_preview,
                                                PikaImageEditor  *editor);


G_DEFINE_TYPE_WITH_CODE (PikaImageEditor, pika_image_editor, PIKA_TYPE_EDITOR,
                         G_IMPLEMENT_INTERFACE (PIKA_TYPE_DOCKED,
                                                pika_image_editor_docked_iface_init))

#define parent_class pika_image_editor_parent_class


static void
pika_image_editor_class_init (PikaImageEditorClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->dispose = pika_image_editor_dispose;

  klass->set_image      = pika_image_editor_real_set_image;
}

static void
pika_image_editor_init (PikaImageEditor *editor)
{
  editor->image = NULL;

  gtk_widget_set_sensitive (GTK_WIDGET (editor), FALSE);
}

static void
pika_image_editor_docked_iface_init (PikaDockedInterface *iface)
{
  iface->set_context = pika_image_editor_set_context;
}

static void
pika_image_editor_set_context (PikaDocked  *docked,
                               PikaContext *context)
{
  PikaImageEditor *editor = PIKA_IMAGE_EDITOR (docked);
  PikaImage       *image  = NULL;

  if (editor->context)
    {
      g_signal_handlers_disconnect_by_func (editor->context,
                                            pika_image_editor_set_image,
                                            editor);

      g_object_unref (editor->context);
    }

  editor->context = context;

  if (context)
    {
      g_object_ref (editor->context);

      g_signal_connect_swapped (context, "image-changed",
                                G_CALLBACK (pika_image_editor_set_image),
                                editor);

      image = pika_context_get_image (context);
    }

  pika_image_editor_set_image (editor, image);
}

static void
pika_image_editor_dispose (GObject *object)
{
  PikaImageEditor *editor = PIKA_IMAGE_EDITOR (object);

  if (editor->image)
    pika_image_editor_set_image (editor, NULL);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
pika_image_editor_real_set_image (PikaImageEditor *editor,
                                  PikaImage       *image)
{
  if (editor->image)
    g_signal_handlers_disconnect_by_func (editor->image,
                                          pika_image_editor_image_flush,
                                          editor);

  editor->image = image;

  if (editor->image)
    g_signal_connect (editor->image, "flush",
                      G_CALLBACK (pika_image_editor_image_flush),
                      editor);

  gtk_widget_set_sensitive (GTK_WIDGET (editor), image != NULL);
}


/*  public functions  */

void
pika_image_editor_set_image (PikaImageEditor *editor,
                             PikaImage       *image)
{
  g_return_if_fail (PIKA_IS_IMAGE_EDITOR (editor));
  g_return_if_fail (image == NULL || PIKA_IS_IMAGE (image));

  if (image != editor->image)
    {
      PIKA_IMAGE_EDITOR_GET_CLASS (editor)->set_image (editor, image);

      if (pika_editor_get_ui_manager (PIKA_EDITOR (editor)))
        pika_ui_manager_update (pika_editor_get_ui_manager (PIKA_EDITOR (editor)),
                                pika_editor_get_popup_data (PIKA_EDITOR (editor)));
    }
}

PikaImage *
pika_image_editor_get_image (PikaImageEditor *editor)
{
  g_return_val_if_fail (PIKA_IS_IMAGE_EDITOR (editor), NULL);

  return editor->image;
}


/*  private functions  */

static void
pika_image_editor_image_flush (PikaImage       *image,
                               gboolean         invalidate_preview,
                               PikaImageEditor *editor)
{
  if (pika_editor_get_ui_manager (PIKA_EDITOR (editor)))
    pika_ui_manager_update (pika_editor_get_ui_manager (PIKA_EDITOR (editor)),
                            pika_editor_get_popup_data (PIKA_EDITOR (editor)));
}
