/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikaprocedureaction.c
 * Copyright (C) 2004-2016 Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gtk/gtk.h>
#include <gegl.h>

#include "widgets-types.h"

#include "libpikabase/pikabase.h"

#include "pdb/pikaprocedure.h"

#include "pikaaction.h"
#include "pikaaction-history.h"
#include "pikaprocedureaction.h"
#include "pikawidgets-utils.h"


enum
{
  PROP_0,
  PROP_PROCEDURE
};


static void   pika_procedure_action_g_action_iface_init (GActionInterface *iface);

static void   pika_procedure_action_finalize            (GObject          *object);
static void   pika_procedure_action_set_property        (GObject          *object,
                                                         guint             prop_id,
                                                         const GValue     *value,
                                                         GParamSpec       *pspec);
static void   pika_procedure_action_get_property        (GObject          *object,
                                                         guint             prop_id,
                                                         GValue           *value,
                                                         GParamSpec       *pspec);

static void   pika_procedure_action_activate            (GAction          *action,
                                                         GVariant         *parameter);


G_DEFINE_TYPE_WITH_CODE (PikaProcedureAction, pika_procedure_action, PIKA_TYPE_ACTION_IMPL,
                         G_IMPLEMENT_INTERFACE (G_TYPE_ACTION, pika_procedure_action_g_action_iface_init))

#define parent_class pika_procedure_action_parent_class


static void
pika_procedure_action_class_init (PikaProcedureActionClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->finalize      = pika_procedure_action_finalize;
  object_class->set_property  = pika_procedure_action_set_property;
  object_class->get_property  = pika_procedure_action_get_property;

  g_object_class_install_property (object_class, PROP_PROCEDURE,
                                   g_param_spec_object ("procedure",
                                                        NULL, NULL,
                                                        PIKA_TYPE_PROCEDURE,
                                                        PIKA_PARAM_READWRITE));
}

static void
pika_procedure_action_g_action_iface_init (GActionInterface *iface)
{
  iface->activate = pika_procedure_action_activate;
}

static void
pika_procedure_action_init (PikaProcedureAction *action)
{
}

static void
pika_procedure_action_finalize (GObject *object)
{
  PikaProcedureAction *action = PIKA_PROCEDURE_ACTION (object);

  g_clear_object (&action->procedure);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
pika_procedure_action_get_property (GObject    *object,
                                    guint       prop_id,
                                    GValue     *value,
                                    GParamSpec *pspec)
{
  PikaProcedureAction *action = PIKA_PROCEDURE_ACTION (object);

  switch (prop_id)
    {
    case PROP_PROCEDURE:
      g_value_set_object (value, action->procedure);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
pika_procedure_action_set_property (GObject      *object,
                                    guint         prop_id,
                                    const GValue *value,
                                    GParamSpec   *pspec)
{
  PikaProcedureAction *action = PIKA_PROCEDURE_ACTION (object);

  switch (prop_id)
    {
    case PROP_PROCEDURE:
      if (action->procedure)
        g_object_unref (action->procedure);
      action->procedure = g_value_dup_object (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
pika_procedure_action_activate (GAction  *action,
                                GVariant *parameter)
{
  PikaProcedureAction *procedure_action = PIKA_PROCEDURE_ACTION (action);

  /* Not all actions have procedures associated with them, for example
   * unused "filters-recent-[N]" actions, so check for NULL before we
   * invoke the action
   */
  if (procedure_action->procedure)
    {
      gsize hack = GPOINTER_TO_SIZE (procedure_action->procedure);

      pika_action_emit_activate (PIKA_ACTION (action),
                                 g_variant_new_uint64 (hack));

      pika_action_history_action_activated (PIKA_ACTION (action));
    }
}


/*  public functions  */

PikaProcedureAction *
pika_procedure_action_new (const gchar   *name,
                           const gchar   *label,
                           const gchar   *tooltip,
                           const gchar   *icon_name,
                           const gchar   *help_id,
                           PikaProcedure *procedure,
                           PikaContext   *context)
{
  PikaProcedureAction *action;

  action = g_object_new (PIKA_TYPE_PROCEDURE_ACTION,
                         "name",       name,
                         "label",      label,
                         "tooltip",    tooltip,
                         "icon-name",  icon_name,
                         "procedure",  procedure,
                         "context",    context,
                         NULL);

  pika_action_set_help_id (PIKA_ACTION (action), help_id);

  return action;
}
