/* LIBPIKA - The PIKA Library
 * Copyright (C) 1995-1997 Peter Mattis and Spencer Kimball
 *
 * pikaprogressbar.c
 * Copyright (C) 2004 Michael Natterer <mitch@gimp.org>
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gtk/gtk.h>

#ifdef GDK_WINDOWING_WIN32
#include <gdk/gdkwin32.h>
#endif

#ifdef GDK_WINDOWING_X11
#include <gdk/gdkx.h>
#endif

#ifdef GDK_WINDOWING_WAYLAND
#include <gdk/gdkwayland.h>
#endif

#include "pikauitypes.h"

#include "pika.h"

#include "pikaprogressbar.h"


/**
 * SECTION: pikaprogressbar
 * @title: PikaProgressBar
 * @short_description: A widget providing a progress bar.
 *
 * A widget providing a progress bar that automatically redirects any
 * progress calls to itself.
 **/


static void     pika_progress_bar_dispose    (GObject     *object);

static void     pika_progress_bar_start      (const gchar *message,
                                              gboolean     cancelable,
                                              gpointer     user_data);
static void     pika_progress_bar_end        (gpointer     user_data);
static void     pika_progress_bar_set_text   (const gchar *message,
                                              gpointer     user_data);
static void     pika_progress_bar_set_value  (gdouble      percentage,
                                              gpointer     user_data);
static void     pika_progress_bar_pulse      (gpointer     user_data);
static guint64  pika_progress_bar_get_window (gpointer     user_data);


G_DEFINE_TYPE (PikaProgressBar, pika_progress_bar, GTK_TYPE_PROGRESS_BAR)

#define parent_class pika_progress_bar_parent_class


static void
pika_progress_bar_class_init (PikaProgressBarClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->dispose = pika_progress_bar_dispose;
}

static void
pika_progress_bar_init (PikaProgressBar *bar)
{
  PikaProgressVtable vtable = { 0, };

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (bar), " ");
  gtk_progress_bar_set_ellipsize (GTK_PROGRESS_BAR (bar), PANGO_ELLIPSIZE_END);

  vtable.start      = pika_progress_bar_start;
  vtable.end        = pika_progress_bar_end;
  vtable.set_text   = pika_progress_bar_set_text;
  vtable.set_value  = pika_progress_bar_set_value;
  vtable.pulse      = pika_progress_bar_pulse;
  vtable.get_window = pika_progress_bar_get_window;

  bar->progress_callback = pika_progress_install_vtable (&vtable, bar, NULL);
}

static void
pika_progress_bar_dispose (GObject *object)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (object);

  if (bar->progress_callback)
    {
      pika_progress_uninstall (bar->progress_callback);
      bar->progress_callback = NULL;
    }

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
pika_progress_bar_start (const gchar *message,
                         gboolean     cancelable,
                         gpointer     user_data)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (user_data);

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (bar), message ? message : " ");
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (bar), 0.0);

  if (gtk_widget_is_drawable (GTK_WIDGET (bar)))
    while (gtk_events_pending ())
      gtk_main_iteration ();
}

static void
pika_progress_bar_end (gpointer user_data)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (user_data);

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (bar), " ");
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (bar), 0.0);

  if (gtk_widget_is_drawable (GTK_WIDGET (bar)))
    while (gtk_events_pending ())
      gtk_main_iteration ();
}

static void
pika_progress_bar_set_text (const gchar *message,
                            gpointer     user_data)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (user_data);

  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (bar), message ? message : " ");

  if (gtk_widget_is_drawable (GTK_WIDGET (bar)))
    while (gtk_events_pending ())
      gtk_main_iteration ();
}

static void
pika_progress_bar_set_value (gdouble  percentage,
                             gpointer user_data)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (user_data);

  if (percentage >= 0.0)
    gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (bar), percentage);
  else
    gtk_progress_bar_pulse (GTK_PROGRESS_BAR (bar));

  if (gtk_widget_is_drawable (GTK_WIDGET (bar)))
    while (gtk_events_pending ())
      gtk_main_iteration ();
}

static void
pika_progress_bar_pulse (gpointer user_data)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (user_data);

  gtk_progress_bar_pulse (GTK_PROGRESS_BAR (bar));

  if (gtk_widget_is_drawable (GTK_WIDGET (bar)))
    while (gtk_events_pending ())
      gtk_main_iteration ();
}

static guint64
pika_window_get_native_id (GtkWindow *window)
{
  GdkWindow *surface;

  g_return_val_if_fail (GTK_IS_WINDOW (window), 0);

  surface = gtk_widget_get_window (GTK_WIDGET (window));
  if (!surface) /* aka window is not yet realized */
    return 0;

#ifdef GDK_WINDOWING_WIN32
  if (GDK_IS_WIN32_WINDOW (surface))
    return GPOINTER_TO_INT (GDK_WINDOW_HWND (gtk_widget_get_window (GTK_WIDGET (window))));
#endif

#ifdef GDK_WINDOWING_X11
  if (GDK_IS_X11_WINDOW (surface))
    return GPOINTER_TO_INT (GDK_WINDOW_XID (gtk_widget_get_window (GTK_WIDGET (window))));
#endif

#ifdef GDK_WINDOWING_WAYLAND
  if (GDK_IS_WAYLAND_WINDOW (surface))
    g_debug ("Getting window ID for progress not supported on Wayland yet");
#endif

  return 0;
}

static guint64
pika_progress_bar_get_window (gpointer user_data)
{
  PikaProgressBar *bar = PIKA_PROGRESS_BAR (user_data);
  GtkWidget       *toplevel;

  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (bar));

  if (GTK_IS_WINDOW (toplevel))
    return pika_window_get_native_id (GTK_WINDOW (toplevel));

  return 0;
}

/**
 * pika_progress_bar_new:
 *
 * Creates a new #PikaProgressBar widget.
 *
 * Returns: the new widget.
 *
 * Since: 2.2
 **/
GtkWidget *
pika_progress_bar_new (void)
{
  return g_object_new (PIKA_TYPE_PROGRESS_BAR, NULL);
}
